@@ -26,7 +26,6 @@
 
 #include <glib.h>
 #include "gdip.h"
-#include "gdip_win32.h"
 #include "gdipImage.h"
 #include <string.h>
 #include <unistd.h>
@@ -36,11 +35,6 @@
 #include <fcntl.h>
 
 
-/* Pick an unlikely combination. This pixel color will be interpreted as 'transparent' 
-   when mapping Wine pixels onto a cairo image. Need to do this crude hack since there
-   is no bitmap alpha component in regular Win32 bitmaps */
-#define	UNALTERED_PIXEL		0x00959293
-
 /*
 	Those are the only pixel formats that we really support	
 */
@@ -63,6 +57,7 @@ void
 gdip_bitmap_init (GpBitmap *bitmap)
 {
 	gdip_image_init (&bitmap->image);
+
 	bitmap->image.type = imageBitmap;
 	bitmap->data.Width = 0;
 	bitmap->data.Height = 0;
@@ -70,8 +65,12 @@ gdip_bitmap_init (GpBitmap *bitmap)
 	bitmap->data.PixelFormat = Format32bppArgb;
 	bitmap->data.Scan0 = 0;
 	bitmap->data.Reserved = 0;
+	bitmap->data.ByteCount = 0;
+	bitmap->data.Bytes = NULL;
+	bitmap->data.Left = 0;
+	bitmap->data.Top = 0;
 	bitmap->cairo_format = CAIRO_FORMAT_ARGB32; 
-	
+		
 	bitmap->hBitmapDC = 0;
 	bitmap->hInitialBitmap = 0;
 	bitmap->hBitmap = 0;
@@ -85,11 +84,18 @@ gdip_bitmap_clone (GpBitmap *bitmap, GpBitmap **clonedbitmap)
 	GpBitmap *result = (GpBitmap *) GdipAlloc (sizeof (GpBitmap));	
 	memcpy (result, bitmap, sizeof (GpBitmap));
 	
-	result->data.Scan0 = (GpBitmap *) malloc (bitmap->data.Stride * bitmap->data.Height);
+	result->data.Scan0 = malloc (bitmap->data.Stride * bitmap->data.Height);
 	memcpy (result->data.Scan0, bitmap->data.Scan0, bitmap->data.Stride * bitmap->data.Height);
 	*clonedbitmap = result;
+
+	if (bitmap->data.ByteCount > 0 && bitmap->data.Bytes != NULL){
+		result->data.Bytes = malloc (bitmap->data.ByteCount);
+		if (result->data.Bytes == NULL)
+			bitmap->data.ByteCount = 0;
+		else
+			memcpy (result->data.Bytes, bitmap->data.Bytes, bitmap->data.ByteCount);
+	}
 	
-	result->image.surface = NULL;
 	
 	/*TODO: We should also copy palette info when we support it*/
 }
@@ -99,7 +105,10 @@ GpBitmap *
 gdip_bitmap_new ()
 {
 	GpBitmap *result = (GpBitmap *) GdipAlloc (sizeof (GpBitmap));
-	gdip_bitmap_init (result);
+
+	if (result)
+		gdip_bitmap_init (result);
+
 	return result;
 }
 
@@ -110,70 +119,18 @@ gdip_bitmap_new ()
 void 
 gdip_bitmap_dispose (GpBitmap *bitmap)
 {
-	if ((bitmap->data.Reserved & GBD_OWN_SCAN0) != 0)
+	if (((bitmap->data.Reserved & GBD_OWN_SCAN0) != 0) && bitmap->data.Scan0 != NULL) {
 		GdipFree (bitmap->data.Scan0);
-}
-
-
-void *
-gdip_bitmap_create_Win32_HDC (GpBitmap *bitmap)
-{
-	void * result = 0;
-	void * hdc = CreateCompatibleDC_pfn (0);
-	void * hbitmap = 0, * holdbitmap = 0;
-	void * hdcDesc = GetDC_pfn (0);
-	
-	hbitmap = CreateCompatibleBitmap_pfn (hdcDesc, bitmap->data.Width, bitmap->data.Height);
-	if (hbitmap != 0) {
-		BITMAPINFO	bmi;
-		gdip_bitmap_fill_info_header (bitmap, &bmi.bmiHeader);
-
-		SetDIBits_pfn (hdc, hbitmap, 0, bitmap->data.Height, bitmap->data.Scan0, &bmi, 0);
-		holdbitmap = SelectObject_pfn (hdc, hbitmap);
-		bitmap->hBitmapDC = hdc;
-		bitmap->hInitialBitmap = holdbitmap;
-		bitmap->hBitmap = hbitmap;
-		result = hdc;
-	} else {
-		DeleteDC_pfn (hdc);
+		bitmap->data.Scan0 = NULL;
 	}
-	ReleaseDC_pfn (0, hdcDesc);
-	return result;
-}
 
-void 
-gdip_bitmap_destroy_Win32_HDC (GpBitmap *bitmap, void *hdc)
-{
-	if (bitmap->hBitmapDC == hdc) {
-		
-		BITMAPINFO	bmi;
-		int res = 0;
-		unsigned long *array, *end;
-			
-		SelectObject_pfn (bitmap->hBitmapDC, bitmap->hInitialBitmap);
-			
-		gdip_bitmap_fill_info_header (bitmap, &bmi.bmiHeader);
-		res = GetDIBits_pfn (bitmap->hBitmapDC, bitmap->hBitmap, 0, bitmap->data.Height, bitmap->data.Scan0, &bmi, 0);
-
-		if (bitmap->cairo_format == CAIRO_FORMAT_ARGB32) {
-			array = bitmap->data.Scan0;
-			end = array + (bmi.bmiHeader.biSizeImage >> 2);
-			while (array < end) {
-				if (*array!=UNALTERED_PIXEL) {
-					*array |= 0xff000000;
-				}
-				++array;
-			}
-		}
-
-		DeleteObject_pfn (bitmap->hBitmap);
-		DeleteDC_pfn (bitmap->hBitmapDC);
-		bitmap->hBitmapDC = 0;
-		bitmap->hInitialBitmap = 0;
-		bitmap->hBitmap = 0;
+	if ((bitmap->data.ByteCount > 0) && (bitmap->data.Bytes != NULL)) {
+		GdipFree (bitmap->data.Bytes);
+		bitmap->data.Bytes = NULL;
 	}
 }
 
+
 GpStatus 
 GdipCreateBitmapFromScan0 (int width, int height, int stride, int format, void *scan0, GpBitmap **bitmap)
 {
@@ -202,6 +159,9 @@ GdipCreateBitmapFromScan0 (int width, int height, int stride, int format, void *
 	}
 
 	result = gdip_bitmap_new ();
+	if (result == NULL)
+		return OutOfMemory;
+			
 	result->data.Width = result->image.width = width;
 	result->data.Height = result->image.height = height;
 	result->data.PixelFormat = result->image.pixFormat = format;
@@ -211,11 +171,15 @@ GdipCreateBitmapFromScan0 (int width, int height, int stride, int format, void *
 		int bpp  = gdip_get_pixel_format_components (format);
 		bpp = bpp * gdip_get_pixel_format_depth (format);
 		stride = (bpp * width) / 8;
-		stride = (stride + 3) & ~3;		
+		stride = (stride + (sizeof(pixman_bits_t)-1));
+		stride &= ~(sizeof(pixman_bits_t)-1);
 	}
 
 	if (scan0 == NULL) {
 		scan0 = GdipAlloc (stride * height);
+		if (scan0 == NULL)
+			return OutOfMemory;
+		
 		memset (scan0, 0, stride * height);
 		own_scan0 = TRUE;
 	}
@@ -227,6 +191,15 @@ GdipCreateBitmapFromScan0 (int width, int height, int stride, int format, void *
 	if (own_scan0)
 		result->data.Reserved |= GBD_OWN_SCAN0;
 
+	/* Initialize the image frames */
+	if (result->image.frameDimensionCount == 0) {
+		result->image.frameDimensionCount = 1;
+			result->image.frameDimensionList = (FrameInfo *) GdipAlloc (sizeof (FrameInfo));
+			result->image.frameDimensionList[0].count = 1; /*multiple frames are already taken care of in respectic codecs*/
+			memcpy (&(result->image.frameDimensionList[0].frameDimension), &gdip_image_frameDimension_page_guid, sizeof (CLSID));
+			result->image.frameDimensionList[0].frames = &(((GpBitmap *) result)->data);
+		}
+
 	*bitmap = result;
 
 	return Ok;
@@ -238,16 +211,14 @@ GdipCreateBitmapFromGraphics (int width, int height, GpGraphics *graphics, GpBit
 	GpBitmap 		*result = 0;
 	int 			bmpSize = 0;
 	int 			stride = width;
-	unsigned long	*ptr;
-	unsigned char	*endptr;
-
 	/*
 	 * FIXME: should get the stride based on the format of the graphics object.
 	 */
-	while (stride % 4)
+	stride *= 4;
+
+	while (stride % sizeof(pixman_bits_t))
 		stride++;
 	
-	stride *= 4;
 	bmpSize = stride * height;
 	result = gdip_bitmap_new ();
 	result->data.Width = result->image.width = width;
@@ -256,12 +227,9 @@ GdipCreateBitmapFromGraphics (int width, int height, GpGraphics *graphics, GpBit
 	result->data.Stride = stride;
 
 	result->data.Scan0 = GdipAlloc (bmpSize);
-
-	ptr=result->data.Scan0;
-	endptr=result->data.Scan0+bmpSize;
-	while ((unsigned char *)ptr<endptr) {
-		ptr[0]=UNALTERED_PIXEL;
-		ptr++;
+	if (result->data.Scan0 == NULL){
+		GdipFree (result);
+		return OutOfMemory;
 	}
 
 	result->data.Reserved |= GBD_OWN_SCAN0;
@@ -323,6 +291,11 @@ GdipCloneBitmapAreaI (int x, int y, int width, int height, PixelFormat format,
 		return st;
 
 	result = gdip_bitmap_new ();
+	if (result == NULL){
+		GdipFree (bd.Scan0);
+		return OutOfMemory;
+	}
+	
 	result->cairo_format = original->cairo_format;
 	memcpy (&result->data, &bd, sizeof (GdipBitmapData));
 	result->image.pixFormat = format;	/* this is only used by image codecs */
@@ -381,8 +354,16 @@ gdip_bitmap_clone_data_rect (GdipBitmapData *srcData, Rect *srcRect, GdipBitmapD
 
 	
 	if (destData->Scan0 == NULL) {
-		destData->Stride = (((( destRect->Width * dest_components * dest_deph) /8)  + 3) & ~3);
-		destData->Scan0 = GdipAlloc (destData->Stride * destRect->Height);
+		byte *data;
+
+		destData->Stride = (((( destRect->Width * dest_components * dest_deph) /8)  + (sizeof(pixman_bits_t)-1)) & ~(sizeof(pixman_bits_t)-1));
+
+		data = GdipAlloc (destData->Stride * destRect->Height);
+		if (data == NULL) {
+			return OutOfMemory;
+		}
+		
+		destData->Scan0 = data;
 		destData->Width = destRect->Width;
 		destData->Height = destRect->Height;
 		destData->PixelFormat = srcData->PixelFormat;
@@ -477,12 +458,14 @@ gdip_bitmap_change_rect_pixel_format (GdipBitmapData *srcData, Rect *srcRect, Gd
 
 		if (destData->Scan0 == NULL) {
 			outStride = bytesPerPixel * destRect->Width;
-			while (outStride % 4)
+			while (outStride % sizeof(pixman_bits_t))
 				outStride++;		/* dword-align each row */
 
 			/* Allocate the output buffer */
 			outBuffer = GdipAlloc (outStride * destRect->Height);
-
+			if (outBuffer == NULL)
+				return OutOfMemory;
+			
 			destData->Width = destRect->Width;
 			destData->Height = destRect->Height;
 			destData->Stride = outStride;
@@ -518,12 +501,14 @@ gdip_bitmap_change_rect_pixel_format (GdipBitmapData *srcData, Rect *srcRect, Gd
 
 		if (destData->Scan0 == NULL) {
 			outStride = destBytesPerPixel * destRect->Width;
-			while (outStride % 4)
+			while (outStride % sizeof(pixman_bits_t))
 				outStride++;		/* dword-align each row */
 
 			/* Allocate the output buffer */
 			outBuffer = GdipAlloc (outStride * destRect->Height);
-
+			if (outBuffer == NULL)
+				return OutOfMemory;
+			
 			destData->Width = destRect->Width;
 			destData->Height = destRect->Height;
 			destData->Stride = outStride;
@@ -562,7 +547,6 @@ gdip_bitmap_change_rect_pixel_format (GdipBitmapData *srcData, Rect *srcRect, Gd
 				 destBytesPerPixel * destRect->Width,
 				 destRect->Height);
 			if (add_alpha) {
-				g_assert (gdip_get_pixel_format_bpp (destFormat) == 32);
 				unsigned int *ptr = (unsigned int *) outBuffer;
 				int i;
 				for (i = 0; i < destRect->Height * (outStride / sizeof(unsigned int)); i++)
@@ -593,7 +577,7 @@ gdip_bitmap_change_rect_pixel_format (GdipBitmapData *srcData, Rect *srcRect, Gd
 
 
 /* Format24bppRgb is internally stored by Cairo as a four bytes. Convert it to 3-byte (RGB) */	
-void
+int
 gdip_from_ARGB_to_RGB (BYTE *src, int width, int height, int stride, BYTE **dest, int* dest_stride)
 {
 	int x, y, len, r, g, b, a;
@@ -604,9 +588,12 @@ gdip_from_ARGB_to_RGB (BYTE *src, int width, int height, int stride, BYTE **dest
 	
 	*dest_stride = dest_components * 8;
 	*dest_stride = (*dest_stride * width) / 8;
-	*dest_stride = (*dest_stride + 3) & ~3;		
+	*dest_stride = (*dest_stride + (sizeof(pixman_bits_t)-1)) & ~(sizeof(pixman_bits_t)-1);		
+	
+	result = GdipAlloc (*dest_stride * height);
+	if (result == NULL)
+		return OutOfMemory;
 	
-	result = GdipAlloc (*dest_stride * height);	
 	memset (result, 0, *dest_stride * height);
 	
 	for (y = 0, pos_src = src, pos_dest = result; y < height; y++, pos_src += stride, pos_dest += *dest_stride) {		
@@ -617,12 +604,13 @@ gdip_from_ARGB_to_RGB (BYTE *src, int width, int height, int stride, BYTE **dest
 		}
 	}
 	
-	*dest = result;	
+	*dest = result;
+	return Ok;
 }
 
 
 /* Format24bppRgb is internally stored by Cairo as a four bytes. Convert it to 3-byte (RGB) */	
-void
+int
 gdip_from_RGB_to_ARGB (BYTE *src, int width, int height, int stride, BYTE **dest, int* dest_stride)
 {
 	int x, y, len, r, g, b, a;
@@ -633,9 +621,12 @@ gdip_from_RGB_to_ARGB (BYTE *src, int width, int height, int stride, BYTE **dest
 	
 	*dest_stride = dest_components * 8;
 	*dest_stride = (*dest_stride * width) / 8;
-	*dest_stride = (*dest_stride + 3) & ~3;		
+	*dest_stride = (*dest_stride + (sizeof(pixman_bits_t)-1)) & ~(sizeof(pixman_bits_t)-1);		
+	
+	result = GdipAlloc (*dest_stride * height);
+	if (result == NULL)
+		return OutOfMemory;
 	
-	result = GdipAlloc (*dest_stride * height);	
 	memset (result, 0, *dest_stride * height);
 	
 	for (y = 0, pos_src = src, pos_dest = result; y < height; y++, pos_src += stride, pos_dest += *dest_stride) {		
@@ -647,7 +638,8 @@ gdip_from_RGB_to_ARGB (BYTE *src, int width, int height, int stride, BYTE **dest
 		}
 	}
 	
-	*dest = result;	
+	*dest = result;
+	return Ok;
 }
 
 
@@ -657,7 +649,7 @@ GdipBitmapLockBits (GpBitmap *bitmap, Rect *srcRect, int flags, int format, Gdip
 {
 	int src_stride;
 	BYTE *src_scan0, *pos;
-	BOOL alocated = FALSE;
+	BOOL allocated = FALSE;
 
 	g_return_val_if_fail (bitmap != NULL, InvalidParameter);
 	g_return_val_if_fail (srcRect != NULL, InvalidParameter);
@@ -681,10 +673,11 @@ GdipBitmapLockBits (GpBitmap *bitmap, Rect *srcRect, int flags, int format, Gdip
 		return NotImplemented;	
 	
 	if (format == Format24bppRgb) {
-		gdip_from_ARGB_to_RGB (bitmap->data.Scan0, bitmap->data.Width, bitmap->data.Height, 
-			bitmap->data.Stride, &src_scan0, &src_stride);
+		if (gdip_from_ARGB_to_RGB (bitmap->data.Scan0, bitmap->data.Width, bitmap->data.Height, 
+					   bitmap->data.Stride, &src_scan0, &src_stride) != Ok)
+			return OutOfMemory;
 			
-		alocated = TRUE;		
+		allocated = TRUE;		
 	}
 	else {
 		src_stride = bitmap->data.Stride;
@@ -704,15 +697,22 @@ GdipBitmapLockBits (GpBitmap *bitmap, Rect *srcRect, int flags, int format, Gdip
 		result->Stride = src_stride;
 		result->PixelFormat = format;
 		result->Scan0 = GdipAlloc (src_stride * bitmap->data.Height);
+		if (result->Scan0 == NULL)
+			return OutOfMemory;
+
+		if (allocated)
+			GdipFree (src_scan0);
+		
 		memcpy (result->Scan0, src_scan0, src_stride * bitmap->data.Height);
 		
 	} else {
 		/* the user wants a subrect and/or wants a pixel format conversion */
 		GdipBitmapData convert;
-		convert.PixelFormat = format;
-		convert.Scan0 = NULL;
 		Rect destRect = {0, 0, srcRect->Width, srcRect->Height};
 		GdipBitmapData src;
+
+		convert.PixelFormat = format;
+		convert.Scan0 = NULL;
 		
 		memcpy (&src, &bitmap->data, sizeof (GdipBitmapData));
 		src.Scan0 = src_scan0;
@@ -743,7 +743,7 @@ GdipBitmapLockBits (GpBitmap *bitmap, Rect *srcRect, int flags, int format, Gdip
 	result->Reserved |= GBD_OWN_SCAN0;
 	bitmap->data.Reserved |= GBD_LOCKED;
 	
-	if (alocated)
+	if (allocated)
 		GdipFree (src_scan0);
 	
 	return Ok;
@@ -754,7 +754,7 @@ GdipBitmapUnlockBits (GpBitmap *bitmap, GdipBitmapData *bitmap_data)
 {
 	int src_stride;
 	BYTE *src_scan0, *pos;
-	BOOL alocated = FALSE;
+	BOOL allocated = FALSE;
 	
 	g_return_val_if_fail (bitmap != NULL, InvalidParameter);
 	g_return_val_if_fail (bitmap_data != NULL, InvalidParameter);
@@ -764,10 +764,11 @@ GdipBitmapUnlockBits (GpBitmap *bitmap, GdipBitmapData *bitmap_data)
 		return InvalidParameter;
 		
 	if (bitmap_data->PixelFormat == Format24bppRgb) {
-		gdip_from_RGB_to_ARGB (bitmap_data->Scan0, bitmap_data->Width, bitmap_data->Height, 
-			bitmap_data->Stride, &src_scan0, &src_stride);
+		if (gdip_from_RGB_to_ARGB (bitmap_data->Scan0, bitmap_data->Width, bitmap_data->Height, 
+					   bitmap_data->Stride, &src_scan0, &src_stride) != Ok)
+			return OutOfMemory;
 	
-		alocated = TRUE;		
+		allocated = TRUE;		
 	}
 	else {
 		src_stride = bitmap_data->Stride;
@@ -803,7 +804,7 @@ GdipBitmapUnlockBits (GpBitmap *bitmap, GdipBitmapData *bitmap_data)
 	/* unlock the bitmap */
 	bitmap->data.Reserved &= ~GBD_LOCKED;
 		
-	if (alocated)
+	if (allocated)
 		GdipFree (src_scan0);
 	
 	return Ok;
@@ -832,27 +833,48 @@ GdipBitmapSetPixel (GpBitmap *bitmap, int x, int y, ARGB color)
 	switch (data->PixelFormat) {
 		case Format24bppRgb:
 			v += x * 4;
+#ifdef WORDS_BIGENDIAN
+			v [3] = color & 0xff;
+			v [2] = (color >> 8) & 0xff;
+			v [1] = (color >> 16) & 0xff;
+			v [0] = 0xff;
+#else
 			v [0] = color & 0xff;
 			v [1] = (color >> 8) & 0xff;
 			v [2] = (color >> 16) & 0xff;
 			v [3] = 0xff;
+#endif
 			break;
 			
 		case Format32bppArgb:
 		case Format32bppPArgb:
 			v += x * 4;
+#ifdef WORDS_BIGENDIAN
+			v [3] = color & 0xff;
+			v [2] = (color >> 8) & 0xff;
+			v [1] = (color >> 16) & 0xff;
+			v [0] = color >> 24;
+#else
 			v [0] = color & 0xff;
 			v [1] = (color >> 8) & 0xff;
 			v [2] = (color >> 16) & 0xff;
 			v [3] = color >> 24;
+#endif
 			break;
 			
 		case Format32bppRgb:
 			v += x * 4;
+#ifdef WORDS_BIGENDIAN
+			v [3] = color & 0xff;
+			v [2] = (color >> 8) & 0xff;
+			v [1] = (color >> 16) & 0xff;
+			v [0] = 0xff;
+#else
 			v [0] = color & 0xff;
 			v [1] = (color >> 8) & 0xff;
 			v [2] = (color >> 16) & 0xff;
 			v [3] = 0xff;
+#endif
 			break;
 		default:
 			return NotImplemented;
@@ -924,7 +946,7 @@ gdip_bitmap_ensure_surface (GpBitmap *bitmap)
 		switch (bitmap->data.PixelFormat) {
 		case Format24bppRgb:
 			bitmap->image.surface = cairo_surface_create_for_image
-				(bitmap->data.Scan0,
+				((char *)bitmap->data.Scan0,
 				 CAIRO_FORMAT_RGB24,
 				 bitmap->data.Width,
 				 bitmap->data.Height,
@@ -934,7 +956,7 @@ gdip_bitmap_ensure_surface (GpBitmap *bitmap)
 		case Format32bppRgb:
 		case Format32bppPArgb:
 			bitmap->image.surface = cairo_surface_create_for_image
-				(bitmap->data.Scan0,
+				((char *)bitmap->data.Scan0,
 				 CAIRO_FORMAT_ARGB32,
 				 bitmap->data.Width,
 				 bitmap->data.Height,
@@ -948,4 +970,3 @@ gdip_bitmap_ensure_surface (GpBitmap *bitmap)
 
 	return bitmap->image.surface;
 }
-
