@@ -91,6 +91,17 @@ gdip_bitmap_fill_info_header (GpBitmap *bitmap, PBITMAPINFOHEADER bmi)
 {
 	int  bitmapLen = bitmap->data.Stride * bitmap->data.Height;
 	memset (bmi, 0, sizeof (BITMAPINFOHEADER));
+#ifdef WORDS_BIGENDIAN
+	bmi->biSize = GUINT32_FROM_LE (sizeof (BITMAPINFOHEADER));
+	bmi->biWidth = GULONG_FROM_LE (bitmap->data.Width);
+	bmi->biHeight = GULONG_FROM_LE (bitmap->data.Height);
+	bmi->biPlanes = GUINT16_FROM_LE (1);
+	bmi->biBitCount = GUINT16_FROM_LE (32); /*PIXEL_FORMAT_BPP (bitmap->image.pixFormat); */
+	bmi->biCompression = GUINT32_FROM_LE (BI_RGB);
+	bmi->biSizeImage =  GUINT32_FROM_LE (0); /* Many tools expect this may be set to zero for BI_RGB bitmaps */
+	bmi->biXPelsPerMeter = GULONG_FROM_LE ((int) (0.5f + ((gdip_get_display_dpi() * 3937) / 100)));
+	bmi->biYPelsPerMeter = GULONG_FROM_LE ((int) (0.5f + ((gdip_get_display_dpi() * 3937) / 100))); /* 1 meter is = 39.37 */       
+#else
 	bmi->biSize = sizeof (BITMAPINFOHEADER);
 	bmi->biWidth = bitmap->data.Width;
 	bmi->biHeight = bitmap->data.Height;
@@ -100,6 +111,7 @@ gdip_bitmap_fill_info_header (GpBitmap *bitmap, PBITMAPINFOHEADER bmi)
 	bmi->biSizeImage =  0; /* Many tools expect this may be set to zero for BI_RGB bitmaps */
 	bmi->biXPelsPerMeter = (int) (0.5f + ((gdip_get_display_dpi() * 3937) / 100));
 	bmi->biYPelsPerMeter = (int) (0.5f + ((gdip_get_display_dpi() * 3937) / 100)); /* 1 meter is = 39.37 */       
+#endif
 }                                                           
 
 void 
@@ -139,54 +151,77 @@ gdip_load_bmp_image_from_stream_delegate (GetBytesDelegate getBytesFunc,
 							image, FALSE);
 }
 
+#define palette_lookup(x)	img->image.palette->Entries[(x)]
+
 GpStatus 
 gdip_read_bmp_image_from_file_stream (void *pointer, GpImage **image, bool useFile)
 {
 	BITMAPFILEHEADER bmfh;
         BITMAPINFOHEADER bmi;
-        GpBitmap *img = NULL;
-        guchar *pixels = NULL, *linep = NULL;
-        int i, j, format, colours;
-        BOOL os2format = FALSE, upsidedown = TRUE;
-        int size, size_read;
-	byte* data_read;
+        GpBitmap	*img = NULL;
+        guchar		*pixels = NULL;
+        guchar		*linep = NULL;
+        int		i;
+	int		j;
+	int		format;
+	int		colours;
+        BOOL		os2format = FALSE;
+	BOOL		upsidedown = TRUE;
+        int		size;
+	int		size_read;
+	byte		*data_read;
+	int		line;
+	int		loop;
+	long		pixel;
+	long		index;
 		
 	size = sizeof(bmfh);
 	data_read = (byte*) GdipAlloc(size);
 	memset (data_read, 0, size);
 	size_read = gdip_read_bmp_data (pointer, data_read, size, useFile);
-	if (size_read < size)
+	if (size_read < size) {
+		GdipFree(data_read);
 		return InvalidParameter;
+	}
 	
 	bmfh.bfType = ((data_read[1]<<8)|data_read[0]);
-	if (bmfh.bfType != BFT_BITMAP )
+	if (bmfh.bfType != BFT_BITMAP) {
+		GdipFree(data_read);
 		return UnknownImageFormat;
+	}
 
 	bmfh.bfSize = (data_read[5]<<24 | data_read[4]<<16 | data_read[3]<<8 | data_read[2]);
 	bmfh.bfReserved1 = ((data_read[7]<<8)|data_read[6]);
 	bmfh.bfReserved1 = ((data_read[9]<<8)|data_read[8]);
 	bmfh.bfOffBits = (data_read[13]<<24 | data_read[12]<<16 | data_read[11]<<8 | data_read[10]);
-
 	GdipFree(data_read);
 
 	size = sizeof(DWORD);
 	data_read = (byte*)GdipAlloc(size);
 	size_read = gdip_read_bmp_data (pointer, data_read, size, useFile);
-	if (size_read < size)
+
+	if (size_read < size) {
+		GdipFree(data_read);
 		return InvalidParameter;
+	}
+
 	bmi.biSize = (data_read[3]<<24 | data_read[2]<<16 | data_read[1]<<8 | data_read[0]);
 
 	if (bmi.biSize > BITMAPCOREHEADER_SIZE){   /* New Windows headers can be bigger */ 
 		memset (data_read, 0, size);
 		size_read = gdip_read_bmp_data (pointer, data_read, size, useFile);
-		if (size_read < size)
+		if (size_read < size) {
+			GdipFree(data_read);
 			return InvalidParameter;
+		}
 		bmi.biWidth = (data_read[3]<<24 | data_read[2]<<16 | data_read[1]<<8 | data_read[0]);
 
 		memset (data_read, 0, size);
 		size_read = gdip_read_bmp_data (pointer, data_read, size, useFile);
-		if (size_read < size)
+		if (size_read < size) {
+			GdipFree(data_read);
 			return InvalidParameter;
+		}
 		bmi.biHeight = (data_read[3]<<24 | data_read[2]<<16 | data_read[1]<<8 | data_read[0]);
  	}
         else  {
@@ -194,31 +229,41 @@ gdip_read_bmp_image_from_file_stream (void *pointer, GpImage **image, bool useFi
 			/* Old OS/2 format. Width and Height fields are WORDs instead of DWORDS */
                         memset (data_read, 0, size);
 			size_read = gdip_read_bmp_data (pointer, data_read, size, useFile);
-			if (size_read < size)
+			if (size_read < size) {
+				GdipFree(data_read);
 				return InvalidParameter;
+			}
 			bmi.biWidth = (data_read[1]<<8 | data_read[0]);
 			bmi.biHeight = (data_read[3]<<8 | data_read[2]);
                         os2format = TRUE;
                 }
-                else
+                else {
+			GdipFree(data_read);
                         return UnknownImageFormat;
+		}
         }
 
 	memset (data_read, 0, size);
 	size_read = gdip_read_bmp_data (pointer, data_read, size, useFile);
-	if (size_read < size)
+	if (size_read < size) {
+		GdipFree(data_read);
 		return InvalidParameter;
+	}
 	bmi.biPlanes = (data_read[1]<<8 | data_read[0]); 
 	bmi.biBitCount = (data_read[3]<<8 | data_read[2]); 
 
 	memset (data_read, 0, size);
 	size_read = gdip_read_bmp_data (pointer, data_read, size, useFile);
-	if (size_read < size)
+	if (size_read < size) {
+		GdipFree(data_read);
 		return InvalidParameter;
+	}
 	bmi.biCompression = (data_read[3]<<24 | data_read[2]<<16 | data_read[1]<<8 | data_read[0]);
 		
-        if (bmi.biCompression == BI_RLE4 || bmi.biCompression == BI_RLE8)
+        if (bmi.biCompression == BI_RLE4 || bmi.biCompression == BI_RLE8) {
+		GdipFree(data_read);
                 return NotImplemented; /* We do not support RLE for now*/
+	}
 
         if (bmi.biHeight < 0) { /* Negative height indicates that the bitmap is sideup*/
                 upsidedown = FALSE;
@@ -227,56 +272,66 @@ gdip_read_bmp_image_from_file_stream (void *pointer, GpImage **image, bool useFi
 
 	memset (data_read, 0, size);
 	size_read = gdip_read_bmp_data (pointer, data_read, size, useFile);
-	if (size_read < size)
+	if (size_read < size) {
+		GdipFree(data_read);
 		return InvalidParameter;
+	}
 	bmi.biSizeImage = (data_read[3]<<24 | data_read[2]<<16 | data_read[1]<<8 | data_read[0]);
 		
 	memset (data_read, 0, size);
 	size_read = gdip_read_bmp_data (pointer, data_read, size, useFile);
-	if (size_read < size)
+	if (size_read < size) {
+		GdipFree(data_read);
 		return InvalidParameter;
+	}
 	bmi.biXPelsPerMeter = (data_read[3]<<24 | data_read[2]<<16 | data_read[1]<<8 | data_read[0]);
 	
 	memset (data_read, 0, size);
 	size_read = gdip_read_bmp_data (pointer, data_read, size, useFile);
-	if (size_read < size)
+	if (size_read < size) {
+		GdipFree(data_read);
 		return InvalidParameter;
+	}
 	bmi.biYPelsPerMeter = (data_read[3]<<24 | data_read[2]<<16 | data_read[1]<<8 | data_read[0]);
 		
 	memset (data_read, 0, size);
 	size_read = gdip_read_bmp_data (pointer, data_read, size, useFile);
-	if (size_read < size)
+	if (size_read < size) {
+		GdipFree(data_read);
 		return InvalidParameter;
+	}
 	bmi.biClrUsed = (data_read[3]<<24 | data_read[2]<<16 | data_read[1]<<8 | data_read[0]);
 		
 	memset (data_read, 0, size);
 	size_read = gdip_read_bmp_data (pointer, data_read, size, useFile);
-	if (size_read < size)
+	if (size_read < size) {
+		GdipFree(data_read);
 		return InvalidParameter;
+	}
+
 	bmi.biClrImportant = (data_read[3]<<24 | data_read[2]<<16 | data_read[1]<<8 | data_read[0]);
 		
 	colours =  (bmi.biClrUsed == 0 && bmi.biBitCount <= 8) ? (1 << bmi.biBitCount) : bmi.biClrUsed;
         
         format = gdip_get_pixelformat (bmi.biBitCount);
                                                
-        if (format == 0)
-                return NotImplemented; /* Unsuported pixel format*/
-
         img = gdip_bitmap_new ();
-        img->image.pixFormat = format;
         img->image.type = imageBitmap;
-        img->image.graphics = 0;
         img->image.width = bmi.biWidth;
         img->image.height = bmi.biHeight;
 
-        img->data.PixelFormat = img->image.pixFormat;
+        img->image.pixFormat = img->data.PixelFormat = Format32bppArgb;
         img->data.Width = img->image.width;
         img->data.Height = img->image.height;
-        img->data.Stride = (bmi.biBitCount * img->image.width) / 8;
+
+	// We always assume 32 bit and translate into 32 bit from source format
+        img->data.Stride = (32 * img->image.width) / 8;
         img->data.Stride = (img->data.Stride + 3) & ~3;
+
+	GdipFree(data_read);
      
-        if (colours){               
-                img->image.palette = g_malloc (sizeof(ColorPalette) + sizeof(ARGB) * colours);
+        if (colours) {
+                img->image.palette = GdipAlloc (sizeof(ColorPalette) + sizeof(ARGB) * colours);
                 img->image.palette->Flags = 0;
                 img->image.palette->Count = colours;
                        
@@ -287,10 +342,14 @@ gdip_read_bmp_image_from_file_stream (void *pointer, GpImage **image, bool useFi
                         for (i = 0; i < colours; i++) {
 				memset(data_read, 0, size);
 				size_read = gdip_read_bmp_data (pointer, data_read, size, useFile);
-				if (size_read < size)
+				if (size_read < size) {
+					GdipFree(data_read);
 					return InvalidParameter;
-				img->image.palette->Entries[i] = (((data_read[2]&0xff)<<16) | 
-					((data_read[1]&0xff)<<8) | (data_read[0]&0xff));
+				}
+				img->image.palette->Entries[i] = 
+					(((data_read[0]&0xff)<<16) | 	// R
+					((data_read[1]&0xff)<<8) | 	// G
+					(data_read[2]&0xff));		// B
                         }
 			GdipFree(data_read);
                 }
@@ -298,60 +357,179 @@ gdip_read_bmp_image_from_file_stream (void *pointer, GpImage **image, bool useFi
 			size = sizeof(byte)*4;
 			data_read = (byte*) GdipAlloc(size);
                         for (i = 0; i < colours; i++) {
-				memset(data_read, 0, size);
                                 size_read = gdip_read_bmp_data (pointer, data_read, size, useFile);
-				if (size_read < size)
+				if (size_read < size) {
 					return InvalidParameter;
-                                img->image.palette->Entries[i] = (((data_read[3]&0xff)<<24)  | 
-					((data_read[2]&0xff)<<16) | ((data_read[1]&0xff)<<8) | 
-						(data_read[0]& 0xff));                       
+				}
+
+                                img->image.palette->Entries[i] = 
+					(((data_read[0]&0xff)<<16)  | 	// R
+					((data_read[1]&0xff)<<8) | 	// G
+					((data_read[2]&0xff)) | 	// B
+					((data_read[3]& 0xff)<<24));	// Alpha
                         }
 			GdipFree(data_read);
                 }
-        }
-        else
-                img->image.palette = NULL;                                                                      
+        } else {
+                img->image.palette = NULL;
+	}
        
         pixels = GdipAlloc (img->data.Stride * img->data.Height);
 
-	size = img->data.Stride;
+	// Size contains the size of the lines on disk
+	switch (bmi.biBitCount) {
+		case 1: {
+			size = ((img->image.width + 31) & ~31) / 8;
+			loop = img->image.width / 8;
+			// we assume 0s for 1 bit, only setting 1s
+			memset(pixels, 0, img->data.Stride * img->data.Height);
+			break;
+		}
+
+		case 4: {
+			size = ((bmi.biBitCount * img->image.width + 31) & ~31) / 8;
+			loop = ((img->image.width + 1) &~1) / 2;
+			break;
+		}
+
+		case 8: {
+			size = (((bmi.biBitCount * img->image.width) + 31) & ~31) / 8;
+			loop = img->image.width;
+			break;
+		}
+
+		case 24: {
+			size = (((bmi.biBitCount * img->image.width) + 31) & ~31) / 8;
+			loop = img->image.width * 3;
+			break;
+		}
+
+		default: {
+			size = (((bmi.biBitCount * img->image.width) + 31) & ~31) / 8;
+			loop = size; 
+			break;
+		}
+	}
+
 	data_read = (byte*) GdipAlloc(size);
-        if (upsidedown) { 
-		for (i = img->data.Height - 1; i >= 0; i--){ 
-			memset(data_read, 0, size);
-			size_read = gdip_read_bmp_data (pointer, data_read, size, useFile);
-			if (size_read < size)
-				return InvalidParameter;                        
-			memcpy(pixels + i*size, data_read, size);			
+
+	for (i = 0; i < img->data.Height; i++){ 
+		if (upsidedown) {
+			line = img->data.Height - i - 1;
+		} else {
+			line = i;
 		}
-        }
-        else {
-		for (i = 0; i < img->data.Height; i++) {
-			memset(data_read, 0, size);
-			size_read = gdip_read_bmp_data (pointer, data_read, size, useFile);
-			if (size_read < size)
-				return InvalidParameter;                        			
-                        memcpy(pixels + i*size, data_read, size);			
+
+		size_read = gdip_read_bmp_data (pointer, data_read, size, useFile);
+
+		if (size_read < size) {
+			GdipFree(data_read);
+			return InvalidParameter;
 		}
-        }
+
+		switch(bmi.biBitCount) {
+			case 1: {
+				int	c;
+				int	bit;
+
+				for (c = 0; c < loop; c++) {
+					for (bit = 0; bit < 8; bit++) {
+						index = (line * img->data.Stride) + (c*8 + bit) * 4;
+
+						if (((data_read[c] << bit) & 0x80) != 0) {
+							set_pixel_bgra(pixels, index, 0xff, 0xff, 0xff, 0xff);
+						} else {
+							set_pixel_bgra(pixels, index, 0x00, 0x00, 0x00, 0xff);
+						}
+					}
+				}
+
+				for (bit = 0; bit < img->image.width % 8; bit++) {
+					index = (line * img->data.Stride) + (c*8 + bit) * 4;
+
+					if (((data_read[c] << bit) & 0x80) != 0) {
+						set_pixel_bgra(pixels, index, 0xff, 0xff, 0xff, 0xff);
+					} else {
+						set_pixel_bgra(pixels, index, 0x00, 0x00, 0x00, 0xff);
+					}
+				}
+				continue;
+			}
+
+			case 4: {
+				int c;
+
+				for (c = 0; c < loop; c++) {
+					pixel = palette_lookup((data_read[c] & 0xf0) >> 4);
+
+					index = (line * img->data.Stride) + c*8;
+
+					set_pixel_bgra(pixels, index, (pixel & 0xff0000) >> 16, (pixel & 0xff00) >> 8, pixel & 0xff, 0xff);
+
+					pixel = palette_lookup(data_read[c] & 0xf);
+
+					set_pixel_bgra(pixels, index+4, (pixel & 0xff0000) >> 16, (pixel & 0xff00) >> 8, pixel & 0xff, 0xff);
+				}
+				continue;
+			}
+
+			case 8: {
+				int	c;
+
+				for (c = 0; c < loop; c++) {
+					pixel = palette_lookup(data_read[c]);
+
+					index = (line * img->data.Stride) + c*4;
+
+					set_pixel_bgra(pixels, index, (pixel & 0xff0000) >> 16, (pixel & 0xff00) >> 8, pixel & 0xff, 0xff);
+				}
+				continue;
+			}
+
+			case 24: {
+				int	src;
+				int	dest;
+
+				src = 0;
+				dest = 0;
+
+				while (src < loop) {
+					index = (line * img->data.Stride);
+					set_pixel_bgra(pixels, index+dest, data_read[src+0], data_read[src+1], data_read[src+2], 0xff);
+					dest += 4;
+					src += 3;
+				}
+				continue;
+			}
+
+			case 32: {
+				int	src;
+				int	dest;
+
+				src = 0;
+				dest = 0;
+
+				while (src < loop) {
+					index = (line * img->data.Stride);
+					set_pixel_bgra(pixels, index+dest, data_read[src+0], data_read[src+1], data_read[src+2], data_read[src+3]);
+					dest += 4;
+					src += 4;
+				}
+				continue;
+			}
+		}
+	}
 
 	GdipFree(data_read);
 
-	/*Cairo stores internally RGB24 as RGB32 */
-	if (bmi.biBitCount == 24) {
-		BYTE *dest;
-		int dest_stride;
-
-		gdip_from_RGB_to_ARGB (pixels, img->image.width, img->image.height,
-						img->data.Stride, &dest, &dest_stride);
-		GdipFree (pixels);
-		pixels = dest;
-		img->data.Stride = dest_stride;
+	if (img->image.palette) {
+		GdipFree(img->image.palette);
+		img->image.palette = NULL;
 	}
 
 	img->data.Scan0 = pixels;
         img->data.Reserved = GBD_OWN_SCAN0;
-        img->image.surface = cairo_surface_create_for_image (pixels,
+        img->image.surface = cairo_surface_create_for_image ((char *)pixels,
                                                          img->cairo_format,
                                                          img->image.width,
                                                          img->image.height,
@@ -365,17 +543,34 @@ gdip_read_bmp_image_from_file_stream (void *pointer, GpImage **image, bool useFi
         img->image.propItems = NULL;
 
         *image = (GpImage *) img;
-
         return Ok;
 }
 
 int 
 gdip_read_bmp_data (void *pointer, byte *data, int size, bool useFile)
 {
-	if (useFile)
+	if (useFile) {
 		return fread (data, 1, size, (FILE*) pointer);
-	else
-		return ((GetBytesDelegate)(pointer))(data, size, 0);
+	} else {
+		// Streams are not required to return the number of bytes
+		// requested, they could return less yet our code seems to assume
+		// it will always get what it's asking for; lets loop until we
+		// get what was requested or we get an error
+		int got;
+		int total;
+
+		total = 0;
+		
+		do {
+			got = ((GetBytesDelegate)(pointer))(data + total, size - total, 0);
+			if (got < 1) {	// 0 = end of stream, -1 = error
+				return total;
+			}
+			total += got;
+		} while (total < size);
+
+		return total;
+	}
 }
 
 GpStatus 
@@ -408,11 +603,18 @@ gdip_save_bmp_image_to_file_stream (void *pointer,
         if (bitmap->image.palette)
                 colours = bitmap->image.palette->Count;
 
+#ifdef WORDS_BIGENDIAN
+	bmfh.bfReserved1 = bmfh.bfReserved2 = GUINT16_FROM_LE (0);
+	bmfh.bfType = GUINT16_FROM_LE (BFT_BITMAP);
+	bmfh.bfOffBits = GUINT32_FROM_LE (14 + 40 + colours * 4);
+	bmfh.bfSize = GUINT32_FROM_LE (bmfh.bfOffBits + bitmapLen);        
+#else
         bmfh.bfReserved1 = bmfh.bfReserved2 = 0;
         bmfh.bfType = BFT_BITMAP;
+        bmfh.bfOffBits = (14 + 40 + colours * 4);
         bmfh.bfSize = (bmfh.bfOffBits + bitmapLen);
 
-        bmfh.bfOffBits = (14 + 40 + colours * 4);
+#endif
         gdip_write_bmp_data (pointer, (byte *)&bmfh, sizeof (bmfh), useFile);
         
 	gdip_bitmap_fill_info_header (bitmap, &bmi);
@@ -423,6 +625,16 @@ gdip_save_bmp_image_to_file_stream (void *pointer,
                 /* Write palette on disk on BGRA*/
                 for (i = 0; bitmap->image.palette->Count; i++) {
                         color = bitmap->image.palette->Entries[i];
+#ifdef WORDS_BIGENDIAN
+                        b = color >> 24;
+			gdip_write_bmp_data (pointer, &b, 1, useFile);
+                        b = (color >> 16) & 0xff;
+			gdip_write_bmp_data (pointer, &b, 1, useFile);
+                        b = (color >> 8) & 0xff;
+			gdip_write_bmp_data (pointer, &b, 1, useFile);
+                        b =  color & 0xff;
+			gdip_write_bmp_data (pointer, &b, 1, useFile);
+#else
                         b =  color & 0xff;
 			gdip_write_bmp_data (pointer, &b, 1, useFile);
                         b = (color >> 8) & 0xff;
@@ -431,12 +643,33 @@ gdip_save_bmp_image_to_file_stream (void *pointer,
 			gdip_write_bmp_data (pointer, &b, 1, useFile);
                         b = color >> 24;
 			gdip_write_bmp_data (pointer, &b, 1, useFile);
+#endif
                 }	
         }
         
         /* Writes bitmap upside down. Many tools can only process bmp stored this way*/        
-        for (i = bitmap->data.Height - 1; i >= 0; i--)
+#ifdef WORDS_BIGENDIAN
+	{
+		int j;
+		guchar *row_pointer = GdipAlloc (image->width * 4);
+		
+		for (i = image->height -1; i >= 0; i--) {
+			for (j = 0; j < image->width; j++) {
+				row_pointer[j*4] = *((guchar *)bitmap->data.Scan0 + (bitmap->data.Stride * i) + (j*4) + 3); 
+				row_pointer[j*4+1] = *((guchar *)bitmap->data.Scan0 + (bitmap->data.Stride * i) + (j*4) + 2); 
+				row_pointer[j*4+2] = *((guchar *)bitmap->data.Scan0 + (bitmap->data.Stride * i) + (j*4) + 1); 
+				row_pointer[j*4+3] = *((guchar *)bitmap->data.Scan0 + (bitmap->data.Stride * i) + (j*4) + 0); 
+			}
+			gdip_write_bmp_data (pointer, row_pointer, bitmap->data.Stride, useFile);
+		}
+		GdipFree (row_pointer);
+	}
+#else
+	
+        for (i = bitmap->data.Height - 1; i >= 0; i--) {
 		gdip_write_bmp_data (pointer, bitmap->data.Scan0 + i *bitmap->data.Stride, bitmap->data.Stride, useFile);
+	}
+#endif
         return Ok;
 }
 
