@@ -1,5 +1,6 @@
 /*
  * Copyright (c) 2004 Ximian
+ * Copyright (c) 2004-2005 Novell, Inc.
  * 
  * Permission is hereby granted, free of charge, to any person obtaining a copy of this software 
  * and associated documentation files (the "Software"), to deal in the Software without restriction, 
@@ -19,15 +20,15 @@
  * Author:
  *          Jordi Mas i Hernandez <jordi@ximian.com>, 2004
  */
-
+#include <stdio.h>
 #include "gdip.h"
-#include "gdip_win32.h"
 #include <math.h>
 #include <glib.h>
 #include <freetype/tttables.h>
+#include <pthread.h>
 
 
-/* Generic fonts */
+/* Generic fonts families */
 static GpFontFamily* familySerif = NULL;
 static GpFontFamily* familySansSerif = NULL;
 static GpFontFamily* familyMonospace = NULL;
@@ -35,9 +36,38 @@ static int ref_familySerif = 0;
 static int ref_familySansSerif = 0;
 static int ref_familyMonospace = 0;
 
+/*
+	How font cache works
+
+	- Create any font and store it in the cache
+	- When GdipDeleteFont is called just decrease the reference count and do not delete the font
+	- All the cached fonts are released when GDI+ is shutdown
+	- When the cache is full we release the first font cached with 0 references
+*/
+
+#define MAX_CACHED_FONTS 128
+static GpCachedFont cached_fonts [MAX_CACHED_FONTS];
+static int cached_fonts_index = 0;
+static pthread_mutex_t	fontcache_mutex = PTHREAD_MUTEX_INITIALIZER;
+
+#define	LOCK_FONTCACHE		pthread_mutex_lock(&fontcache_mutex)
+#define	UNLOCK_FONTCACHE	pthread_mutex_unlock(&fontcache_mutex)
 
 /* Family and collections font functions */
 
+void
+gdip_createFontFamily (GpFontFamily **family)
+{
+	GpFontFamily *result;
+
+	result = (GpFontFamily *) GdipAlloc (sizeof (GpFontFamily));
+	result->height = -1;
+	result->linespacing = -1;
+	result->celldescent = -1;
+	result->cellascent = -1;
+	*family = result;
+}
+
 GpStatus
 GdipNewInstalledFontCollection (GpFontCollection **font_collection)
 {	
@@ -142,8 +172,7 @@ GdipDeleteFontFamily (GpFontFamily *fontFamily)
 		ref_familyMonospace--;
 		if (ref_familyMonospace)
 			delete = FALSE;
-	}
-	
+	}	
 	
 	if (delete) {
 		FcPatternDestroy (fontFamily->pattern);
@@ -201,7 +230,7 @@ GdipGetFontCollectionFamilyList (GpFontCollection *font_collection, int num_soug
 		
 
 	for (i = 0; i < font_collection->fontset->nfont; gpfam++, pattern++, i++) {
-		*gpfam = (GpFontFamily *) GdipAlloc (sizeof (GpFontFamily));
+		gdip_createFontFamily (gpfam);
 		(*gpfam)->pattern = *pattern;
 		(*gpfam)->allocated = FALSE;
 	}
@@ -242,7 +271,7 @@ GdipCreateFontFamilyFromName (GDIPCONST WCHAR *name, GpFontCollection *font_coll
 		FcConfigSubstitute (0, pat, FcMatchPattern);
 		FcDefaultSubstitute (pat);                  
 		
-		*fontFamily = (GpFontFamily *) GdipAlloc (sizeof (GpFontFamily));
+		gdip_createFontFamily (fontFamily);
 		(*fontFamily)->pattern =  FcFontMatch (0, pat, &rlt);
 		(*fontFamily)->allocated = TRUE;
 
@@ -257,8 +286,8 @@ GdipCreateFontFamilyFromName (GDIPCONST WCHAR *name, GpFontCollection *font_coll
 	for (i=0; i < font_collection->fontset->nfont; gpfam++, i++){
 		FcResult r = FcPatternGetString (*gpfam, FC_FAMILY, 0, &str);
 
-		if (strcmp (string, str)==0) {
-			*fontFamily = (GpFontFamily *) GdipAlloc (sizeof (GpFontFamily));
+		if (strcmp ((char *)string, (const char *)str)==0) {
+			gdip_createFontFamily (fontFamily);
 			(*fontFamily)->pattern = *gpfam;
 			(*fontFamily)->allocated = FALSE;
 			g_free (string);
@@ -347,24 +376,33 @@ GdipGetEmHeight (GDIPCONST GpFontFamily *family, GpFontStyle style, short *EmHei
 	if (!family || !EmHeight)
 		return InvalidParameter;
 
+	if (family->height != -1) {
+		*EmHeight = family->height;
+		return Ok;
+	}
+
 	GdipCreateFont (family, 0.0f, style, UnitPoint, &font);
 
 	if (font) {
 		FT_Face	face;
+		TT_VertHeader *pVert;
 
-		face = cairo_ft_font_face(font->cairofnt);
+		face = gdip_cairo_ft_font_lock_face(font->cairofnt);
 
-		TT_VertHeader *pVert = FT_Get_Sfnt_Table (face, ft_sfnt_vhea);
-		if (pVert)
+		pVert = FT_Get_Sfnt_Table(face, ft_sfnt_vhea);
+		if (pVert) {
 			rslt = pVert->yMax_Extent;
-		else if (face)
+		} else if (face) {
 			rslt = face->units_per_EM;
-		else
+		} else {
 			rslt = 0;
+		}
+		gdip_cairo_ft_font_unlock_face(font->cairofnt);
 		GdipDeleteFont (font);
 	}
 
 	*EmHeight = rslt;
+	((GpFontFamily *)family)->height = rslt;
 	return Ok;
 }
 
@@ -377,22 +415,30 @@ GdipGetCellAscent (GDIPCONST GpFontFamily *family, GpFontStyle style, short *Cel
 	if (!family || !CellAscent)
 		return InvalidParameter;
 
+	if (family->cellascent != -1) {
+		*CellAscent = family->cellascent;
+		return Ok;
+	}
+
 	GdipCreateFont (family, 0.0f, style, UnitPoint, &font);
 
 	if (font){
                 FT_Face face;
+		TT_HoriHeader *pHori;
 
-                face = cairo_ft_font_face(font->cairofnt);
+                face = gdip_cairo_ft_font_lock_face(font->cairofnt);
 
-		TT_HoriHeader *pHori = FT_Get_Sfnt_Table (face, ft_sfnt_hhea);
+		pHori = FT_Get_Sfnt_Table (face, ft_sfnt_hhea);
 
 		if (pHori)
 			rslt = pHori->Ascender;
 
+		gdip_cairo_ft_font_lock_face(font->cairofnt);
 		GdipDeleteFont (font);
 	}
 
 	*CellAscent = rslt;
+	((GpFontFamily *)family)->cellascent = rslt;
 	return Ok;         
 }
 
@@ -405,24 +451,30 @@ GdipGetCellDescent (GDIPCONST GpFontFamily *family, GpFontStyle style, short *Ce
 	if (!family || !CellDescent)
 		return InvalidParameter;
 
-	*CellDescent = 0;
+	if (family->celldescent != -1) {
+		*CellDescent = family->celldescent;
+		return Ok;
+	}
 
 	GdipCreateFont (family, 0.0f, style, UnitPoint, &font);
 
 	if (font){
                 FT_Face face;
+		TT_HoriHeader* pHori;
 
-                face = cairo_ft_font_face(font->cairofnt);
+                face = gdip_cairo_ft_font_lock_face(font->cairofnt);
 
-		TT_HoriHeader *pHori = FT_Get_Sfnt_Table (face, ft_sfnt_hhea);
+		pHori = FT_Get_Sfnt_Table (face, ft_sfnt_hhea);
 
 		if (pHori)
 			rslt = -pHori->Descender;
 
+		gdip_cairo_ft_font_unlock_face(font->cairofnt);
 		GdipDeleteFont (font);
 	}
 
 	*CellDescent = rslt;
+	((GpFontFamily *)family)->celldescent = rslt;
 	return Ok;         
 }
 
@@ -435,24 +487,34 @@ GdipGetLineSpacing (GDIPCONST GpFontFamily *family, GpFontStyle style, short *Li
 	if (!family || !LineSpacing)
 		return InvalidParameter;
 
+	if (family->linespacing != -1) {
+		*LineSpacing = family->linespacing;
+		return Ok;
+	}
+
 	GdipCreateFont (family, 0.0f, style, UnitPoint, &font);
 
 	if (font){
                 FT_Face face;
+		TT_HoriHeader *pHori;
 
-                face = cairo_ft_font_face(font->cairofnt);
+                face = gdip_cairo_ft_font_lock_face(font->cairofnt);
 
-		TT_HoriHeader *pHori = FT_Get_Sfnt_Table (face, ft_sfnt_hhea);
-		if (pHori)
+		pHori = FT_Get_Sfnt_Table (face, ft_sfnt_hhea);
+		if (pHori) {
 			rslt = pHori->Ascender + (-pHori->Descender) + pHori->Line_Gap;
-		else if (face)
+		} else if (face) {
 			rslt = face->height;
-		else
+		} else {
 			rslt = 0;
+		}
+
+		gdip_cairo_ft_font_unlock_face(font->cairofnt);
 		GdipDeleteFont (font);
 	}
 
 	*LineSpacing = rslt;
+	((GpFontFamily *)family)->linespacing = rslt;
 	return Ok;
 }
 
@@ -467,7 +529,6 @@ GdipIsStyleAvailable (GDIPCONST GpFontFamily *family, int style, BOOL *IsStyleAv
 }
 
 
-
 /* Font functions */
 
 int
@@ -503,10 +564,11 @@ gdip_font_create (const unsigned char *family, int fcslant, int fcweight, GpFont
 	result->cairofnt = font;
 	result->ft_library = ft_library;
 
-	FT_Set_Char_Size (cairo_ft_font_face(font),
+	FT_Set_Char_Size (gdip_cairo_ft_font_lock_face(font),
 			  DOUBLE_TO_26_6 (1.0),
 			  DOUBLE_TO_26_6 (1.0),
 			  0, 0);
+	gdip_cairo_ft_font_unlock_face(font);
 
 	FcPatternDestroy (pat);
 	return 1;
@@ -538,6 +600,35 @@ gdip_font_drawstrikeout (GpGraphics *graphics, GpBrush *brush, float x, float y,
         GdipFillRectangle (graphics, brush, x, y -pos, width, size);
 }
 
+void
+gdip_release_font (GpFont* font)
+{
+	if (!font)
+		return;
+
+	cairo_font_destroy ((cairo_font_t *)font->cairofnt);
+
+	if (font->ft_library)
+		FT_Done_FreeType(font->ft_library);
+	
+	GdipFree ((void *)font);	
+}
+
+void
+gdip_release_cachedfonts ()
+{
+	int i;
+
+	/* Loop all the cached fonts */
+	
+	LOCK_FONTCACHE;
+	for (i = 0; i < cached_fonts_index; i++) {
+		gdip_release_font (cached_fonts[i].font);
+	}
+	UNLOCK_FONTCACHE;
+}
+
+
 GpStatus
 GdipCreateFont (GDIPCONST GpFontFamily* family, float emSize, GpFontStyle style, Unit unit,  GpFont **font)
 {
@@ -546,15 +637,42 @@ GdipCreateFont (GDIPCONST GpFontFamily* family, float emSize, GpFontStyle style,
 	GpFont *result;
 	int slant = 0;
 	int weight = FC_WEIGHT_LIGHT;
+	int i;
+	float sizeInPixels;
 	
 	if (!family || !font)
 		return InvalidParameter;
 
 	r = FcPatternGetString (family->pattern, FC_FAMILY, 0, &str);
+	
+	gdip_unitConversion (unit, UnitPixel, emSize, &sizeInPixels);
+
+	
+	/* Is it already in the cache */
+	LOCK_FONTCACHE;
+	for (i = 0; i < cached_fonts_index; i++) {
+		if (sizeInPixels != cached_fonts[i].sizeInPixels)
+			continue;
+
+		if (style != cached_fonts[i].style)
+			continue;
+		
+		if (strcmp ((char *)str, (const char *)cached_fonts[i].szFamily) != 0)
+			continue;
+
+		if (cached_fonts[i].font == NULL)
+			continue;
+
+		/* Found in cache */
+		*font = cached_fonts[i].font;
+		cached_fonts[i].refcount++;
+		UNLOCK_FONTCACHE;
+		return Ok;
+	}
+	UNLOCK_FONTCACHE;
 
 	result = (GpFont *) GdipAlloc (sizeof (GpFont));
-
-	gdip_unitConversion (unit, UnitPixel, emSize, &result->sizeInPixels);
+	result->sizeInPixels = sizeInPixels;
 
         if ((style & FontStyleBold) == FontStyleBold)
                 weight = FC_WEIGHT_BOLD;
@@ -563,12 +681,40 @@ GdipCreateFont (GDIPCONST GpFontFamily* family, float emSize, GpFontStyle style,
                 slant = FC_SLANT_ITALIC;
         
 	if (!gdip_font_create (str, slant, weight, result)) {
-		return InvalidParameter;	// FIXME -  wrong return code
+		return InvalidParameter;	/* FIXME -  wrong return code */
 	}
         result->style = style;
 	cairo_font_reference ((cairo_font_t *)result->cairofnt);
-	result->wineHfont=CreateWineFont(str, style, emSize, unit);
 	*font=result;
+
+	if (strlen ((const char *)str) > 127) /* Cannot cache this font */
+		return Ok;
+	
+	/* Cache entry */
+	LOCK_FONTCACHE;
+	if (cached_fonts_index < MAX_CACHED_FONTS) {
+		strcpy (cached_fonts[cached_fonts_index].szFamily, (const char *)str);
+		cached_fonts[cached_fonts_index].sizeInPixels = sizeInPixels;
+		cached_fonts[cached_fonts_index].style = style;
+		cached_fonts[cached_fonts_index].font = result;
+		cached_fonts[cached_fonts_index].refcount = 1;
+		cached_fonts_index++;
+	} else {
+
+		/* Look for the first non-referenced font */
+		for (i = 0; i < MAX_CACHED_FONTS; i++) {
+			if (cached_fonts[i].refcount > 0)
+				continue;
+
+			gdip_release_font (cached_fonts[i].font); /* release previous cached font */
+			strcpy (cached_fonts[i].szFamily, (const char *)str);
+			cached_fonts[i].sizeInPixels = sizeInPixels;
+			cached_fonts[i].style = style;
+			cached_fonts[i].font = result;
+			cached_fonts[i].refcount = 1;
+		}	
+	}
+	UNLOCK_FONTCACHE;
         
 	return Ok;
 }
@@ -583,190 +729,100 @@ GdipGetHfont(GpFont* font, void **Hfont)
 	return InvalidParameter;
 }
 
+
 GpStatus
 GdipDeleteFont (GpFont* font)
 {
-	if (font) {
-		cairo_font_destroy ((cairo_font_t *)font->cairofnt);
-		if (font->ft_library) {
-			FT_Done_FreeType(font->ft_library);
-		}
-		if (font->wineHfont) {
-			DeleteWineFont(font->wineHfont);
-		}
-		GdipFree ((void *)font);
-                return Ok;
-	}
-        return InvalidParameter;
-}
+	int i;
 
-static GpStatus
-CreateFontFromHDCorHfont(void *hdcIn, void *hfont, GpFont **font, LOGFONTA *lf)
-{
-	FcResult	r;
-	GpFont		*result;
-	int		slant	= 0;
-	int		weight	= FC_WEIGHT_LIGHT;
-	int		style	= 0;
-	TEXTMETRICA	tm;
-	unsigned char	FaceName[33];
-	void		*hdc;
-	void		*oldFont;
-	
-	if (!font) {
+	if (!font)
 		return InvalidParameter;
-	}
 
-	if (!hdcIn) {
-		hdc=GetDC_pfn(NULL);
-		oldFont=SelectObject_pfn(hdc, hfont);
-	} else {
-		hdc=hdcIn;
-		oldFont=0;	/* Make the compiler happy, we don't care about oldFont in this path */
-	}
-	SetMapMode_pfn(hdc, 1);
-	if (GetTextMetrics_pfn(hdc, &tm)==0 || GetTextFace_pfn(hdc, sizeof(FaceName), FaceName)==0) {
-		ReleaseDC_pfn(NULL, hdc);
-		return InvalidParameter;
-	}
-	if (!hdcIn) {
-		SelectObject_pfn(hdc, oldFont);
-		ReleaseDC_pfn(NULL, hdc);
-	}
+	/* Is is in the cache */
+	LOCK_FONTCACHE;
+	for (i = 0; i < cached_fonts_index; i++) {
+		if (font != cached_fonts[i].font)
+			continue;		
 
-	result = (GpFont *) GdipAlloc (sizeof (GpFont));
-
-	if (tm.tmHeight<0) {
-	   result->sizeInPixels=tm.tmHeight*-1;
-	} else {
-	   result->sizeInPixels=tm.tmHeight;
+		/* Found in cache */
+		cached_fonts[i].refcount--;
+		UNLOCK_FONTCACHE;
+		return Ok;
 	}
+	UNLOCK_FONTCACHE;
 
-	if (tm.tmStruckOut!=0) {
-		style |= FontStyleStrikeout;
-	}
-
-	if (tm.tmUnderlined!=0) {
-		style |= FontStyleUnderline;
-	}
-
-	if (tm.tmItalic!=0) {
-		style |= FontStyleItalic;
-		slant = FC_SLANT_ITALIC;
-	}
-
-	if (tm.tmWeight>400) {
-		style |= FontStyleBold;
-		weight = FC_WEIGHT_BOLD;
-	}
-
-	if (!gdip_font_create (FaceName, slant, weight, result)) {
-		return InvalidParameter;	// FIXME - pick right return code
-	}
-	
-	result->style = style;
-	cairo_font_reference ((cairo_font_t *)result->cairofnt);
-	result->wineHfont=CreateWineFont(FaceName, style, result->sizeInPixels, UnitPixel);
-
-	/* Assign our results */
-	*font=result;
-	if (lf) {
-		lf->lfHeight=result->sizeInPixels;
-		lf->lfItalic=tm.tmItalic;
-		lf->lfStrikeOut=tm.tmStruckOut;
-		lf->lfUnderline=tm.tmUnderlined;
-		lf->lfWeight=tm.tmWeight;
-		strcpy(lf->lfFaceName, FaceName);
-	}
-
-	return Ok;
+	gdip_release_font (font);
+	return Ok;	       
 }
 
 GpStatus
 GdipCreateFontFromDC(void *hdc, GpFont **font)
 {
-	return(CreateFontFromHDCorHfont(hdc, NULL, font, NULL));
+	return(NotImplemented);
 }
 
 GpStatus
-GdipCreateFontFromHfont(void *hfont, GpFont **font, LOGFONTA *lf)
+GdipCreateFontFromHfont(void *hfont, GpFont **font, void *lf)
 {
-	return(CreateFontFromHDCorHfont(NULL, hfont, font, lf));
+	return(NotImplemented);
 }
 
 GpStatus
-GdipGetLogFontA(GpFont *font, GpGraphics *graphics, LOGFONTA *lf)
+GdipGetLogFontA(GpFont *font, GpGraphics *graphics, void *lf)
 {
-	void		*hdc;
-	void		*oldFont;
-	TEXTMETRICA	tm;
-	unsigned char	FaceName[33];
-
 	if (!font || !lf) {
 		return(InvalidParameter);
 	}
-
-	hdc=GetDC_pfn(NULL);
-	oldFont=SelectObject_pfn(hdc, font->wineHfont);
-
-	if (GetTextMetrics_pfn(hdc, &tm)==0 || GetTextFace_pfn(hdc, sizeof(FaceName), FaceName)==0) {
-		SelectObject_pfn(hdc, oldFont);
-		ReleaseDC_pfn(NULL, hdc);
-		return(Win32Error);
-	}
-
-	/* We assign what we know and default the rest */
-	lf->lfHeight=font->sizeInPixels;
-	lf->lfWidth=tm.tmAveCharWidth;
-	lf->lfEscapement=0;
-	lf->lfOrientation=0;
-	lf->lfWeight=tm.tmWeight;
-	lf->lfItalic=tm.tmItalic;
-	lf->lfUnderline=tm.tmUnderlined;
-	lf->lfStrikeOut=tm.tmStruckOut;
-	lf->lfCharSet=tm.tmCharSet;
-	lf->lfOutPrecision=0;		/* 0 = OUT_DEFAULT_PRECIS */
-	lf->lfClipPrecision=0;		/* 0 = CLIP_DEFAULT_PRECIS */
-	lf->lfQuality=4;		/* 4 = ANTIALIASED_QUALITY */
-	lf->lfPitchAndFamily=tm.tmPitchAndFamily;
-	strcpy(lf->lfFaceName, FaceName);
-
-	/* Clean up */
-	SelectObject_pfn(hdc, oldFont);
-	ReleaseDC_pfn(NULL, hdc);
-
-	return(Ok);
+	
+	return NotImplemented;
 }
 
 GpStatus
 GdipPrivateAddMemoryFont(GpFontCollection *fontCollection, GDIPCONST void *memory, int length)
 {
-	char	*fontfile;
-	FILE	*f;
+	FcChar8 fontfile[256];
+	int	f;
 
-	fontfile=tempnam(NULL, NULL);
-	if (!fontfile) {
-		return(OutOfMemory);
-	}
+	strcpy((char *) fontfile, "/tmp/ffXXXXXX");
 
-	f=fopen(fontfile, "wb");
-	if (!f) {
-		free(fontfile);
+	f = mkstemp((char*)fontfile);
+	if (f == -1) {
 		return(GenericError);
 	}
 
-	if (fwrite(memory, 1, length, f)!=length) {
-		fclose(f);
-		free(fontfile);
+	if (write(f, memory, length)!=length) {
+		close(f);
 		return(GenericError);
 	}
-	fclose(f);
+	close(f);
 
 	FcConfigAppFontAddFile(fontCollection->config, fontfile);
 
 	/* FIXME - May we delete our temporary font file or does 
 	   FcConfigAppFontAddFile just reference our file?  */
+	/* unlink(fontfile); */
 
-	free(fontfile);
 	return(Ok);
 }
+
+
+bool
+GetFontMetrics(GpGraphics *graphics, GpFont *font, int *ascent, int *descent)
+{
+	cairo_font_extents_t	font_extent;
+
+	cairo_set_font (graphics->ct, (cairo_font_t*) font->cairofnt);
+	cairo_scale_font (graphics->ct, font->sizeInPixels);
+	cairo_current_font_extents (graphics->ct, &font_extent);
+
+	if (ascent) {
+		*ascent = (int)font_extent.ascent;
+	}
+
+	/* Descent is negative for descent below the baseline */
+	if (descent) {
+		*descent = -1 * (int)font_extent.descent;
+	}
+
+	return TRUE;
+}
