@@ -8,8 +8,9 @@
  *      Ravindra (rkumar@novell.com)
  *  	Sanjay Gupta (gsanjay@novell.com)
  *	Vladimir Vukicevic (vladimir@pobox.com)
+ *	Geoff Norton (gnorton@customerdna.com)
  *
- * Copyright (C) Novell, Inc. 2003-2004. http://www.novell.com
+ * Copyright (C) Novell, Inc. 2003-2005. http://www.novell.com
  */
 
 #ifndef _GDIP_H
@@ -18,10 +19,64 @@
 #include <stdlib.h>
 #include <stdio.h>
 #include <glib.h>
+#include <unistd.h>
 
+#include "config.h"
+
+#include "pixman.h"
+
+#ifdef USE_INCLUDED_CAIRO
 #include <cairo.h>
-#ifndef CAIRO_HAS_XLIB_SURFACE
+#else
+#include <cairo/cairo.h>
+#endif
+#include <X11/Xlib.h>
+
+#ifdef CAIRO_HAS_FT_FONT
+#ifdef USE_INCLUDED_CAIRO
+#include <cairo-ft.h>
+#else
+#include <cairo/cairo-ft.h>
+#endif
+#endif
+
+#define	gdip_cairo_ft_font_lock_face(font)	cairo_ft_font_face(font)
+#define gdip_cairo_ft_font_unlock_face(font)
+
+#ifdef WORDS_BIGENDIAN
+#define set_pixel_bgra(pixel,index,b,g,r,a) { \
+		pixel[index+0] = a; \
+		pixel[index+1] = r; \
+		pixel[index+2] = g; \
+		pixel[index+3] = b; \
+	}
+#define get_pixel_bgra(color, b, g, r, a) { \
+		a = (color & 0x000000ff); \
+		r = (color & 0x0000ff00) >> 8; \
+		g = (color & 0x00ff0000) >> 16; \
+		b = (color & 0xff000000) >> 24; \
+	}
+#else
+#define set_pixel_bgra(pixel,index,b,g,r,a) { \
+		pixel[index+0] = b; \
+		pixel[index+1] = g; \
+		pixel[index+2] = r; \
+		pixel[index+3] = a; \
+	}
+#define get_pixel_bgra(color, b, g, r, a) { \
+		a = ((color & 0xff000000) >> 24); \
+		r = ((color & 0x00ff0000) >> 16); \
+		g = ((color & 0x0000ff00) >> 8); \
+		b = (color & 0x000000ff); \
+	}
+#endif
+
+#ifdef CAIRO_HAS_XLIB_SURFACE
+#ifdef USE_INCLUDED_CAIRO
 #include <cairo-xlib.h>
+#else
+#include <cairo/cairo-xlib.h>
+#endif
 #endif
 
 /* mono/io-layer/uglify.h also has these typedefs.
@@ -90,10 +145,10 @@ typedef ImageAbort GetThumbnailImageAbort;
 
 typedef unsigned char byte;
 typedef int bool;
-typedef unsigned short WCHAR; /* 16-bits unicode */
-typedef unsigned int UINT;
-typedef unsigned int ARGB;
-typedef int PROPID;
+typedef guint16 WCHAR; /* 16-bits unicode */
+typedef guint32 UINT;
+typedef guint32 ARGB;
+typedef gint32 PROPID;
 
 #define MAKE_ARGB_RGB(r,g,b)  (((r&0xff) << 16) | ((g&0xff) << 8) | (b&0xff))
 #define MAKE_ARGB_ARGB(a,r,g,b)  (((a&0xff) << 24) | ((r&0xff) << 16) | ((g&0xff) << 8) | (b&0xff))
@@ -160,15 +215,12 @@ typedef enum {
         DigitSubstituteTraditional = 3
 } DigitSubstitute;
 
-
-
 /* The pixel format spec is:
  * [0-7 format index] [8-15 pixel size, bits] [16-23 flags] [24-31 reserved]
  * so,
  * (ID | (bpp << 8) | flags)
  */
 
-
 typedef enum {
 	PixelFormatAlpha = 0x00040000,     /* flag: format has alpha */
 	PixelFormatCanonical = 0x00200000, /* flag: unknown */
@@ -291,7 +343,8 @@ typedef enum {
 typedef enum {
 	gtUndefined,
 	gtX11Drawable,
-	gtMemoryBitmap
+	gtMemoryBitmap,
+	gtOSXDrawable
 } GraphicsType;
 
 typedef enum{
@@ -378,7 +431,6 @@ typedef struct {
 	int		LineLen;	/* If LineStart how many chars is the line long? */
 } GpStringDetailStruct;
 
-
 typedef enum {
 	ImageFlagsNone = 0,
 	ImageFlagsScalable = 1,
@@ -415,7 +467,6 @@ typedef enum
     CombineModeXor              = 3,       
     CombineModeExclude          = 4,   
     CombineModeComplement       = 5  
-    
 } CombineMode;
 
 typedef enum
@@ -497,7 +548,7 @@ typedef enum
 typedef enum {
         CoordinateSpaceWorld,
         CoordinateSpacePage,
-        CoordinateSpaceDevice,
+        CoordinateSpaceDevice
 } GpCoordinateSpace;
 
 typedef enum {
@@ -566,6 +617,13 @@ typedef enum {
         RegionTypePath  = 3
 } RegionType;
 
+typedef enum {
+	X11DRV_GET_DISPLAY,			/* get X11 display for a DC */
+	X11DRV_GET_DRAWABLE,		/* get current drawable for a DC */
+	X11DRV_GET_FONT			/* get current X font for a DC */
+} x11drv_escape_codes;
+
+
 /*
  * Structures
  *
@@ -580,8 +638,14 @@ typedef struct {    /* Keep in sync with BitmapData.cs */
 	unsigned int Height;
 	int          Stride;
 	int          PixelFormat;
-	void         *Scan0;
+	byte* 	     Scan0;
 	unsigned int Reserved;
+	/*Added to keep track of position for displaying next frame
+	in a gif*/
+	unsigned int Top; 
+	unsigned int Left;
+	int	     ByteCount;
+	char*	     Bytes;
 } GdipBitmapData, BitmapData;
 
 typedef struct {
@@ -606,17 +670,8 @@ typedef struct {
 } GpRegion;
 
 typedef struct {
-	cairo_t         *ct;
-	cairo_matrix_t  *copy_of_ctm;
-	void            *hdc;
-	int             hdc_busy_count;
-	void            *image;
-	int             type; 
-	GpRegion*	clip;
-	GpRect		bounds;
-	GpUnit		page_unit;
-	float		scale;
-} GpGraphics;
+	int first, length;
+} CharacterRange;
 
 typedef struct _CustomLineCap GpCustomLineCap;
 typedef struct _AdjustableArrowCap GpAdjustableArrowCap;
@@ -631,6 +686,7 @@ typedef struct {
         GpLineJoin line_join;
         GpDashStyle dash_style;
 	GpLineCap line_cap;  /* Cairo supports only same kind of end caps for both the ends. */
+	GpLineCap end_cap;
 	int compound_count;
 	float *compound_array;
         GpPenAlignment mode;
@@ -643,6 +699,25 @@ typedef struct {
         BOOL changed; /* flag to mark if pen is changed and needs setup */
 } GpPen;
 
+typedef struct {
+	cairo_t         *ct;
+	cairo_matrix_t	*copy_of_ctm;
+	Display		*display;
+	Drawable	drawable;
+	void            *image;
+	int             type; 
+	GpRegion*	clip;
+	GpRect		bounds;
+	GpUnit		page_unit;
+	float		scale;
+	InterpolationMode interpolation;
+	GpPen*		last_pen;	/* caching pen and brush to avoid unnecessary sets */
+	GpBrush*	last_brush;
+	
+	GpCompositingMode    composite_mode;
+	GpCompositingQuality composite_quality;
+} GpGraphics;
+
 typedef struct {
 	cairo_matrix_t		*matrix;
 } GpState;
@@ -654,10 +729,10 @@ typedef struct {
 } ColorPalette;
 
 typedef struct {
-    unsigned long  Data1;
-    unsigned short Data2;
-    unsigned short Data3;
-    byte           Data4 [8];
+    DWORD Data1;
+    WORD  Data2;
+    WORD  Data3;
+    byte  Data4 [8];
 } GUID, Guid, CLSID;
 
 typedef struct {
@@ -682,12 +757,12 @@ typedef struct {
 typedef struct {
 	GUID frameDimension;
 	int count;
-} FrameDimensionInfo;
+	BitmapData *frames;
+} FrameInfo;
 
 typedef struct {
 	ImageType     	type;
 	cairo_surface_t *surface;
-	GpGraphics  	*graphics;		/* created by GdipGetImageGraphicsContext */
 	int 		imageFlags;
 	int 		height;
 	int 		width;
@@ -698,23 +773,24 @@ typedef struct {
 	float 		verticalResolution;
 	ImageFormat     format;
 	int		frameDimensionCount;
-	FrameDimensionInfo *frameDimensionList;
+	FrameInfo	*frameDimensionList;
 } GpImage;
 
 typedef struct {
-	GpImage	image;
+	GpImage		image;
         int             cairo_format;
 	BitmapData	data;
-	void *hBitmapDC;
-	void *hInitialBitmap;
-	void *hBitmap;
+	void		*hBitmapDC;
+	void		*hInitialBitmap;
+	void		*hBitmap;	
 } GpBitmap;
 
 typedef struct {
-        GpFillMode fill_mode;
-        int count;
-        GByteArray *types;
-        GArray *points;
+	GpFillMode fill_mode;
+	int count;
+	GByteArray *types;
+	GArray *points;
+	BOOL start_new_fig; /* Flag to keep track if we need to start a new figure */
 } GpPath;
 
 typedef struct {
@@ -723,6 +799,13 @@ typedef struct {
         byte *Types;
 } GpPathData;
 
+typedef struct {
+	GpPath *path;
+	int markerPosition; /* The start position of next marker, index of (marker type) + 1  */
+	int subpathPosition; /* The start position of next subpath, index of (start type) */
+	int pathTypePosition; /* The position to get the next path type inside a subpath */
+} GpPathIterator;
+
 typedef struct {
         FcFontSet*  fontset;
         /* Only for private collections */
@@ -733,6 +816,10 @@ typedef struct {
 typedef struct {
         FcPattern*	pattern;
 	bool		allocated;
+	short 		height;
+	short 		linespacing;
+	short		celldescent;
+	short		cellascent;
 } GpFontFamily;
 
 
@@ -744,7 +831,15 @@ typedef struct {
 	FT_Library          ft_library;
 } GpFont;
 
-
+ 
+typedef struct
+{
+	char		szFamily [128];
+	float		sizeInPixels;
+	GpFontStyle	style;
+	GpFont *	font;
+	int		refcount;
+} GpCachedFont;
 
 typedef struct {
         StringAlignment alignment;
@@ -753,10 +848,12 @@ typedef struct {
         StringFormatFlags formatFlags;
         StringTrimming  trimming;
         DigitSubstitute substitute;
+	CharacterRange *charRanges;
         float firstTabOffset;
         float *tabStops;
         int numtabStops;
-}GpStringFormat;
+	int charRangeCount;
+} GpStringFormat;
 
 typedef struct {
         float X, Y;
@@ -771,9 +868,10 @@ typedef struct {
 	ARGB key_colorhigh;
 	BOOL key_enabled;
 	BOOL no_op;
+	GpColorMatrix *colormatrix;
+	BOOL colormatrix_enabled;
 } GpImageAttribute;
 
-
 typedef struct {
 	GpImageAttribute def;
 	GpImageAttribute bitmap;
@@ -808,17 +906,12 @@ typedef struct {
  * 
  */
 void gdip_image_init              (GpImage *image);
-void *gdip_image_create_Win32_HDC (GpImage *image);
-void gdip_image_destroy_Win32_HDC (GpImage *image, void *hdc);
 
 void gdip_bitmap_init  (GpBitmap *bitmap);
 GpBitmap *gdip_bitmap_new   (void);
 void gdip_bitmap_dispose (GpBitmap *bitmap);
 void gdip_bitmap_clone (GpBitmap *bitmap, GpBitmap **clonedbitmap);
 
-void *gdip_bitmap_create_Win32_HDC (GpBitmap *bitmap);
-void gdip_bitmap_destroy_Win32_HDC (GpBitmap *bitmap, void *hdc);
-
 void gdip_graphics_init (GpGraphics *graphics);
 GpGraphics *gdip_graphics_new (void);
 void gdip_graphics_attach_bitmap (GpGraphics *graphics, GpBitmap *image);
@@ -827,18 +920,14 @@ void gdip_graphics_detach_bitmap (GpGraphics *graphics, GpBitmap *image);
 GpStatus gdip_bitmap_clone_data_rect (GdipBitmapData *srcData, Rect *srcRect, GdipBitmapData *destData, Rect *destRect);
 GpStatus gdip_bitmap_change_rect_pixel_format (GdipBitmapData *srcData, Rect *srcRect, GdipBitmapData *destData, Rect *destRect);
 
-void gdip_from_RGB_to_ARGB (BYTE *src, int width, int height, int stride, BYTE **dest, int* dest_stride);
-void gdip_from_ARGB_to_RGB (BYTE *src, int width, int height, int stride, BYTE **dest, int* dest_stride);
+int gdip_from_RGB_to_ARGB (BYTE *src, int width, int height, int stride, BYTE **dest, int* dest_stride);
+int gdip_from_ARGB_to_RGB (BYTE *src, int width, int height, int stride, BYTE **dest, int* dest_stride);
 
 /* Pen */
 void gdip_pen_init (GpPen *pen);
 GpPen *gdip_pen_new (void);
 GpStatus gdip_pen_setup (GpGraphics *graphics, GpPen *pen);
 
-extern Display *GDIP_display;
-
-void initializeGdipWin32 (void);
-void shutdownGdipWin32 (void);
 void initCodecList (void);
 void releaseCodecList (void);
 void gdip_process_bitmap_attributes (GpBitmap *bitmap, void **dest, GpImageAttributes* attr, bool *allocated);
@@ -854,10 +943,10 @@ GpStatus GdipBitmapGetPixel ( GpBitmap *bmp, int x, int y, ARGB *color);
 
 
 /* Graphics */
-GpStatus GdipCreateFromHDC (int hDC, GpGraphics **graphics);
+GpStatus GdipCreateFromHDC (void *hDC, GpGraphics **graphics);
 GpStatus GdipDeleteGraphics (GpGraphics *graphics);
-GpStatus GdipGetDC (GpGraphics *graphics, int *hDC);
-GpStatus GdipReleaseDC (GpGraphics *graphics, int hDC);
+GpStatus GdipGetDC (GpGraphics *graphics, void **hDC);
+GpStatus GdipReleaseDC (GpGraphics *graphics, void *hDC);
 GpStatus GdipRestoreGraphics (GpGraphics *graphics, unsigned int graphicsState);
 GpStatus GdipSaveGraphics(GpGraphics *graphics, unsigned int * state);
 GpStatus GdipRotateWorldTransform (GpGraphics *graphics, float angle, GpMatrixOrder order);
@@ -934,6 +1023,7 @@ GpStatus GdipSetTextContrast(GpGraphics *graphics, UINT contrast);
 GpStatus GdipGetTextContrast(GpGraphics *graphics, UINT *contrast);
 GpStatus GdipSetSmoothingMode(GpGraphics *graphics, SmoothingMode mode);
 GpStatus GdipGetSmoothingMode(GpGraphics *graphics, SmoothingMode *mode);
+GpStatus GdipSetVisibleClip_linux (GpGraphics *graphics, GpRect *rect);
 
 /* Status */
 GpStatus gdip_get_status (cairo_status_t status);
@@ -1014,7 +1104,7 @@ GpStatus GdipGetAdjustableArrowCapFillState (GpAdjustableArrowCap *arrowCap, boo
 /* Text */
 GpStatus GdipDrawString (GpGraphics *graphics, GDIPCONST WCHAR *string, int len, GDIPCONST GpFont *font, GDIPCONST RectF *rc, GDIPCONST GpStringFormat *format, GpBrush *brush);
 GpStatus GdipMeasureString(GpGraphics *graphics, GDIPCONST WCHAR *string, int length, GDIPCONST GpFont *font, GDIPCONST RectF *layoutRect, GDIPCONST GpStringFormat *stringFormat,  RectF *boundingBox, int *codepointsFitted, int *linesFilled);
-
+GpStatus GdipMeasureCharacterRanges(GpGraphics *graphics, GDIPCONST WCHAR *string, int length, GDIPCONST GpFont *font, GDIPCONST GpRectF *layoutRect, GDIPCONST GpStringFormat *stringFormat, int regionCount, GpRegion **regions);
 
 /* Matrix */
 GpStatus GdipCreateMatrix (GpMatrix **matrix);
@@ -1088,6 +1178,8 @@ GpStatus GdipGetStringFormatDigitSubstitution(GDIPCONST GpStringFormat *format,
 GpStatus GdipSetStringFormatDigitSubstitution(GpStringFormat *format, int language, DigitSubstitute substitute);      
 GpStatus GdipGetStringFormatTabStopCount(GDIPCONST GpStringFormat *format, int *count);
 GpStatus GdipGetStringFormatTabStops(GDIPCONST GpStringFormat *format, int count, float *firstTabOffset, float *tabStops);
+GpStatus GdipGetStringFormatMeasurableCharacterRangeCount(GDIPCONST GpStringFormat *format, int *count);
+GpStatus GdipSetStringFormatMeasurableCharacterRanges(GpStringFormat *format, int rangeCount, GDIPCONST CharacterRange *ranges);
 
 /* ImageAttributes */
 GpStatus GdipCreateImageAttributes(GpImageAttributes **imageattr);
@@ -1162,8 +1254,6 @@ void GdipFree (void *ptr);
 int fcmp (double x1, double x2, double epsilon);
 float gdip_get_display_dpi();
 void gdip_unitConversion(Unit fromUnit, Unit toUnit, float nSrc, float* nTrg);
-int gdpi_utf8_to_glyphs (cairo_font_t* font, cairo_matrix_t matrix, const unsigned char* utf8, double	x0,
-   double y0, cairo_glyph_t** glyphs, size_t* nglyphs);
 
 void gdip_font_drawunderline (GpGraphics *graphics, GpBrush *brush, float x, float y, float width);
 void gdip_font_drawstrikeout (GpGraphics *graphics, GpBrush *brush, float x, float y, float width);
@@ -1177,6 +1267,7 @@ cairo_surface_t * gdip_bitmap_ensure_surface (GpBitmap *bitmap);
 const EncoderParameter *gdip_find_encoder_parameter (GDIPCONST EncoderParameters *eps, const GUID *guid);
 
 /* Stream handling bits */
+typedef int (*GetHeaderDelegate) (unsigned char *, int);
 typedef int (*GetBytesDelegate) (unsigned char *, int, BOOL);
 typedef long (*SeekDelegate) (int, int);
 typedef int (*PutBytesDelegate) (unsigned char *, int);
