@@ -45,80 +45,23 @@ struct startupOutput
 	void *unhook;
 };
 
-Display *GDIP_display = 0;
-int      closeDisplay = 0;
-
-void * x11drvHandle = 0;
-
-static void _load_x11drv ()
-{
-	if (x11drvHandle == 0) {
-		char *path = getenv ("_WINE_SHAREDLIB_PATH");
-		char *full;
-
-		/* We are called also without Wine */
-		if (path == NULL)
-			return;
-
-		full = malloc (strlen (path) + 30);
-		strcpy (full, path);
-		strcat (full, "/");
-		strcat (full, "x11drv.dll.so");
-		
-		x11drvHandle = dlopen (full, 1);
-		free (full);
-	}
-}
-
-static void _unload_x11drv ()
-{
-	if (x11drvHandle != 0) {
-		dlclose (x11drvHandle);
-	}
-}
-
-Display *_get_wine_display ()
-{
-	Display * result = 0;
-	_load_x11drv ();
-	if (x11drvHandle != 0) {
-		Display **addr = dlsym(x11drvHandle,"gdi_display");
-		if (addr) {
-			result = *addr;
-		}
-	}
-	return result;
-}
+extern void gdip_release_cachedfonts ();
 
 
 GpStatus 
 GdiplusStartup(unsigned long *token, const struct startupInput *input, struct startupOutput *output)
 {
-    if (getenv ("GDIPLUS_NOX") == NULL) {
-        GDIP_display = _get_wine_display ();
-        if (GDIP_display == 0){
-            GDIP_display = XOpenDisplay(0);
-            closeDisplay = 1;
-        }
-    } else {
-        GDIP_display = 0;
-    }
-
         initCodecList (); 
-        initializeGdipWin32 ();
 	*token = 1;
+	gdip_get_display_dpi();
 	return Ok;
 }
 
 void 
 GdiplusShutdown(unsigned long *token)
 {
-	if (closeDisplay && GDIP_display) {
-		XCloseDisplay(GDIP_display);
-	}
-	_unload_x11drv ();
-        releaseCodecList ();
-	shutdownGdipWin32();
+	releaseCodecList ();
+	gdip_release_cachedfonts ();
 }
 
 
@@ -163,25 +106,70 @@ gdip_get_status (cairo_status_t status)
         }
 }
 
+/* Converts the given interpolation value to cairo_filter_t */
+cairo_filter_t
+gdip_get_cairo_filter (InterpolationMode imode)
+{
+	cairo_filter_t filter;
+
+	switch (imode) {
+
+	case InterpolationModeHighQuality:
+	case InterpolationModeHighQualityBilinear:
+	case InterpolationModeHighQualityBicubic:
+		filter = CAIRO_FILTER_BEST;
+		break;
+
+	case InterpolationModeNearestNeighbor:
+		filter = CAIRO_FILTER_NEAREST;
+		break;
+
+	case InterpolationModeBilinear:
+		filter = CAIRO_FILTER_BILINEAR;
+		break;
+
+	case InterpolationModeBicubic:
+		filter = CAIRO_FILTER_GAUSSIAN;
+		break;
+		
+	case InterpolationModeLowQuality:
+		filter = CAIRO_FILTER_FAST;
+		break;
+
+	case InterpolationModeDefault:
+	default:
+		filter = CAIRO_FILTER_GOOD;
+		break;
+	}
+
+	return filter;
+}
+
 float
 gdip_get_display_dpi()
 {
-    static float dpis = 0;
+	static float dpis = 0;
+	Display* display;
 
-    if (dpis == 0) {
-        if (GDIP_display == NULL) {
-            dpis = 72.0f;
-        } else {
-            char *val = XGetDefault(GDIP_display, "Xft", "dpi");
-            if (val) {
-                dpis = atof(val);
-            } else {
-		dpis = 72.0f;
-	    }
-        }
-    }
+	if (dpis == 0) {
+		char *val;
 
-    return dpis;
+		display = XOpenDisplay (0);
+		// If the display is openable lets try to read dpi from it; otherwise use a default of 96.0f
+		if (display) {
+			val = XGetDefault (display, "Xft", "dpi");
+			XCloseDisplay (display);
+			if (val) {
+				dpis = atof (val);
+			} else {
+				dpis = 96.0f;
+			}
+		} else {
+			dpis = 96.0f;
+		}
+	}
+
+	return dpis;
 }
 
 void gdip_unitConversion(Unit fromUnit, Unit toUnit, float nSrc, float* nTrg)
@@ -240,108 +228,6 @@ void gdip_unitConversion(Unit fromUnit, Unit toUnit, float nSrc, float* nTrg)
 }
 
 
-
-
-/*
-    Functionality to be replaced by Cairo when available from
-    the standard API
-*/
-
-/* Compute the amount that each basis vector is scaled by. */
-cairo_status_t
-_gdip_cairo_matrix_compute_scale_factors (cairo_matrix_t *matrix, double *sx, double *sy)
-{
-    double x, y;
-
-    x = 1.0;
-    y = 0.0;
-    cairo_matrix_transform_distance (matrix, &x, &y);
-    *sx = sqrt(x*x + y*y);
-
-    x = 0.0;
-    y = 1.0;
-    cairo_matrix_transform_distance (matrix, &x, &y);
-    *sy = sqrt(x*x + y*y);
-
-    return CAIRO_STATUS_SUCCESS;
-}
-
-static void
-_install_font_matrix(cairo_matrix_t *matrix, FT_Face face)
-{
-    cairo_matrix_t normalized;
-    double scale_x, scale_y;
-    double xx, xy, yx, yy, tx, ty;
-    FT_Matrix mat;
-
-    _gdip_cairo_matrix_compute_scale_factors (matrix, &scale_x, &scale_y);
-
-    cairo_matrix_copy (&normalized, matrix);
-
-    cairo_matrix_scale (&normalized, 1.0 / scale_x, 1.0 / scale_y);
-    cairo_matrix_get_affine (&normalized,
-                             &xx /* 00 */ , &yx /* 01 */,
-                             &xy /* 10 */, &yy /* 11 */,
-                             &tx, &ty);
-
-    mat.xx = DOUBLE_TO_16_16(xx);
-    mat.xy = -DOUBLE_TO_16_16(xy);
-    mat.yx = -DOUBLE_TO_16_16(yx);
-    mat.yy = DOUBLE_TO_16_16(yy);
-
-    FT_Set_Transform(face, &mat, NULL);
-    FT_Set_Char_Size(face,
-		     DOUBLE_TO_26_6(scale_x),
-		     DOUBLE_TO_26_6(scale_y),
-		     0, 0);
-}
-
-int
-gdpi_utf8_to_glyphs (cairo_font_t	*font,
-                 cairo_matrix_t		matrix,
-		 const unsigned char	*utf8,
-		 double			x0,
-		 double			y0,
-		 cairo_glyph_t		**glyphs,
-		 size_t			*nglyphs)
-{
-    FT_Face face = cairo_ft_font_face(font);
-    double		x = 0.0;
-    double		y = 0.0;
-    size_t		i;
-    FT_ULong		*ucs4 = NULL;
-
-    ucs4 = (FT_ULong *)g_utf8_to_ucs4 (utf8, (glong)-1, NULL, (glong *)nglyphs, NULL);
-
-    if (ucs4 == NULL)
-        return 0;
-
-    *glyphs = (cairo_glyph_t *) malloc ((*nglyphs) * (sizeof (cairo_glyph_t)));
-    if (*glyphs == NULL)
-    {
-        g_free (ucs4);
-        return 0;
-    }
-
-    _install_font_matrix (&matrix, face);
-
-    for (i = 0; i < *nglyphs; i++)
-    {
-        (*glyphs)[i].index = FT_Get_Char_Index (face, ucs4[i]);
-        (*glyphs)[i].x = x0 + x;
-        (*glyphs)[i].y = y0 + y;
-
-        FT_Load_Glyph (face, (*glyphs)[i].index, FT_LOAD_DEFAULT);
-
-        x += DOUBLE_FROM_26_6 (face->glyph->advance.x);
-        y -= DOUBLE_FROM_26_6 (face->glyph->advance.y);
-    }
-
-    g_free (ucs4);
-    return 1;
-}
-
-
 /* Curve handling
  *
  * See Richard Rasala, Explicit Cubic Spline Interpolation Formulas, in
@@ -509,3 +395,30 @@ fcmp (double x1, double x2, double epsilon)
         else /* -delta <= difference <= delta */
                 return 0;  /* x1 == x2 */
 }
+
+/* This function is used by gradient brushes for calculating the erf required for 
+ * calculating the integral of the normal distribution equation.
+ */
+float gdip_erf (float x, float std, float mean)
+{
+	/* std refers to standard deviation
+	 *
+	 * ERF: http://mathworld.wolfram.com/Erf.html
+	 * 
+	 * erf (z) = (2 / sqrt (pi)) * infinite sum of [(pow (-1, n) * pow (z, 2n+1))/(n! * (2n+1))]
+	 * using Maclaurin series.
+	 */
+	float series;
+	float constant = 2.0 / sqrt (PI);
+	float z = (x - mean) / std;
+	z /= sqrt (2.0);
+
+	series = (z
+		  - pow (z, 3) / 3.0
+		  + pow (z, 5) / 10.0
+		  - pow (z, 7) / 42.0
+		  + pow (z, 9) / 216.0
+		  - pow (z, 11) / 1420.0);
+
+	return constant * series;
+}
