@@ -81,6 +81,12 @@ gdip_getcodecinfo_gif ()
 }
 
 /* Read callback function for the gif libbrary*/
+static int 
+gdip_gif_fileinputfunc (GifFileType *gif, GifByteType *data, int len) 
+{
+	return fread (data, 1, len, (FILE*) gif->UserData);
+}
+
 static int 
 gdip_gif_inputfunc (GifFileType *gif, GifByteType *data, int len) 
 {
@@ -119,18 +125,28 @@ gdip_load_gif_image (void *stream, GpImage **image, bool from_file)
 	
 	GpBitmap *img = NULL;
 	ColorMapObject *pal = NULL;
+	ColorMapObject *localPalObj = NULL;
+	ColorPalette *localPal = NULL;
 	guchar *pixels = NULL;
 	guchar *readptr, *writeptr;
-	int i, j, l;
+	int i, j, l, pixelLength;
 	int imgCount;
 	int extBlockCount;
 	bool pageDimensionFound = FALSE;
 	bool timeDimensionFound = FALSE;
 	int pageDimensionCount = 0;
 	int timeDimensionCount = 0;
+	bool timeDimension = FALSE;
+	int timeCounter = 0;
+	int pageCounter = 0;
+	BitmapData data;
+	GifImageDesc imgDesc;
+
+	data.ByteCount = 0;
+	data.Bytes = NULL;	
 
 	if (from_file) 
-		gif = DGifOpenFileHandle (fileno ((FILE*)stream));
+		gif = DGifOpen(stream, &gdip_gif_fileinputfunc);
 	else 		
 		gif = DGifOpen (stream, &gdip_gif_inputfunc);
 	
@@ -143,7 +159,6 @@ gdip_load_gif_image (void *stream, GpImage **image, bool from_file)
 
 	img = gdip_bitmap_new ();
 	img->image.type = imageBitmap;
-	img->image.graphics = 0;
 	img->image.width = gif->SWidth;
 	img->image.height = gif->SHeight;
 
@@ -175,21 +190,25 @@ gdip_load_gif_image (void *stream, GpImage **image, bool from_file)
 		
 	if (pageDimensionFound && timeDimensionFound){
 		img->image.frameDimensionCount = 2;
-		img->image.frameDimensionList = (FrameDimensionInfo *) GdipAlloc (sizeof (FrameDimensionInfo)*2);
+		img->image.frameDimensionList = (FrameInfo *) GdipAlloc (sizeof (FrameInfo)*2);
 		img->image.frameDimensionList[0].count = pageDimensionCount;
 		memcpy (&(img->image.frameDimensionList[0].frameDimension), &gdip_image_frameDimension_page_guid, sizeof (CLSID));
+		img->image.frameDimensionList[0].frames = (BitmapData *) GdipAlloc (sizeof (BitmapData)*pageDimensionCount);
 		img->image.frameDimensionList[1].count = timeDimensionCount;
 		memcpy (&(img->image.frameDimensionList[1].frameDimension), &gdip_image_frameDimension_time_guid, sizeof (CLSID));
+		img->image.frameDimensionList[1].frames = (BitmapData *) GdipAlloc (sizeof (BitmapData)*timeDimensionCount);
 	} else if (pageDimensionFound) {
 		img->image.frameDimensionCount = 1;
-		img->image.frameDimensionList = (FrameDimensionInfo *) GdipAlloc (sizeof (FrameDimensionInfo));
+		img->image.frameDimensionList = (FrameInfo *) GdipAlloc (sizeof (FrameInfo));
 		img->image.frameDimensionList[0].count = pageDimensionCount;
 		memcpy (&(img->image.frameDimensionList[0].frameDimension), &gdip_image_frameDimension_page_guid, sizeof (CLSID));
+		img->image.frameDimensionList[0].frames = (BitmapData *) GdipAlloc (sizeof (BitmapData)*pageDimensionCount);
 	} else if (timeDimensionFound) {
 		img->image.frameDimensionCount = 1;
-		img->image.frameDimensionList = (FrameDimensionInfo *) GdipAlloc (sizeof (FrameDimensionInfo));
+		img->image.frameDimensionList = (FrameInfo *) GdipAlloc (sizeof (FrameInfo));
 		img->image.frameDimensionList[0].count = timeDimensionCount;
 		memcpy (&(img->image.frameDimensionList[0].frameDimension), &gdip_image_frameDimension_time_guid, sizeof (CLSID));
+		img->image.frameDimensionList[0].frames = (BitmapData *) GdipAlloc (sizeof (BitmapData)*timeDimensionCount);
 	}
 
 	/* Note that Cairo/libpixman does not have support for indexed
@@ -207,7 +226,7 @@ gdip_load_gif_image (void *stream, GpImage **image, bool from_file)
 						gif->SColorMap->Colors[i].Green,
 						gif->SColorMap->Colors[i].Blue);
 		}
-	
+		
 		img->image.palette = pal;
 	}
 	img->image.pixFormat = Format8bppIndexed;
@@ -218,39 +237,88 @@ gdip_load_gif_image (void *stream, GpImage **image, bool from_file)
 	img->image.pixFormat = Format32bppArgb;
 	
 	img->cairo_format = CAIRO_FORMAT_ARGB32;
-	img->data.PixelFormat = img->image.pixFormat;
-	img->data.Width = img->image.width;
-	img->data.Height = img->image.height;
-	img->data.Stride = img->data.Width * 4;
-	
-	pixels = GdipAlloc (img->data.Stride * img->data.Height);
-	
-	readptr = gif->SavedImages->RasterBits;
-	writeptr = pixels;
-	for (i = 0; i < img->image.width * img->image.height; i++) {
-		guchar pix = *readptr++;
-		if (pal) {
-		*writeptr++ = pal->Colors[pix].Blue;
-		*writeptr++ = pal->Colors[pix].Green;
-		*writeptr++ = pal->Colors[pix].Red;
-		} else {
-		*writeptr++ = pix;
-		*writeptr++ = pix;
-		*writeptr++ = pix;
-		}
-		*writeptr++ = 255;  /* A */
-	}
 
-	img->data.Scan0 = pixels;
-	img->data.Reserved = GBD_OWN_SCAN0;
+	/* 
+	 * Now populate the frames associated with each frame dimension
+	 */
+	imgCount = gif->ImageCount;
+	for (i=0; i<imgCount; i++) {
+		SavedImage si = gif->SavedImages[i];
+		extBlockCount = si.ExtensionBlockCount;
+		for (l =0; l<extBlockCount; l++) {
+			ExtensionBlock eb = si.ExtensionBlocks[l];
+			if (eb.Function == 249){
+				data.ByteCount= eb.ByteCount;
+				data.Bytes = (char *) GdipAlloc (data.ByteCount);
+				memcpy (data.Bytes, eb.Bytes, data.ByteCount);
+				timeDimension = TRUE;
+				break; /*there can be only one Graphic Control Extension before ImageData*/
+			}			
+		}
+		
+		/* copy the local color map if there is one*/
+		localPalObj= si.ImageDesc.ColorMap;
+		if (localPalObj != NULL) {
+			localPal = g_malloc (sizeof(ColorPalette) + sizeof(ARGB) * localPalObj->ColorCount);
+			localPal->Flags = 0;
+			localPal->Count = localPalObj->ColorCount;
+			for (j = 0; j < localPalObj->ColorCount; j++) {
+				localPal->Entries[j] = MAKE_ARGB_RGB(localPalObj->Colors[j].Red,
+								localPalObj->Colors[j].Green,
+								localPalObj->Colors[j].Blue);
+			}
+		}
+
+                data.PixelFormat = img->image.pixFormat;
+		imgDesc = si.ImageDesc;
+		data.Width = imgDesc.Width;
+		data.Height = imgDesc.Height;
+		data.Stride = data.Width * 4;
+		data.Top = imgDesc.Top;
+		data.Left = imgDesc.Left;
 	
-	img->image.surface = cairo_surface_create_for_image (pixels, img->cairo_format,
+		pixels = GdipAlloc (data.Stride * data.Height);
+	
+		readptr = si.RasterBits;
+		writeptr = pixels;
+		pixelLength = data.Width * data.Height;
+		/*
+		 * While loading images, local color map if present takes precedence
+		 * over global color map.
+		 */
+		for (j = 0; j < pixelLength; j++) {
+			guchar pix = *readptr++;
+			if (localPalObj) {
+				set_pixel_bgra(writeptr, 0, localPalObj->Colors[pix].Blue, localPalObj->Colors[pix].Green, localPalObj->Colors[pix].Red, 0xff);
+			} else if (pal) {
+				set_pixel_bgra(writeptr, 0, pal->Colors[pix].Blue, pal->Colors[pix].Green, pal->Colors[pix].Red, 0xff);
+			} else {
+				set_pixel_bgra(writeptr, 0, pix, pix, pix, 0xff);
+			}
+			writeptr += 4;
+		}
+
+		data.Scan0 = pixels;
+		data.Reserved = GBD_OWN_SCAN0;
+
+		if (pageDimensionFound && timeDimensionFound){
+			if (timeDimension)
+				img->image.frameDimensionList [1].frames [timeCounter ++] = data;
+			else
+				img->image.frameDimensionList [0].frames [pageCounter ++] = data;
+		} else if (pageDimensionFound){
+			img->image.frameDimensionList[0].frames [pageCounter++] = data;
+		} else if (timeDimensionFound){
+			img->image.frameDimensionList[0].frames [timeCounter++] = data;
+		}				
+	}
+	
+	img->data = img->image.frameDimensionList[0].frames[0];
+	
+	img->image.surface = cairo_surface_create_for_image ((char *)img->data.Scan0, img->cairo_format,
 							img->image.width, img->image.height,
 							img->data.Stride);
-	img->image.imageFlags =
-	ImageFlagsReadOnly |
-	ImageFlagsHasRealPixelSize |
-	ImageFlagsColorSpaceRGB;
+	img->image.imageFlags = ImageFlagsReadOnly | ImageFlagsHasRealPixelSize | ImageFlagsColorSpaceRGB;
 	img->image.horizontalResolution = 0;
 	img->image.verticalResolution = 0;
 	img->image.propItems = NULL;
@@ -260,7 +328,7 @@ gdip_load_gif_image (void *stream, GpImage **image, bool from_file)
 	return Ok;
 	error:	
 	if (pixels)
-	GdipFree (pixels);
+		GdipFree (pixels);
 	if (img)
 		gdip_bitmap_dispose (img);
 	if (gif)
@@ -282,7 +350,7 @@ gdip_gif_outputfunc (GifFileType *gif,  const GifByteType *data, int len)
 
 
 GpStatus 
-gdip_save_gif_image_to_file (unsigned char *filename, GpImage *image)
+gdip_save_gif_image_to_file (char *filename, GpImage *image)
 {
 	return gdip_save_gif_image ( (void *)filename, image, TRUE);
 }
@@ -300,71 +368,128 @@ gdip_save_gif_image (void *stream, GpImage *image, bool from_file)
 {
 	GifFileType *fp;
 	int i, row, x, y, size;
-	ColorMapObject *color_map;
 	GpBitmap *bitmap = (GpBitmap *) image;
-	GifByteType *red, *green, *blue, *pixels;
+	GifByteType *red = NULL, *green = NULL, *blue = NULL, *pixels = NULL;
 	GifByteType *ptr_red, *ptr_green, *ptr_blue, *ptr_pixels;
-	ARGB color;	
+	ARGB color = 0;	
 	int cmap_size = 256;	
-	ColorMapObject* cmap = NULL;	
+	ColorMapObject *cmap = NULL;	
 	BOOL error = FALSE;
+	BitmapData data;
+	int dimensionCount;
+	int frameCount;
+	int j,k;
+	int sizeAlloc;
+	unsigned char * v;
+	BOOL animationFlag = FALSE;
 	
 	if (!stream) 
 		return InvalidParameter;
-		
+
 	if (from_file)
-		fp = EGifOpenFileName ((unsigned char *) stream, 0);
+		fp = EGifOpenFileName (stream, 0);
 	else
 		fp = EGifOpen (stream, gdip_gif_outputfunc);
 		
 	if (!fp)
 		return FileNotFound;
 		
-	size = bitmap->data.Height * bitmap->data.Width;
-	ptr_red = red = malloc (sizeof (GifByteType)* size);
-	ptr_green = green = malloc (sizeof (GifByteType)* size);
-	ptr_blue = blue = malloc (sizeof (GifByteType)* size);
-	ptr_pixels = pixels = malloc (sizeof (GifByteType)* bitmap->data.Height * bitmap->data.Stride);
-	cmap = MakeMapObject (cmap_size, 0);	
-	
-	for (y = 0; y <bitmap->data.Height; y++) {	
-		for (x = 0; x <bitmap->data.Width; x++) {
-		
-			GdipBitmapGetPixel (bitmap, x, y, &color);		
+	dimensionCount = image->frameDimensionCount;
+	for (j = 0; j < dimensionCount; j++) {
+		frameCount = image->frameDimensionList [j].count;
+		if (!memcmp (&(image->frameDimensionList [j].frameDimension), 
+				&gdip_image_frameDimension_time_guid, sizeof (CLSID)))
+			animationFlag = TRUE;
+		else
+			animationFlag = FALSE;
+
+		for (k = 0; k < frameCount; k++) {
+			cmap_size = 256;
+			cmap  = MakeMapObject (cmap_size, 0);
 			
-			*ptr_red++ = (color & 0x00ff0000) >> 16;
-			*ptr_green++ = (color & 0x0000ff00) >> 8;
-			*ptr_blue++ =  (color & 0x000000ff);
-		}	
-	}	
+			data = image->frameDimensionList [j].frames [k]; 
+			size = data.Height * data.Width;
+			sizeAlloc = sizeof (GifByteType)* size;
+			ptr_red  = red = GdipAlloc (sizeAlloc);
+			ptr_green = green = GdipAlloc (sizeAlloc);
+			ptr_blue = blue = GdipAlloc (sizeAlloc);
+			ptr_pixels = pixels = GdipAlloc (sizeAlloc);
 
-	if (QuantizeBuffer (bitmap->data.Width, bitmap->data.Height, &cmap_size,
-		red, green, blue, pixels, cmap->Colors) == GIF_ERROR) 
-		error = TRUE;
+			
+			for (y = 0; y < data.Height; y++) {	
+				v = ((unsigned char *)data.Scan0) + y * data.Stride;
+				for (x = 0; x < data.Width; x++) {
+					v += 4;
+					color = (v [0]) | (v [1] << 8) | (v [2] << 16) | (v [3] << 24);
+										
+#ifdef WORDS_BIGENDIAN
+					*ptr_red++ = (color >> 8) & 0xff;
+					*ptr_green++ = (color >> 16) & 0xff;
+					*ptr_blue++ =  (color >> 24) & 0xff;
+#else
+					*ptr_red++ = (color & 0x00ff0000) >> 16;
+					*ptr_green++ = (color & 0x0000ff00) >> 8;
+					*ptr_blue++ =  (color & 0x000000ff);
+#endif
+				}	
+			}	
+			
+			if (QuantizeBuffer (data.Width, data.Height, &cmap_size, red, 
+					green, blue, pixels, cmap->Colors) == GIF_ERROR) 
+				error = TRUE;
+			
+			cmap->BitsPerPixel = BitSize (cmap_size);
+			cmap->ColorCount = 1 << cmap->BitsPerPixel;
+
+			/*Make info from first frame as Global color map*/
+			if (k == 0){
+				if (j == 0){
+					if (EGifPutScreenDesc (fp, image->width, image->height,
+								8, 0, cmap) == GIF_ERROR) 
+						error = TRUE;
+				}
+
+				/*Uses information obtained from util directory of libungif 
+				originally written by Gershon Elber*/
+				if (animationFlag) {
+					/*Put application extension code*/
+					unsigned char extStr [3];
+					extStr [0] = 1 % 256;
+					extStr [1] = 1 / 256;
+					extStr [2] = 0;
+
+					EGifPutExtension (fp, APPLICATION_EXT_FUNC_CODE,
+										3, extStr);
+				}
+			}
 	
-	if (EGifPutScreenDesc (fp, bitmap->data.Width, bitmap->data.Height,
-		8, 0, cmap) == GIF_ERROR) 
-		error = TRUE;
+			/*Put Graphic control Extension*/
+			if (animationFlag) {
+				EGifPutExtension(fp, GRAPHICS_EXT_FUNC_CODE, data.ByteCount, data.Bytes);
+			}
+
+			if (EGifPutImageDesc (fp, data.Left, data.Top, data.Width, 
+						data.Height, FALSE, cmap) == GIF_ERROR) 
+				error = TRUE;
 	
-	if (EGifPutImageDesc (fp, 0, 0, bitmap->data.Width, bitmap->data.Height,
-		FALSE, NULL) == GIF_ERROR) 
-		error = TRUE;
-	
-	for (i = 0;  i < bitmap->data.Height;  ++i) {
-		if (EGifPutLine (fp, ptr_pixels, bitmap->data.Width) == GIF_ERROR) {
-			error = TRUE;
-			break;
+			for (i = 0;  i < data.Height;  ++i) {
+				if (EGifPutLine (fp, ptr_pixels, data.Width) == GIF_ERROR) {
+					error = TRUE;
+					break;
+				}
+				ptr_pixels += data.Width;
+			}
+
+			FreeMapObject (cmap);
+			GdipFree (red);
+			GdipFree (green);
+			GdipFree (blue);	
+			GdipFree (pixels);		
 		}
-		ptr_pixels += bitmap->data.Width;
 	}
-
+	
 	EGifCloseFile (fp);	
-	free (red);
-	free (green);
-	free (blue);	
-	free (pixels);		
-	free (cmap);
-
+	
 	return (error == FALSE) ? Ok : GenericError;
 }
 
@@ -386,7 +511,7 @@ gdip_load_gif_image_from_file (FILE *fp, GpImage **image)
 }
 
 GpStatus 
-gdip_save_gif_image_to_file (unsigned char *filename, GpImage *image)
+gdip_save_gif_image_to_file (char *filename, GpImage *image)
 {
 	return UnknownImageFormat;
 }
