@@ -11,7 +11,6 @@
 #ifndef _GIFCODEC_H
 #define _GIFCODEC_H
 #include "gdip.h"
-#include "gdip_win32.h"
 
 GpStatus 
 gdip_load_gif_image_from_file (FILE *fp, GpImage **image);
@@ -22,7 +21,7 @@ gdip_load_gif_image_from_stream_delegate (GetBytesDelegate getBytesFunc,
                                            GpImage **image);
 					   
 GpStatus 
-gdip_save_gif_image_to_file (unsigned char *filename, GpImage *image);
+gdip_save_gif_image_to_file (char *filename, GpImage *image);
 
 
 GpStatus
