@@ -18,9 +18,11 @@
  * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
  * OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
  *
- * Author: Duncan Mak (duncan@ximian.com)
+ * Authors:
+ *      Duncan Mak (duncan@ximian.com)
+ *      Ravindra (rkumar@novell.com)
  *
- * Copyright (C) 2003, Novell Inc.
+ * Copyright (C) 2003, Novell Inc. (http://www.novell.com)
  *
  */
  
@@ -73,9 +75,9 @@ int_to_float (const GpPoint *pts, int count)
         GpPoint *tmp = (GpPoint *) pts;
         int i;
 
-        for (i = 0; i < count; i++, p++, tmp++) {
-                p->X = (float) tmp->X;
-                p->Y = (float) tmp->Y;
+        for (i = 0; i < count; i++) {
+                p[i].X = (float) tmp[i].X;
+                p[i].Y = (float) tmp[i].Y;
         }
 
         return p;
@@ -85,33 +87,24 @@ static void
 append (GpPath *path, float x, float y, GpPathPointType type)
 {
         byte t = (byte) type;
-        GpPointF pt;
-        int count;
+	GpPointF pt;
 
-        GdipGetPointCount (path, &count);
-        pt.X = x;
-        pt.Y = y;
+	pt.X = x; pt.Y = y;
 
-        /* if we're adding a Start and we're already at pt, then drop it  */
-        if (type == PathPointTypeStart && count > 0) {
-
-                GpPointF current;
-                GdipGetPathLastPoint (path, &current);
-
-                /* if we closed a subpath, then append anyway */
-                byte type = g_array_index (path->types, byte, path->count);
-                if (type & PathPointTypeCloseSubpath)
-                        goto append;
-
-                if ((fcmp (x, current.X, FLT_EPSILON) == 0)
-                        && (fcmp (y, current.Y, FLT_EPSILON) == 0))
-                                return;
+	if (path->start_new_fig)
+		t = PathPointTypeStart;
+	/* if we closed a subpath, then start new figure and append */
+	else if (path->count > 0) {
+		type = g_array_index (path->types, byte, path->count - 1);
+		if (type & PathPointTypeCloseSubpath)
+		    t = PathPointTypeStart;
         }
 
-append:
-        g_array_append_val (path->points, pt);
-        g_byte_array_append (path->types, &t, 1);
-        path->count++;
+	g_array_append_val (path->points, pt);
+	g_byte_array_append (path->types, &t, 1);
+	path->count++;
+
+	path->start_new_fig = FALSE;
 }
 
 static void
@@ -160,24 +153,39 @@ append_curve (GpPath *path, const GpPointF *points, GpPointF *tangents, int coun
 }
 
 GpStatus
-GdipCreatePath (GpFillMode brushMode, GpPath **path)
+GdipCreatePath (GpFillMode fillMode, GpPath **path)
 {
-        *path = (GpPath *) GdipAlloc (sizeof (GpPath));
+	g_return_val_if_fail (path != NULL, InvalidParameter);
 
-        (*path)->fill_mode = brushMode;
-        (*path)->points = g_array_new (FALSE, FALSE, sizeof (GpPointF));
-        (*path)->types = g_byte_array_new ();
-        (*path)->count = 0;
+	*path = (GpPath *) GdipAlloc (sizeof (GpPath));
 
-        return Ok;
+	(*path)->fill_mode = fillMode;
+	(*path)->points = g_array_new (FALSE, FALSE, sizeof (GpPointF));
+	(*path)->types = g_byte_array_new ();
+	(*path)->count = 0;
+	(*path)->start_new_fig = TRUE;
+
+	return Ok;
 }
 
 GpStatus
 GdipCreatePath2 (const GpPointF *points, const byte *types,
                 int count, GpFillMode fillMode, GpPath **path)
 {
-        GArray *pts = array_to_g_array (points, count);
-        GByteArray *t = array_to_g_byte_array (types, count);
+	GArray *pts;
+	GByteArray *t;
+
+	g_return_val_if_fail (path != NULL, InvalidParameter);
+	g_return_val_if_fail (points != NULL, InvalidParameter);
+	g_return_val_if_fail (types != NULL, InvalidParameter);
+
+	/* FIXME: Check whether path types are valid before adding them. MS does
+	 * some checking and does not use the points passed in, if they look
+	 * invalid. Some cases are like last type being PathPointTypeStart or
+	 * PathPointTypeCloseSubpath etc.
+	 */
+	pts = array_to_g_array (points, count);
+	t = array_to_g_byte_array (types, count);
         
         *path = (GpPath *) GdipAlloc (sizeof (GpPath));
         (*path)->fill_mode = fillMode;
@@ -192,9 +200,14 @@ GpStatus
 GdipCreatePath2I (const GpPoint *points, const byte *types,
                 int count, GpFillMode fillMode, GpPath **path)
 {
-        GpPointF *pt = convert_points (points, count);
+	GpPointF *pt;
+        GpStatus s;
 
-        GpStatus s = GdipCreatePath2 (pt, types, count, fillMode, path);
+	g_return_val_if_fail (points != NULL, InvalidParameter);
+
+	pt = convert_points (points, count);
+	
+	s = GdipCreatePath2 (pt, types, count, fillMode, path);
 
         GdipFree (pt);
 
@@ -204,51 +217,83 @@ GdipCreatePath2I (const GpPoint *points, const byte *types,
 GpStatus
 GdipClonePath (GpPath *path, GpPath **clonePath)
 {
+	int i;
+	byte type;
+	GpPointF point;
+
+	g_return_val_if_fail (path != NULL, InvalidParameter);
+	g_return_val_if_fail (clonePath != NULL, InvalidParameter);
+
         *clonePath = (GpPath *) GdipAlloc (sizeof (GpPath));
         (*clonePath)->fill_mode = path->fill_mode;
         (*clonePath)->count = path->count;
-        (*clonePath)->points = path->points;
-        (*clonePath)->types = path->types;
-        
+        (*clonePath)->points = g_array_new (FALSE, FALSE, sizeof (GpPointF));
+        (*clonePath)->types = g_byte_array_new ();
+	for (i = 0; i < path->count; i++) {
+		point = g_array_index (path->points, GpPointF, i);
+		type = g_array_index (path->types, byte, i);
+		g_array_append_val ((*clonePath)->points, point);
+		g_byte_array_append ((*clonePath)->types, &type, 1);
+	}
+
+	(*clonePath)->start_new_fig = path->start_new_fig;
+
         return Ok;
 }
 
 GpStatus
 GdipDeletePath (GpPath *path)
 {
-        if (path->count != 0) {
-                if (path->points != NULL)
-                        g_array_free (path->points, TRUE);
-                
-                if (path->types != NULL)
-                        g_byte_array_free (path->types, TRUE);
-        }
-        
-        GdipFree (path);
-        return Ok;
+	g_return_val_if_fail (path != NULL, InvalidParameter);
+
+	if (path->points != NULL)
+		g_array_free (path->points, TRUE);
+	path->points = NULL;
+
+	if (path->types != NULL)
+		g_byte_array_free (path->types, TRUE);
+	path->types = NULL;
+
+	GdipFree (path);
+	return Ok;
 }
 
 GpStatus
 GdipResetPath (GpPath *path)
 {
-        path->points = g_array_new (FALSE, FALSE, sizeof (GpPointF));
-        path->types = g_byte_array_new ();
-        path->count = 0;
-        
-        return Ok;
+	g_return_val_if_fail (path != NULL, InvalidParameter);
+
+	if (path->points != NULL)
+		g_array_free (path->points, TRUE);
+
+	if (path->types != NULL)
+		g_byte_array_free (path->types, TRUE);
+
+	path->count = 0;
+	path->points = g_array_new (FALSE, FALSE, sizeof (GpPointF));
+	path->types = g_byte_array_new ();
+	path->fill_mode = FillModeAlternate;
+	path->start_new_fig = TRUE;
+
+	return Ok;
 }
 
 GpStatus
-GdipGetPointCount (GpPath *path, int *count)
+GdipGetPointCount (GDIPCONST GpPath *path, int *count)
 {
-        *count = path->count;
-        return Ok;
+	g_return_val_if_fail (path != NULL, InvalidParameter);
+	g_return_val_if_fail (count != NULL, InvalidParameter);
+
+	*count = path->count;
+	return Ok;
 }
 
 GpStatus
 GdipGetPathTypes (GpPath *path, byte *types, int count)
 {
-        int i;
+	int i;
+	g_return_val_if_fail (path != NULL, InvalidParameter);
+	g_return_val_if_fail (types != NULL, InvalidParameter);
         
         for (i = 0; i < count; i++)
                 types [i] = path->types->data [i];
@@ -257,9 +302,11 @@ GdipGetPathTypes (GpPath *path, byte *types, int count)
 }
 
 GpStatus
-GdipGetPathPoints (GpPath *path, GpPointF *points, int count)
+GdipGetPathPoints (GDIPCONST GpPath *path, GpPointF *points, int count)
 {
-        int i;
+	int i;
+	g_return_val_if_fail (path != NULL, InvalidParameter);
+	g_return_val_if_fail (points != NULL, InvalidParameter);
 
         for (i = 0; i < count; i++) {
 		GpPointF point = g_array_index (path->points, GpPointF, i);
@@ -271,9 +318,11 @@ GdipGetPathPoints (GpPath *path, GpPointF *points, int count)
 }
 
 GpStatus
-GdipGetPathPointsI (GpPath *path, GpPoint *points, int count)
+GdipGetPathPointsI (GDIPCONST GpPath *path, GpPoint *points, int count)
 {
-        int i;
+	int i;
+	g_return_val_if_fail (path != NULL, InvalidParameter);
+	g_return_val_if_fail (points != NULL, InvalidParameter);
 
         for (i = 0; i < count; i++) {
                 GpPoint point = g_array_index (path->points, GpPoint, i);
@@ -285,24 +334,32 @@ GdipGetPathPointsI (GpPath *path, GpPoint *points, int count)
 }
 
 GpStatus
-GdipGetPathFillMode (GpPath *path, GpFillMode *fillmode)
+GdipGetPathFillMode (GpPath *path, GpFillMode *fillMode)
 {
-        *fillmode = path->fill_mode;
+	g_return_val_if_fail (path != NULL, InvalidParameter);
+	g_return_val_if_fail (fillMode != NULL, InvalidParameter);
+
+        *fillMode = path->fill_mode;
         
         return Ok;
 }
 
 GpStatus
-GdipSetPathFillMode (GpPath *path, GpFillMode fillmode)
+GdipSetPathFillMode (GpPath *path, GpFillMode fillMode)
 {
-        path->fill_mode = fillmode;
+	g_return_val_if_fail (path != NULL, InvalidParameter);
+
+        path->fill_mode = fillMode;
         
         return Ok;
 }
 
 GpStatus
-GdipGetPathData (GpPath *path, GpPathData *pathData)
+GdipGetPathData (GDIPCONST GpPath *path, GpPathData *pathData)
 {
+	g_return_val_if_fail (path != NULL, InvalidParameter);
+	g_return_val_if_fail (pathData != NULL, InvalidParameter);
+
         pathData->Count = path->count;
         pathData->Points = g_array_to_array (path->points);
         pathData->Types = g_byte_array_to_array (path->types);
@@ -310,33 +367,85 @@ GdipGetPathData (GpPath *path, GpPathData *pathData)
         return Ok;
 }
 
-/* MonoTODO */
 GpStatus
 GdipStartPathFigure (GpPath *path)
 {
-        return Ok;
+	g_return_val_if_fail (path != NULL, InvalidParameter);
+
+	path->start_new_fig = TRUE;
+
+	return Ok;
 }
 
-/* MonoTODO */
 GpStatus
 GdipClosePathFigure (GpPath *path)
 {
-        return Ok;
+	byte current;
+	g_return_val_if_fail (path != NULL, InvalidParameter);
+
+	current = g_array_index (path->types, byte, path->count - 1);
+	g_byte_array_remove_index (path->types, path->count - 1);
+	current |= PathPointTypeCloseSubpath;
+	g_byte_array_append (path->types, &current, 1);
+
+	path->start_new_fig = TRUE;
+
+	return Ok;
 }
 
-/* MonoTODO */
 GpStatus
 GdipClosePathFigures (GpPath *path)
 {
+	int index = 0;
+	byte currentType;
+	byte lastType;
+	GByteArray *oldTypes;
+
+	g_return_val_if_fail (path != NULL, InvalidParameter);
+
+	/* first point is not closed */
+	if (path->count <= 1)
+		return Ok;
+
+	oldTypes = path->types;
+	path->types = g_byte_array_new ();
+
+	lastType = g_array_index (oldTypes, byte, index);
+	index++;
+
+	for (index = 1; index < path->count; index++) {
+		currentType = g_array_index (oldTypes, byte, index);
+		/* we dont close on the first point */
+		if ((currentType == PathPointTypeStart) && (index > 1)) {
+			lastType |= PathPointTypeCloseSubpath;
+			g_byte_array_append (path->types, &lastType, 1);
+		}
+		else
+			g_byte_array_append (path->types, &lastType, 1);
+
+		lastType = currentType;
+	}
+
+	/* close at the end */
+	lastType |= PathPointTypeCloseSubpath;
+	g_byte_array_append (path->types, &lastType, 1);
+
+	path->start_new_fig = TRUE;
+
+	g_byte_array_free (oldTypes, TRUE);
+
         return Ok;
 }
 
 GpStatus
 GdipSetPathMarker (GpPath *path)
 {
-        byte current = g_array_index (path->types, byte, path->count);
+	byte current;
+	g_return_val_if_fail (path != NULL, InvalidParameter);
 
-        g_byte_array_remove_index (path->types, path->count);
+	current = g_array_index (path->types, byte, path->count - 1);
+
+        g_byte_array_remove_index (path->types, path->count - 1);
 
         current |= PathPointTypePathMarker;
 
@@ -349,10 +458,15 @@ GpStatus
 GdipClearPathMarkers (GpPath *path)
 {
         int i;
-        GByteArray *cleared = g_byte_array_new ();
+	byte current;
+	GByteArray *cleared;
+
+	g_return_val_if_fail (path != NULL, InvalidParameter);
+
+	cleared = g_byte_array_new ();
 
         for (i = 0; i < path->count; i++) {
-                byte current = g_array_index (path->types, byte, i);
+                current = g_array_index (path->types, byte, i);
 
                 /* take out the marker if there is one */
                 if (current & PathPointTypePathMarker)
@@ -362,6 +476,7 @@ GdipClearPathMarkers (GpPath *path)
         }
 
         /* replace the existing with the cleared array */
+	g_byte_array_free (path->types, TRUE);
         path->types = cleared;
 
         return Ok;
@@ -370,10 +485,17 @@ GdipClearPathMarkers (GpPath *path)
 GpStatus
 GdipReversePath (GpPath *path)
 {
-        int length= path->count;
-        GByteArray *types = g_byte_array_sized_new (length);
-        GArray *points = g_array_sized_new (FALSE, TRUE, sizeof (GpPointF), length);
-        int i;
+	int length, i;
+	GByteArray *types;
+	GArray *points;
+
+	g_return_val_if_fail (path != NULL, InvalidParameter);
+
+	length = path->count;
+
+	types = g_byte_array_sized_new (length);
+	points = g_array_sized_new (FALSE, TRUE, sizeof (GpPointF), length);
+
         for (i = length; i > 0; i--) {
                 byte t = g_array_index (path->types, byte, i);
                 GpPointF pt = g_array_index (path->points, GpPointF, i);
@@ -390,24 +512,34 @@ GdipReversePath (GpPath *path)
 GpStatus
 GdipGetPathLastPoint (GpPath *path, GpPointF *lastPoint)
 {
-        *lastPoint = g_array_index (path->points, GpPointF, path->count);
-        return Ok;
+	g_return_val_if_fail (path != NULL, InvalidParameter);
+	g_return_val_if_fail (lastPoint != NULL, InvalidParameter);
+	g_return_val_if_fail (path->count > 0, InvalidParameter);
+
+	*lastPoint = g_array_index (path->points, GpPointF, path->count - 1);
+	return Ok;
 }
 
 GpStatus
 GdipAddPathLine (GpPath *path, float x1, float y1, float x2, float y2)
 {
-        append (path, x1, y1, PathPointTypeStart);
-        append (path, x2, y2, PathPointTypeLine);
+	g_return_val_if_fail (path != NULL, InvalidParameter);
 
-        return Ok;
+	append (path, x1, y1, PathPointTypeLine);
+	append (path, x2, y2, PathPointTypeLine);
+
+	return Ok;
 }
 
 GpStatus
 GdipAddPathLine2 (GpPath *path, const GpPointF *points, int count)
 {
-        int i;
-        GpPointF *tmp = (GpPointF *) points;
+	int i;
+	GpPointF *tmp;
+	g_return_val_if_fail (path != NULL, InvalidParameter);
+	g_return_val_if_fail (points != NULL, InvalidParameter);
+
+	tmp = (GpPointF *) points;
 
         for (i = 0; i < count; i++, tmp++)
                 append_point (path, *tmp, PathPointTypeLine);
@@ -459,6 +591,7 @@ GdipAddPathArc (GpPath *path, float x, float y,
                 float width, float height, float startAngle, float sweepAngle)
 {
         float endAngle = startAngle + sweepAngle;
+	g_return_val_if_fail (path != NULL, InvalidParameter);
         
         if (sweepAngle < 180)
                 append_arc (path, TRUE, x, y, width, height, startAngle, endAngle);
@@ -477,7 +610,9 @@ GdipAddPathBezier (GpPath *path,
                    float x1, float y1, float x2, float y2, 
                    float x3, float y3, float x4, float y4)
 {
-        append (path, x1, y1, PathPointTypeStart);
+	g_return_val_if_fail (path != NULL, InvalidParameter);
+
+        append (path, x1, y1, PathPointTypeLine);
         append_bezier (path, x2, y2, x3, y3, x4, y4);
         
         return Ok;
@@ -486,11 +621,12 @@ GdipAddPathBezier (GpPath *path,
 GpStatus
 GdipAddPathBeziers (GpPath *path, const GpPointF *points, int count)
 {
-        int i;
+	int i;
+	GpPointF *tmp = (GpPointF *) points;
+	g_return_val_if_fail (path != NULL, InvalidParameter);
+	g_return_val_if_fail (points != NULL, InvalidParameter);
         
-        GpPointF *tmp = (GpPointF *) points;
-        
-        append_point (path, *tmp, PathPointTypeStart);
+        append_point (path, *tmp, PathPointTypeLine);
         tmp++;
 
         for (i = 1; i < count; i++, tmp++)
@@ -516,6 +652,8 @@ GdipAddPathCurve3 (GpPath *path, const GpPointF *points, int count,
         int offset, int numberOfSegments, float tension)
 {
         GpPointF *tangents;
+	g_return_val_if_fail (path != NULL, InvalidParameter);
+	g_return_val_if_fail (points != NULL, InvalidParameter);
 
         tangents = gdip_open_curve_tangents (CURVE_MIN_TERMS, points, count, tension);
 
@@ -536,12 +674,15 @@ GpStatus
 GdipAddPathClosedCurve2 (GpPath *path, const GpPointF *points, int count, float tension)
 {
         GpPointF *tangents;
+	g_return_val_if_fail (points != NULL, InvalidParameter);
 
         tangents = gdip_closed_curve_tangents (CURVE_MIN_TERMS, points, count, tension);
 
         append_curve (path, points, tangents, count, CURVE_CLOSE);
 
-        GdipFree (tangents);
+	/* close the path */
+	GdipClosePathFigure (path);
+	GdipFree (tangents);
 
         return Ok;
 }
@@ -549,10 +690,12 @@ GdipAddPathClosedCurve2 (GpPath *path, const GpPointF *points, int count, float
 GpStatus
 GdipAddPathRectangle (GpPath *path, float x, float y, float width, float height)
 {
-        append (path, x, y, PathPointTypeLine);
+	g_return_val_if_fail (path != NULL, InvalidParameter);
+
+        append (path, x, y, PathPointTypeStart);
         append (path, x + width, y, PathPointTypeLine);
         append (path, x + width, y + height, PathPointTypeLine);
-        append (path, x, y + height, PathPointTypeLine);
+        append (path, x, y + height, PathPointTypeLine | PathPointTypeCloseSubpath);
         
         return Ok;
 }
@@ -561,11 +704,14 @@ GpStatus
 GdipAddPathRectangles (GpPath *path, const GpRectF *rects, int count)
 {
         int i;
-        for (i = 0; i < count; i++, rects++) {
-                float x = rects->X;
-                float y = rects->Y;
-                float width = rects->Width;
-                float height = rects->Height;
+	g_return_val_if_fail (path != NULL, InvalidParameter);
+	g_return_val_if_fail (rects != NULL, InvalidParameter);
+
+        for (i = 0; i < count; i++) {
+                float x = rects[i].X;
+                float y = rects[i].Y;
+                float width = rects[i].Width;
+                float height = rects[i].Height;
                 GdipAddPathRectangle (path, x, y, width, height);
         }
         
@@ -580,6 +726,7 @@ GdipAddPathEllipse (GpPath *path, float x, float y, float width, float height)
         double ry = height / 2;
         double cx = x + rx;
         double cy = y + ry;
+	g_return_val_if_fail (path != NULL, InvalidParameter);
 
         /* origin */
         append (path, cx + rx, cy, PathPointTypeStart);
@@ -607,6 +754,9 @@ GdipAddPathEllipse (GpPath *path, float x, float y, float width, float height)
                        cx + C1 * rx, cy + ry, 
                        cx + rx, cy + C1 * ry, 
                        cx + rx, cy);
+
+	/* close the path */
+	GdipClosePathFigure (path);
         
         return Ok;
 }
@@ -629,6 +779,8 @@ GdipAddPathPie (GpPath *path, float x, float y, float width, float height, float
         float sin_alpha = sin (alpha);
         float cos_alpha = cos (alpha);
 
+	g_return_val_if_fail (path != NULL, InvalidParameter);
+
         /* move to center */
         append (path, cx, cy, PathPointTypeStart);
         
@@ -659,6 +811,8 @@ GdipAddPathPolygon (GpPath *path, const GpPointF *points, int count)
 {
         int i;
         GpPointF *tmp = (GpPointF *) points;
+	g_return_val_if_fail (path != NULL, InvalidParameter);
+	g_return_val_if_fail (points != NULL, InvalidParameter);
         
         append_point (path, *tmp, PathPointTypeStart);
         tmp ++;
@@ -680,28 +834,48 @@ GpStatus
 GdipAddPathPath (GpPath *path, GpPath *addingPath, bool connect)
 {
         int i, length;
+	GpPathPointType first = PathPointTypeStart;
+        GpPointF *pts;
+        byte *types;
+
+	g_return_val_if_fail (path != NULL, InvalidParameter);
+	g_return_val_if_fail (addingPath != NULL, InvalidParameter);
 
         GdipGetPointCount (addingPath, &length);
 
         if (length < 1)
                 return Ok;
         
-        GpPointF pts [length];
-        byte types [length];
+        pts = calloc(sizeof(GpPointF), length);
+        types = calloc(sizeof(byte), length);
 
         GdipGetPathPoints (addingPath, pts, length);
         GdipGetPathTypes (addingPath, types, length);
 
-        if (connect)
-                append_point (path, pts [0], PathPointTypeLine);
-        else                             
-                append_point (path, pts [0], PathPointTypeStart);
-                
+	/* We can connect only open figures. If first figure is closed
+	 * it can't be connected.
+	 */
+        if (connect) {
+		for (i = 1; i < length; i++) {
+			if (types [i] == PathPointTypeStart) {
+				if (types [i - 1] & PathPointTypeCloseSubpath)
+					first = PathPointTypeStart;
+				else
+					first = PathPointTypeLine;
+				break;
+			}
+		}
+	}
 
-        for (i = 1; i < length; i++)
-                append_point (path, pts [i], types [i]);
+	append_point (path, pts [0], first); 
 
-        return Ok;
+	for (i = 1; i < length; i++)
+		append_point (path, pts [i], types [i]);
+
+	free(pts);
+	free(types);
+
+	return Ok;
 }
 
 /* XXX: This one is really hard. They really translate a string into
@@ -712,7 +886,7 @@ GdipAddString (GpPath *path, const char *string, int length,
                 const GpFontFamily *family, int style, float emSize,
                 const GpRectF *layoutRect, const GpStringFormat *format)
 { 
-        return Ok;
+	return NotImplemented;
 }
 
 /* MonoTODO */
@@ -721,34 +895,27 @@ GdipAddStringI (GpPath *path, const char *string, int length,
                 const GpFontFamily *family, int style, float emSize,
                 const GpRect *layoutRect, const GpStringFormat *format)
 {
-        return Ok;
+	return NotImplemented;
 }
 
 GpStatus
 GdipAddPathLineI (GpPath *path, int x1, int y1, int x2, int y2)
 {
-        append (path, x1, y1, PathPointTypeStart);
-        
-        append (path, x2, y2, PathPointTypeLine);
-
-        return Ok;
+	return GdipAddPathLine (path, x1, y1, x2, y2);
 }
 
 GpStatus
 GdipAddPathLine2I (GpPath* path, const GpPoint *points, int count)
 {
-        int i;
-        GpPointF *tmp = int_to_float (points, count);
-        
-        append_point (path, *tmp, PathPointTypeStart);
-        tmp++;
+	GpPointF *tmp;
+	GpStatus status;
+	g_return_val_if_fail (points != NULL, InvalidParameter);
 
-        for (i = 1; i < count; i++, tmp++)
-                append_point (path, *tmp, PathPointTypeLine);
+	tmp = int_to_float (points, count);
+	status = GdipAddPathLine2 (path, tmp, count);
+	GdipFree (tmp);
 
-        GdipFree (tmp);
-
-        return Ok;
+	return status;
 }
 
 GpStatus
@@ -766,8 +933,13 @@ GdipAddPathBezierI (GpPath *path, int x1, int y1, int x2, int y2, int x3, int y3
 GpStatus
 GdipAddPathBeziersI (GpPath *path, const GpPoint *points, int count)
 {
-        GpPointF *tmp = int_to_float (points, count);
-        Status s = GdipAddPathBeziers (path, tmp, count);
+	GpPointF *tmp;
+        Status s;
+
+	g_return_val_if_fail (points != NULL, InvalidParameter);
+
+	tmp = int_to_float (points, count);
+	s = GdipAddPathBeziers (path, tmp, count);
 
         GdipFree (tmp);
 
@@ -790,9 +962,14 @@ GpStatus
 GdipAddPathCurve3I (GpPath *path, const GpPoint *points,
                     int count, int offset, int numberOfSegments, float tension)
 {
-        GpPointF *pt = convert_points (points, count);
+        GpPointF *pt;
+        Status s;
 
-        Status s = GdipAddPathCurve3 (path, pt, count, offset, numberOfSegments, tension);
+	g_return_val_if_fail (points != NULL, InvalidParameter);
+
+	pt = convert_points (points, count);
+
+	s = GdipAddPathCurve3 (path, pt, count, offset, numberOfSegments, tension);
 
         GdipFree (pt);
 
@@ -808,9 +985,14 @@ GdipAddPathClosedCurveI (GpPath *path, const GpPoint *points, int count)
 GpStatus
 GdipAddPathClosedCurve2I (GpPath *path, const GpPoint *points, int count, float tension)
 {
-        GpPointF *pt = convert_points (points, count);
+	GpPointF *pt;
+	Status s;
 
-        Status s = GdipAddPathClosedCurve2 (path, pt, count, tension);
+	g_return_val_if_fail (points != NULL, InvalidParameter);
+
+	pt = convert_points (points, count);
+
+	s = GdipAddPathClosedCurve2 (path, pt, count, tension);
 
         GdipFree (pt);
 
@@ -827,11 +1009,13 @@ GpStatus
 GdipAddPathRectanglesI (GpPath *path, const GpRect *rects, int count)
 {
         int i;
-        for (i = 0; i < count; i++, rects++) {
-                float x = (float) rects->X;
-                float y = (float) rects->Y;
-                float width =  (float) rects->Width;
-                float height =  (float) rects->Height;
+	g_return_val_if_fail (rects != NULL, InvalidParameter);
+
+        for (i = 0; i < count; i++) {
+                float x = (float) rects[i].X;
+                float y = (float) rects[i].Y;
+                float width =  (float) rects[i].Width;
+                float height =  (float) rects[i].Height;
                 GdipAddPathRectangle (path, x, y, width, height);
         }
 
@@ -853,9 +1037,14 @@ GdipAddPathPieI (GpPath *path, int x, int y, int width, int height, float startA
 GpStatus
 GdipAddPathPolygonI (GpPath *path, const GpPoint *points, int count)
 {
-        GpPointF *tmp = int_to_float (points, count);
+	GpPointF *tmp;	
+        Status s;
 
-        Status s = GdipAddPathPolygon (path, tmp, count);
+	g_return_val_if_fail (points != NULL, InvalidParameter);
+
+	tmp = int_to_float (points, count);
+
+	s = GdipAddPathPolygon (path, tmp, count);
 
         GdipFree (tmp);
 
@@ -866,21 +1055,21 @@ GdipAddPathPolygonI (GpPath *path, const GpPoint *points, int count)
 GpStatus 
 GdipFlattenPath (GpPath *path, GpMatrix *matrix, float flatness)
 {
-        return Ok;
+	return NotImplemented;
 }
 
 /* MonoTODO */
 GpStatus 
 GdipWindingModeOutline (GpPath *path, GpMatrix *matrix, float flatness)
 {
-        return Ok;
+	return NotImplemented;
 }
 
 /* MonoTODO */
 GpStatus 
 GdipWidenPath (GpPath *nativePath, GpPen *pen, GpMatrix *matrix, float flatness)
 {
-        return Ok;
+	return NotImplemented;
 }
 
 /* MonoTODO */
@@ -889,16 +1078,22 @@ GdipWarpPath (GpPath *nativePath, GpMatrix *matrix, const GpPointF *points, int
               float srcx, float srcy, float srcwidth, float srcheight,
               WarpMode warpMode, float flatness)
 {
-        return Ok;
+	return NotImplemented;
 }
 
 GpStatus 
 GdipTransformPath (GpPath* path, GpMatrix *matrix)
 {
-        PointF *points = g_array_to_array (path->points);
-        int count = path->count;
+        PointF *points;
+	int count;
+        Status s;
 
-        Status s = GdipTransformMatrixPoints (matrix, points, count);
+	g_return_val_if_fail (path != NULL, InvalidParameter);
+
+	points = g_array_to_array (path->points);
+	count = path->count;
+
+	s = GdipTransformMatrixPoints (matrix, points, count);
 
         path->points = array_to_g_array (points, count);
 
@@ -911,40 +1106,76 @@ GdipTransformPath (GpPath* path, GpMatrix *matrix)
 GpStatus 
 GdipGetPathWorldBounds (GpPath *path, GpRectF *bounds, const GpMatrix *matrix, const GpPen *pen)
 {
-        return Ok;
+	static int called = 0;
+
+	if (!called) {
+		printf("NOT IMPLEMENTED: GdipGetPathWorldBounds (GpPath *path, GpRectF *bounds, const GpMatrix *matrix, const GpPen *pen)\n");
+		called = 1;
+	}
+	return Ok;
 }
 
 /* MonoTODO */
 GpStatus 
 GdipGetPathWorldBoundsI (GpPath *path, GpRect *bounds, const GpMatrix *matrix, const GpPen *pen)
 {
-        return Ok;
+	static int called = 0;
+
+	if (!called) {
+		printf("NOT IMPLEMENTED: GdipGetPathWorldBoundsI (GpPath *path, GpRect *bounds, const GpMatrix *matrix, const GpPen *pen)\n");
+		called = 1;
+	}
+	return Ok;
 }
 
 /* MonoTODO */
 GpStatus 
 GdipIsVisiblePathPoint (GpPath *path, float x, float y, GpGraphics *graphics, bool *result)
 {
-        return Ok;
+	static int called = 0;
+
+	if (!called) {
+		printf("NOT IMPLEMENTED: GdipIsVisiblePathPoint (GpPath *path, float x, float y, GpGraphics *graphics, bool *result)\n");
+		called = 1;
+	}
+	return Ok;
 }
 
 /* MonoTODO */
 GpStatus 
 GdipIsVisiblePathPointI (GpPath *path, int x, int y, GpGraphics *graphics, bool *result)
 {
-        return Ok;
+	static int called = 0;
+
+	if (!called) {
+		printf("NOT IMPLEMENTED: GdipIsVisiblePathPointI (GpPath *path, int x, int y, GpGraphics *graphics, bool *result)\n");
+		called = 1;
+	}
+	return Ok;
 }
 
 /* MonoTODO */
 GpStatus 
 GdipIsOutlineVisiblePathPoint (GpPath *path, float x, float y, GpGraphics *graphics, bool *result)
 {
-        return Ok;
+	static int called = 0;
+
+	if (!called) {
+		printf("NOT IMPLEMENTED: GdipIsOutlineVisiblePathPoint (GpPath *path, float x, float y, GpGraphics *graphics, bool *result)\n");
+		called = 1;
+	}
+	return Ok;
 }
 
 /* MonoTODO */
 GpStatus 
 GdipIsOutlineVisiblePathPointI (GpPath *path, int x, int y, GpGraphics *graphics, bool *result)
 {
-        return Ok;
+	static int called = 0;
+
+	if (!called) {
+		printf("NOT IMPLEMENTED: GdipIsOutlineVisiblePathPointI (GpPath *path, int x, int y, GpGraphics *graphics, bool *result)\n");
+		called = 1;
+	}
+	return Ok;
 }
