@@ -29,11 +29,12 @@
 #include <stdio.h>
 #include "gdipImage.h"
 #include <math.h>
-#include "gdip_win32.h"
+
 
 static char *guid_to_string_hack (GDIPCONST CLSID *clsid);
 void gdip_FlipX (GpImage *image);
 void gdip_rotate_180_FlipX (GpImage *image);
+cairo_filter_t gdip_get_cairo_filter (InterpolationMode imode);
 
 /*
  * format guids
@@ -55,11 +56,10 @@ GUID GdipEncoderQuality = {0x1d5be4b5U, 0xfa4aU, 0x452dU, {0x9c, 0xdd, 0x5d, 0xb
 GUID GdipEncoderCompression = {0xe09d739dU, 0xccd4U, 0x44eeU, {0x8e, 0xba, 0x3f, 0xbf, 0x8b, 0xe4, 0xfc, 0x58}};
 
 void 
-gdip_image_init(GpImage *image)
+gdip_image_init (GpImage *image)
 {
 	image->type = imageUndefined;
 	image->surface = 0;
-	image->graphics = 0;
 	image->imageFlags = 0;
 	image->height = 0;
 	image->width = 0;
@@ -73,67 +73,48 @@ gdip_image_init(GpImage *image)
 	image->frameDimensionList = NULL;
 } 
 
-void *
-gdip_image_create_Win32_HDC (GpImage *image)
-{
-	void *result = 0;
-	switch (image->type) {
-		case imageBitmap:
-			result = gdip_bitmap_create_Win32_HDC ((GpBitmap *) image);
-			break;
-		case imageMetafile:
-			break;
-		case imageUndefined:
-			break;
-		default:
-			break;
-	}
-	return result;
-}
-
-void 
-gdip_image_destroy_Win32_HDC (GpImage *image, void *hdc)
-{
-	if (!image || !hdc)
-		return;
-	
-	switch (image->type) {
-		case imageBitmap:
-			gdip_bitmap_destroy_Win32_HDC ((GpBitmap *) image, hdc);
-			break;
-		case imageMetafile:
-			break;
-		case imageUndefined:
-			break;
-		default:
-			break;
-	}
-}
 
 GpStatus 
 GdipDisposeImage (GpImage *image)
 {
+	int i = 0, j = 0, count = 0, dataCount = 0;
+	BitmapData *data;
+
 	if (!image)
 		return InvalidParameter;
 
-	switch (image->type){
-		case imageBitmap:
-			gdip_bitmap_dispose ((GpBitmap *) image);
-			break;
-		case imageMetafile:
-                	break;
-		case imageUndefined:
-			break;
-		default:
-			break;
-	}
-	cairo_surface_destroy (image->surface);
-	image->surface = 0;
-	if (image->frameDimensionList != NULL)
+	if (image->surface)
+		cairo_surface_destroy (image->surface);
+	image->surface = NULL;
+
+	count = image->frameDimensionCount;
+
+	if (count > 0 && image->frameDimensionList != NULL) {
+		for (i = 0; i < count; i++) {
+			dataCount = image->frameDimensionList [i].count;
+			data = image->frameDimensionList [i].frames;
+				for (j = 0; j < dataCount; j++) {
+					if (data [j].Scan0) {
+						GdipFree (data [j].Scan0);
+						data [j].Scan0 = NULL;
+					}
+					if ((data [j].ByteCount) > 0 && (data [j].Bytes != NULL)) {
+						GdipFree (data [j].Bytes);
+						data [j].ByteCount = 0;
+						data [j].Bytes = NULL;
+					}
+				}
+		}
 		GdipFree (image->frameDimensionList);
-	GdipFree (image);
+	}
 	
-	return Ok;
+	/* Nothing more to be done here... We have already
+	 * cleaned the memory while looping in FrameDimension List
+	 * and hence we dont need to do anything in gdip_bitmap_dispose ()
+	 */
+	 GdipFree (image);
+
+	 return Ok;
 }
 
 /*
@@ -143,18 +124,19 @@ GdipDisposeImage (GpImage *image)
 GpStatus 
 GdipGetImageGraphicsContext (GpImage *image, GpGraphics **graphics)
 {
+	GpGraphics *gfx;
 	if (!image || !graphics)
 		return InvalidParameter;
 
-	if (image->graphics == 0) {
-		image->graphics = gdip_graphics_new ();
-		if (image->type == imageBitmap) {
-			gdip_graphics_attach_bitmap (image->graphics, (GpBitmap *) image);
-		}
-		else if (image->type == imageMetafile) {
-		}
+	gfx = gdip_graphics_new ();
+	if (image->type == imageBitmap) {
+		gdip_graphics_attach_bitmap (gfx, (GpBitmap *) image);
 	}
-	*graphics = image->graphics;
+	else if (image->type == imageMetafile) {
+	}
+
+	*graphics = gfx;
+
 	return Ok;
 }
 
@@ -183,27 +165,21 @@ GdipDrawImageRect (GpGraphics *graphics, GpImage *image, float x, float y, float
 	g_return_val_if_fail (image != NULL, InvalidParameter);
 	g_return_val_if_fail (image->type == imageBitmap, InvalidParameter);
 	
-	/* Always to re-attach the image */	
-	if (image->surface)  {
-		cairo_surface_destroy (image->surface);
-		image->surface = NULL;
-	}
+	cairo_new_path(graphics->ct);
 
-	/* Create a surface for this bitmap if one doesn't exist */    
+	/* Create a surface for this bitmap if one doesn't exist */    	
 	gdip_bitmap_ensure_surface ((GpBitmap*) image);
-	
-	gdip_cairo_set_surface_pattern (graphics->ct, image->surface);
+	cairo_surface_set_filter (((GpBitmap*) image)->image.surface, gdip_get_cairo_filter (graphics->interpolation));
+
 	cairo_translate (graphics->ct, x, y);
-	cairo_rectangle (graphics->ct, 0, 0, width, height);
-	
+
 	if (width != image->width || height != image->height) {
 		cairo_scale (graphics->ct,
 				(double) width / image->width,
 				(double) height / image->height);
-		cairo_fill (graphics->ct);
-	} else 
-		cairo_fill (graphics->ct);
-
+	}
+	
+	cairo_show_surface (graphics->ct, image->surface, image->width, image->height);
 	cairo_default_matrix (graphics->ct);
 
 	return Ok;
@@ -226,7 +202,7 @@ GdipDrawImagePoints (GpGraphics *graphics, GpImage *image, GDIPCONST GpPointF *d
 	height = dstPoints[2].Y - dstPoints[2].Y;
 	
 	return GdipDrawImageRect (graphics, image, x, y, width, height);
-	}
+}
 
 GpStatus
 GdipDrawImagePointsI (GpGraphics *graphics, GpImage *image, GDIPCONST GpPoint *dstPoints, int count)
@@ -280,7 +256,7 @@ GdipDrawImageRectRect (GpGraphics *graphics, GpImage *image,
 	g_return_val_if_fail (graphics != NULL, InvalidParameter);
 	g_return_val_if_fail (image != NULL, InvalidParameter);
 	g_return_val_if_fail (image->type == imageBitmap, InvalidParameter);
-    
+
 	if (srcUnit != UnitPixel && srcUnit != UnitWorld) {
 		gdip_unitConversion(srcUnit, UnitPixel, dstx, &dstx);
 		gdip_unitConversion(srcUnit, UnitPixel, dsty, &dsty);
@@ -301,8 +277,8 @@ GdipDrawImageRectRect (GpGraphics *graphics, GpImage *image,
 	if (allocated) 				
 		bitmap->data.Scan0 = dest;
 	
-	/* Always to re-attach the image */	
-	if (bitmap->image.surface)  {
+	/* Re-attach the image if the image has attribues */
+	if ((bitmap->image.surface && imageAttributes != NULL) || bitmap->image.surface == NULL)  {
 		cairo_surface_destroy (bitmap->image.surface);
 		bitmap->image.surface = NULL;
 	}
@@ -339,7 +315,8 @@ GdipDrawImageRectRect (GpGraphics *graphics, GpImage *image,
 		}
 		
 		/* Create a surface for this bitmap if one doesn't exist */
-		gdip_bitmap_ensure_surface ((GpBitmap*) image);			
+		gdip_bitmap_ensure_surface ((GpBitmap*) image);
+		cairo_surface_set_filter (((GpBitmap*) image)->image.surface, gdip_get_cairo_filter (graphics->interpolation));
 
 		for (posy = 0; posy < dstheight; posy += img_height) {
 		
@@ -392,6 +369,7 @@ GdipDrawImageRectRect (GpGraphics *graphics, GpImage *image,
 		
 		/* Create a surface for this bitmap if one doesn't exist */
 		gdip_bitmap_ensure_surface ((GpBitmap*) image);
+		cairo_surface_set_filter (((GpBitmap*) image)->image.surface, gdip_get_cairo_filter (graphics->interpolation));
 		
 		mat = cairo_matrix_create ();
 		cairo_matrix_scale (mat, srcwidth / dstwidth, srcheight / dstheight);
@@ -409,6 +387,12 @@ GdipDrawImageRectRect (GpGraphics *graphics, GpImage *image,
 		/* unref */
 		cairo_matrix_destroy (mat);
 	}
+
+	/* The default surface is no longer valid*/
+	if (bitmap->image.surface && imageAttributes != NULL)  {
+		cairo_surface_destroy (bitmap->image.surface);
+		bitmap->image.surface = NULL;
+	}
 	
 	if (allocated) {
 		bitmap->data.Scan0 = org;
@@ -476,7 +460,7 @@ GdipLoadImageFromFile (GDIPCONST WCHAR *file, GpImage **image)
 	GpImage *result = 0;
 	GpStatus status = 0;
 	ImageFormat format;
-	unsigned char *file_name = NULL;
+	char *file_name = NULL;
 	char format_peek[10];
 	int format_peek_sz;
 	
@@ -485,7 +469,7 @@ GdipLoadImageFromFile (GDIPCONST WCHAR *file, GpImage **image)
 	if (!image || !file)
 		return InvalidParameter;
 	
-	file_name = (unsigned char *) g_utf16_to_utf8 ((const gunichar2 *)file, -1, NULL, NULL, &err);
+	file_name = (char *) g_utf16_to_utf8 ((const gunichar2 *)file, -1, NULL, NULL, &err);
 	if (file_name == NULL || err != NULL) {
 		*image = NULL;
 		return InvalidParameter;
@@ -539,11 +523,12 @@ GdipLoadImageFromFile (GDIPCONST WCHAR *file, GpImage **image)
 	if (status != Ok) {
 		*image = NULL;
 	} else {
-		if (result->frameDimensionCount == 0){
+		if (result->frameDimensionCount == 0) {
 			result->frameDimensionCount = 1;
-			result->frameDimensionList = (FrameDimensionInfo *) GdipAlloc (sizeof (FrameDimensionInfo));
+			result->frameDimensionList = (FrameInfo *) GdipAlloc (sizeof (FrameInfo));
 			result->frameDimensionList[0].count = 1; /*multiple frames are already taken care of in respectic codecs*/
 			memcpy (&(result->frameDimensionList[0].frameDimension), &gdip_image_frameDimension_page_guid, sizeof (CLSID));
+			result->frameDimensionList[0].frames = &(((GpBitmap *) result)->data);
 		}
 		*image = result;
 	}
@@ -582,7 +567,7 @@ GdipSaveImageToFile (GpImage *image, GDIPCONST WCHAR *file, GDIPCONST CLSID *enc
 {
 	FILE *fp = 0;
 	GpStatus status = 0;
-	unsigned char *file_name;
+	char *file_name;
 	GError *err = NULL;
 	ImageFormat format;
 	
@@ -596,7 +581,7 @@ GdipSaveImageToFile (GpImage *image, GDIPCONST WCHAR *file, GDIPCONST CLSID *enc
 	if (format == INVALID)
 		return UnknownImageFormat;
 	
-	file_name = (unsigned char *) g_utf16_to_utf8 ((const gunichar2 *)file, -1, NULL, NULL, &err);
+	file_name = (char *) g_utf16_to_utf8 ((const gunichar2 *)file, -1, NULL, NULL, &err);
 	if (file_name == NULL || err != NULL) 
 		return InvalidParameter;
 	
@@ -654,17 +639,10 @@ GdipGetImageDimension (GpImage *image, float *width, float *height)
 {
 	if (!image || !width || !height)
 		return InvalidParameter;
-        switch (image->type){
-		case imageBitmap:
-			gdip_bitmap_dispose ((GpBitmap *) image);
-			break;
-		case imageMetafile:
-			break;
-		case imageUndefined:
-			break;
-		default:
-			break;
-	}
+
+	*width = image->width;
+	*height = image->height;
+
 	return Ok;
 }
 
@@ -788,11 +766,15 @@ GdipImageGetFrameDimensionsCount (GpImage *image, UINT *count)
 GpStatus 
 GdipImageGetFrameDimensionsList (GpImage *image, GUID *dimensionGUID, UINT count)
 {
-	if (!image || !dimensionGUID || !count)
-		return InvalidParameter;
 	int i;
-	int countReturn = image->frameDimensionCount;
+	int countReturn;
 	GUID guid [count];
+
+	if (!image || !dimensionGUID)
+		return InvalidParameter;
+
+	countReturn = image->frameDimensionCount;
+	
 	if (countReturn < count)
 		countReturn = count;
 	for (i=0; i<countReturn; i++){
@@ -806,9 +788,11 @@ GdipImageGetFrameDimensionsList (GpImage *image, GUID *dimensionGUID, UINT count
 GpStatus
 GdipImageGetFrameCount(GpImage *image, GDIPCONST GUID *dimensionGUID, UINT* count)
 {
+	int i;
+
 	if (!image || !dimensionGUID)
 		return InvalidParameter;
-	int i;
+
 	for (i=0; i<image->frameDimensionCount; i++){
 		if (memcmp(dimensionGUID, &(image->frameDimensionList[i].frameDimension), sizeof(CLSID)) == 0) {
 			*count = image->frameDimensionList[i].count;
@@ -822,13 +806,27 @@ GdipImageGetFrameCount(GpImage *image, GDIPCONST GUID *dimensionGUID, UINT* coun
 GpStatus
 GdipImageSelectActiveFrame(GpImage *image, GDIPCONST GUID *dimensionGUID, UINT index)
 {
-        if (!image || !dimensionGUID)
+	int i=0;
+
+        if (!image || !dimensionGUID || (index <0))
                 return InvalidParameter;
-
-        return NotImplemented;        
+		
+	for (i=0; i<image->frameDimensionCount; i++) {
+		if (memcmp(dimensionGUID, &(image->frameDimensionList[i].frameDimension), sizeof(CLSID)) == 0) {
+			if (image->frameDimensionList[i].count >= index){
+				GpBitmap *bitmap = (GpBitmap *) image;
+				bitmap->data = image->frameDimensionList[i].frames[index];
+				/*printf ("\n image.c selected frame is %d returning ok", index);*/
+				return Ok;
+			} else {
+				return InvalidParameter;
+				/*printf ("\n image.c not selected frame is %d and returning error", index);*/
+			}
+		}
+	}
+        return InvalidParameter;        
 }
 
-
 void copy_pixel (BYTE *src, BYTE *trg, int components)
 {
 	memcpy (trg, src, components);
@@ -861,27 +859,58 @@ void
 gdip_rotate_90 (GpImage *image)
 {
 	BYTE *src, *trg, *rotated;
-	int stride, height, line, x;	
+	int line, x, stride_trg, height_trg, width_trg;
 	GpBitmap *bitmap = (GpBitmap *) image;
-	int components = gdip_get_pixel_format_components (bitmap->data.PixelFormat);		         
-	 
-	bitmap = (GpBitmap *) image;	
-        stride = bitmap->data.Stride;
-	height = bitmap->data.Height;
-	rotated = malloc (stride * height);	
+	int components = gdip_get_pixel_format_components (bitmap->data.PixelFormat);		         	 
+
+	height_trg = bitmap->data.Width;
+	width_trg = bitmap->data.Height;
+	stride_trg = components * gdip_get_pixel_format_depth (bitmap->data.PixelFormat);
+	stride_trg = (stride_trg * width_trg) / 8;
+	stride_trg = (stride_trg + 3) & ~3;		
+	
+	rotated = malloc (stride_trg * height_trg);
 	src = (BYTE *) bitmap->data.Scan0;
 	
-	for (line = 0; line < height; line++, src += stride) {		
-		for (x = 0; x < bitmap->data.Width; x++) {		
-			trg = rotated;	
-			trg += x * stride;
-			trg += (bitmap->data.Width - 1  - line) *components;		
-			copy_pixel (src + (x*components), trg, components);			
+	for (line = 0; line < bitmap->data.Height; line++, src += bitmap->data.Stride) {		
+		for (x = 0; x < bitmap->data.Width; x++) {
+
+			if (x >= height_trg)
+				continue;			
+
+			if (line >= width_trg)
+				continue;
+
+			trg = rotated;
+			trg += x * stride_trg;
+			trg += (width_trg - 1  - line) * components;		
+			copy_pixel (src + (x * components), trg, components);			
 		}	
-	}	
-	
-	memcpy (bitmap->data.Scan0, rotated, stride * height);	
-	free (rotated);
+	}
+
+	bitmap->data.Stride = stride_trg;
+	bitmap->data.Height = image->height = height_trg;
+	bitmap->data.Width = image->width = width_trg;
+
+	if ((bitmap->data.Reserved & GBD_OWN_SCAN0) != 0) {
+		int k,j; 
+
+		/* Do the free first since the frame point can be for some codes a reference to bitmap->data*/
+		free (bitmap->data.Scan0);
+		for (j = 0; j < image->frameDimensionCount; j++) {
+			for (k = 0; k < image->frameDimensionList [j].count; k++) {
+				if (image->frameDimensionList[j].frames[k].Scan0 == bitmap->data.Scan0)  {
+					image->frameDimensionList[j].frames[k].Scan0 = rotated;
+					image->frameDimensionList[j].frames[k].Stride = stride_trg;
+					image->frameDimensionList[j].frames[k].Height = image->height = height_trg;
+					image->frameDimensionList[j].frames[k].Width = image->width = width_trg;
+				}				
+			}
+		}		
+	}
+
+	bitmap->data.Scan0 = rotated;
+	bitmap->data.Reserved |= GBD_OWN_SCAN0;	
 }
 
 void
@@ -915,29 +944,59 @@ void
 gdip_rotate_270 (GpImage *image)
 {
 	BYTE *src, *trg, *rotated;
-	int stride, height, line, x;	
+	int line, x, stride_trg, height_trg, width_trg;
 	GpBitmap *bitmap = (GpBitmap *) image;
-	int components = gdip_get_pixel_format_components (bitmap->data.PixelFormat);		         
-	 
-	bitmap = (GpBitmap *) image;	
-        stride = bitmap->data.Stride;
-	height = bitmap->data.Height;
-	rotated = malloc (stride * height);
+	int components = gdip_get_pixel_format_components (bitmap->data.PixelFormat);
+
+	height_trg = bitmap->data.Width;
+	width_trg = bitmap->data.Height;
+	stride_trg = components * gdip_get_pixel_format_depth (bitmap->data.PixelFormat);
+	stride_trg = (stride_trg * width_trg) / 8;
+	stride_trg = (stride_trg + 3) & ~3;		
+
+	rotated = malloc (stride_trg * height_trg);
 	
 	src = (BYTE *) bitmap->data.Scan0;
 	
-	for (line = 0; line < height; line++, src += stride) {		
+	for (line = 0; line < bitmap->data.Height; line++, src += bitmap->data.Stride) {		
 		for (x = 0; x < bitmap->data.Width; x++) {
-		
+
+			if (x >= height_trg)
+				continue;			
+
+			if (line >= width_trg)
+				continue;
+
 			trg = rotated;	
-			trg += (height-1-x)  * stride;
+			trg += (height_trg -1 -x) * stride_trg;
 			trg += line * components;	
-			copy_pixel (src + (x*components), trg, components);			
+			copy_pixel (src + (x * components), trg, components);			
 		}	
 	}	
-	
-	memcpy (bitmap->data.Scan0, rotated, stride * height);	
-	free (rotated);
+
+	bitmap->data.Stride = stride_trg;
+	bitmap->data.Height = image->height = height_trg;
+	bitmap->data.Width = image->width = width_trg;
+
+	if ((bitmap->data.Reserved & GBD_OWN_SCAN0) != 0) {
+		int k,j; 
+
+		/* Do the free first since the frame point can be for some codes a reference to bitmap->data*/
+		free (bitmap->data.Scan0);
+		for (j = 0; j < image->frameDimensionCount; j++) {
+			for (k = 0; k < image->frameDimensionList [j].count; k++) {
+				if (image->frameDimensionList[j].frames[k].Scan0 == bitmap->data.Scan0)  {
+					image->frameDimensionList[j].frames[k].Scan0 = rotated;
+					image->frameDimensionList[j].frames[k].Stride = stride_trg;
+					image->frameDimensionList[j].frames[k].Height = image->height = height_trg;
+					image->frameDimensionList[j].frames[k].Width = image->width = width_trg;
+				}				
+			}
+		}		
+	}
+
+	bitmap->data.Scan0 = rotated;
+	bitmap->data.Reserved |= GBD_OWN_SCAN0;
 }
 
 void
@@ -1068,6 +1127,45 @@ GdipSetPropertyItem(GpImage *image, GDIPCONST PropertyItem *item)
 	return NotImplemented; /* GdipSetPropertyItem */
 }
 
+void
+gdip_image_clone (GpImage* image, GpImage* clonedImage)
+{
+	int i = 0, j = 0, count = 0, dataCount = 0;
+	BitmapData *data, *clonedData;
+
+	clonedImage->surface = NULL;
+
+	if (image->frameDimensionCount) {
+		clonedImage->frameDimensionCount = image->frameDimensionCount;
+		clonedImage->frameDimensionList = malloc (sizeof (FrameInfo) * image->frameDimensionCount);
+
+		for (i = 0; i < image->frameDimensionCount; i++) {
+			clonedImage->frameDimensionList[i].count = image->frameDimensionList[i].count;
+			memcpy (&clonedImage->frameDimensionList[i].frameDimension, 
+				&image->frameDimensionList[i].frameDimension, sizeof (GUID));
+
+			dataCount = image->frameDimensionList[i].count;
+			data = image->frameDimensionList[i].frames;
+			clonedImage->frameDimensionList[i].frames = malloc (sizeof (BitmapData) * dataCount);
+			clonedData = clonedImage->frameDimensionList[i].frames;
+			/* Copy all BitmapData */
+			memcpy (clonedImage->frameDimensionList[i].frames, 
+				image->frameDimensionList[i].frames,  sizeof (BitmapData) * dataCount);
+
+			for (j = 0; j < dataCount; j++) {
+				if (data[j].Scan0) {
+					clonedData[j].Scan0 = malloc (data[j].Stride * data[j].Height);
+					memcpy (clonedData[j].Scan0, data[j].Scan0, data[j].Stride * data[j].Height);				
+				}
+				if ((data[j].ByteCount) > 0 && (data[j].Bytes != NULL)) {				
+					clonedData[j].Bytes = malloc (data[j].ByteCount);
+					memcpy (clonedData[j].Bytes, data[j].Bytes, data[j].ByteCount);
+				}
+			}
+		}
+	}
+}
+
 GpStatus
 GdipCloneImage(GpImage *image, GpImage **cloneImage)
 {
@@ -1077,6 +1175,7 @@ GdipCloneImage(GpImage *image, GpImage **cloneImage)
 	switch (image->type){
 		case imageBitmap:
                 	gdip_bitmap_clone ((GpBitmap *) image, (GpBitmap **) cloneImage);
+			gdip_image_clone (image, *cloneImage);
 			break;
 		case imageMetafile:
         	        return NotImplemented; /* GdipCloneImage - imageMetafile */
@@ -1211,7 +1310,8 @@ gdip_get_pixel_format_components(PixelFormat pixfmt)
 }
 
 GpStatus
-GdipLoadImageFromDelegate_linux (GetBytesDelegate getBytesFunc,
+GdipLoadImageFromDelegate_linux (GetHeaderDelegate getHeaderFunc,
+				 GetBytesDelegate getBytesFunc,
 				 PutBytesDelegate putBytesFunc,
                                  SeekDelegate seekFunc,
 				 CloseDelegate closeFunc,
@@ -1222,11 +1322,11 @@ GdipLoadImageFromDelegate_linux (GetBytesDelegate getBytesFunc,
 	GpStatus status = 0;
 	ImageFormat format;
 	
-	char format_peek[10];
+	byte format_peek[10];
 	int format_peek_sz;
 	
-	format_peek_sz = getBytesFunc (format_peek, 10, 1);
-	format = get_image_format (format_peek, format_peek_sz);
+	format_peek_sz = getHeaderFunc (format_peek, 10);
+	format = get_image_format ((char *)format_peek, format_peek_sz);
 	
 	switch (format) {
 		case JPEG:
@@ -1265,9 +1365,10 @@ GdipLoadImageFromDelegate_linux (GetBytesDelegate getBytesFunc,
 	} else {
 		if (result->frameDimensionCount == 0){
 			result->frameDimensionCount = 1;
-			result->frameDimensionList = (FrameDimensionInfo *) GdipAlloc (sizeof (FrameDimensionInfo));
+			result->frameDimensionList = (FrameInfo *) GdipAlloc (sizeof (FrameInfo));
 			result->frameDimensionList[0].count = 1; /*multiple frames are already taken care of in respectic codecs*/
 			memcpy (&(result->frameDimensionList[0].frameDimension), &gdip_image_frameDimension_page_guid, sizeof (CLSID));
+			result->frameDimensionList[0].frames = &(((GpBitmap *) result)->data);
 		}
 		*image = result;
 	}
@@ -1455,10 +1556,12 @@ GdipGetAllPropertyItems(GpImage *image, UINT totalBufferSize, UINT numProperties
 GpStatus
 GdipGetEncoderParameterListSize(GpImage *image, GDIPCONST CLSID *clsidEncoder, UINT *size)
 {
+	ImageFormat fmt;
+
 	g_return_val_if_fail (clsidEncoder != NULL, InvalidParameter);
 	g_return_val_if_fail (size != NULL, InvalidParameter);
 
-	ImageFormat fmt = gdip_get_imageformat_from_codec_clsid ((CLSID *) clsidEncoder);
+	fmt = gdip_get_imageformat_from_codec_clsid ((CLSID *) clsidEncoder);
 
 	switch (fmt) {
 		case JPEG:
@@ -1477,10 +1580,12 @@ GdipGetEncoderParameterListSize(GpImage *image, GDIPCONST CLSID *clsidEncoder, U
 GpStatus
 GdipGetEncoderParameterList(GpImage *image, GDIPCONST CLSID *clsidEncoder, UINT size, EncoderParameters *buffer)
 {
+	ImageFormat fmt;
+
 	g_return_val_if_fail (clsidEncoder != NULL, InvalidParameter);
 	g_return_val_if_fail (buffer != NULL, InvalidParameter);
 
-	ImageFormat fmt = gdip_get_imageformat_from_codec_clsid ((CLSID *) clsidEncoder);
+	fmt = gdip_get_imageformat_from_codec_clsid ((CLSID *) clsidEncoder);
 
 	switch (fmt) {
 		case JPEG:
