@@ -311,9 +311,9 @@ gdip_load_jpeg_image_internal (struct jpeg_source_mgr *src,
     cinfo.do_block_smoothing = FALSE;
 
     img->image.type = imageBitmap;
-    img->image.graphics = 0;
     img->image.width = cinfo.image_width;
     img->image.height = cinfo.image_height;
+
     if (cinfo.density_unit == 1) { /* dpi */
         img->image.horizontalResolution = cinfo.X_density;
         img->image.verticalResolution = cinfo.Y_density;
@@ -442,10 +442,7 @@ gdip_load_jpeg_image_internal (struct jpeg_source_mgr *src,
 			r = (255 - k) * (255 - y) / 255;
 		    }
 
-		    lineptr [0] = r;
-		    lineptr [1] = g;
-		    lineptr [2] = b;
-		    lineptr [3] = 255;
+		    set_pixel_bgra(lineptr, 0, b, g, r, 0xff);
 		    lineptr += 4;
 		}
 	    }
@@ -456,19 +453,15 @@ gdip_load_jpeg_image_internal (struct jpeg_source_mgr *src,
 		guchar *inptr, *outptr;
 		JOCTET r, g, b;
 
-		inptr = lines[i] + (img->image.width) * 3 - 1;
-		outptr = lines[i] + stride - 1;
+		inptr = lines[i] + (img->image.width) * 3;
+		outptr = lines[i] + stride;
 		for (j = 0; j < img->image.width; j++) {
 		    /* Note the swapping of R and B, to get ARGB from what
 		     * looks like BGR data.
 		     */
-		    r = *inptr--;
-		    g = *inptr--;
-		    b = *inptr--;
-		    *outptr-- = 255;
-		    *outptr-- = b;
-		    *outptr-- = g;
-		    *outptr-- = r;
+		    inptr -= 3;
+		    outptr -= 4;
+		    set_pixel_bgra(outptr, 0, inptr[2], inptr[1], inptr[0], 0xff);
 		}
 	    }
         }
@@ -480,7 +473,7 @@ gdip_load_jpeg_image_internal (struct jpeg_source_mgr *src,
     img->data.Scan0 = destbuf;
     img->data.Reserved = GBD_OWN_SCAN0;
 
-    img->image.surface = cairo_surface_create_for_image (destbuf, img->cairo_format,
+    img->image.surface = cairo_surface_create_for_image ((char *)destbuf, img->cairo_format,
                                                    img->image.width, img->image.height,
                                                    stride);
 
@@ -676,9 +669,15 @@ gdip_save_jpeg_image_internal (FILE *fp,
             outptr = scanline;
 
             for (j = 0; j < bitmap->data.Width; j++) {
+#ifdef WORDS_BIGENDIAN
+                *outptr++ = inptr[1]; /* R */
+                *outptr++ = inptr[2]; /* G */
+                *outptr++ = inptr[3]; /* B */
+#else
                 *outptr++ = inptr[2]; /* R */
                 *outptr++ = inptr[1]; /* G */
                 *outptr++ = inptr[0]; /* B */
+#endif
                 inptr += 4;        /* skip RGB+A */
             }
 
