@@ -40,18 +40,6 @@ static BrushClass vtable = { BrushTypeLinearGradient,
 			     gdip_linear_gradient_clone_brush,
 			     gdip_linear_gradient_destroy };
 
-struct _Blend {
-	float *factors;
-	float *positions;
-	int count;
-};
-
-struct _InterpolationColors {
-	ARGB* colors;
-	float *positions;
-	int count;
-};
- 
 void
 gdip_linear_gradient_init (GpLineGradient *linear)
 {
@@ -70,7 +58,6 @@ gdip_linear_gradient_init (GpLineGradient *linear)
 	linear->blend->positions = (float *) GdipAlloc (sizeof (float));
 	linear->blend->factors [0] = 1.0;
 	linear->blend->positions[0] = 0.0;
-	linear->changed = TRUE;
 	linear->pattern = NULL;
 }
 
@@ -102,12 +89,17 @@ gdip_linear_gradient_clone_brush (GpBrush *brush, GpBrush **clonedBrush)
 	newbrush->base = linear->base;
 	newbrush->wrapMode = linear->wrapMode;
 	GdipCloneMatrix (linear->matrix, &newbrush->matrix);
-	newbrush->rectangle = linear->rectangle;
+	if (linear->rectangle) {
+		newbrush->rectangle = (GpRectF *) GdipAlloc (sizeof (GpRectF));
+		memcpy (newbrush->rectangle, linear->rectangle, sizeof (GpRectF));
+	} else {
+		newbrush->rectangle= NULL;
+	}
 	newbrush->gammaCorrection = linear->gammaCorrection;
 	newbrush->angle = linear->angle;
 	newbrush->isAngleScalable = linear->isAngleScalable;
-	newbrush->changed = linear->changed;
 	/* cloned brush needs to have its own pattern */
+	newbrush->base.changed = TRUE;
 	newbrush->pattern = NULL;
 	newbrush->lineColors [0] = linear->lineColors [0];
 	newbrush->lineColors [1] = linear->lineColors [1];
@@ -119,7 +111,7 @@ gdip_linear_gradient_clone_brush (GpBrush *brush, GpBrush **clonedBrush)
 	if (newbrush->presetColors == NULL) 
 		goto NO_PRESET;
 
-	if (linear->presetColors->count != -1) {
+	if (linear->presetColors->count > 0) {
 		newbrush->presetColors->colors = (ARGB *) GdipAlloc (linear->presetColors->count * sizeof (ARGB));
 		if (newbrush->presetColors->colors == NULL) 
 			goto NO_PRESET_COLORS;
@@ -137,7 +129,7 @@ gdip_linear_gradient_clone_brush (GpBrush *brush, GpBrush **clonedBrush)
 	newbrush->blend = (Blend *) GdipAlloc (sizeof (Blend));
 	if (newbrush->blend == NULL)
 		goto NO_BLEND;
-	if (linear->blend->count != -1) {
+	if (linear->blend->count > 0) {
 		newbrush->blend->factors = (float *) GdipAlloc (linear->blend->count * sizeof (float));
 		if (newbrush->blend->factors == NULL) 
 			goto NO_BLEND_FACTORS;
@@ -699,8 +691,14 @@ gdip_linear_gradient_setup (GpGraphics *graphics, GpBrush *brush)
 	product = cairo_matrix_create ();
 	g_return_val_if_fail (product != NULL, OutOfMemory);
 
-	/* We do setup iff brush is changed or setup has not been done yet. */
-	if (linear->changed || (linear->pattern) == NULL) {
+	/* We create the new pattern for brush, if the brush is changed
+	 * or if pattern has not been created yet.
+	 */
+	if (linear->base.changed || (linear->pattern) == NULL) {
+
+		/* destroy the existing pattern */
+		if (linear->pattern)
+			cairo_pattern_destroy (linear->pattern);
 
 		switch (linear->wrapMode) {
 
@@ -726,21 +724,20 @@ gdip_linear_gradient_setup (GpGraphics *graphics, GpBrush *brush)
 	}
 
 	if (status == Ok) {
-		if (linear->pattern == NULL)
-			return GenericError;
-		/* Use both the matrices */
-		cairo_matrix_multiply (product, linear->matrix, graphics->copy_of_ctm);
-		cairo_matrix_invert (product);
-		cairo_pattern_set_matrix (linear->pattern, product);
-		cairo_pattern_set_extend (linear->pattern, CAIRO_EXTEND_REPEAT);
-		cairo_set_pattern (ct, linear->pattern);
+		if (linear->pattern != NULL) {
+			/* Use both the matrices */
+			cairo_matrix_multiply (product, linear->matrix, graphics->copy_of_ctm);
+			cairo_matrix_invert (product);
+			cairo_pattern_set_matrix (linear->pattern, product);
+			cairo_pattern_set_extend (linear->pattern, CAIRO_EXTEND_REPEAT);
+			cairo_set_pattern (ct, linear->pattern);
 
-		status = gdip_get_status (cairo_status (ct));
+			status = gdip_get_status (cairo_status (ct));
+		}
+		else
+			status = GenericError;
 	}
 
-	if (status == Ok)
-		linear->changed = FALSE;
-
 	cairo_matrix_destroy (product);
 
 	return status;
@@ -1008,7 +1005,7 @@ GdipSetLineBlend (GpLineGradient *brush, GDIPCONST float *blend, GDIPCONST float
 		brush->presetColors->count = 0;
 	}
 
-	brush->changed = TRUE;
+	brush->base.changed = TRUE;
 
 	return Ok;
 }
@@ -1023,6 +1020,13 @@ GdipGetLineBlend (GpLineGradient *brush, float *blend, float *positions, int cou
 	g_return_val_if_fail (blend != NULL, InvalidParameter);
 	g_return_val_if_fail (positions != NULL, InvalidParameter);
 	g_return_val_if_fail (brush->blend->count == count, InvalidParameter);
+
+	/* If count is less than 1, we are not in a proper state 
+	 * to return blend property. By default, we have one blend
+	 * set. Therefore, count of 1 is acceptible.
+	 */
+	if (brush->blend->count < 1)
+		return WrongState;
 	
 	memcpy (blend, brush->blend->factors, count * sizeof (float));
 	memcpy (positions, brush->blend->positions, count * sizeof (float));
@@ -1035,7 +1039,7 @@ GdipSetLineGammaCorrection (GpLineGradient *brush, BOOL useGammaCorrection)
 {
 	g_return_val_if_fail (brush != NULL, InvalidParameter);
 	brush->gammaCorrection = useGammaCorrection;
-	brush->changed = TRUE;
+	brush->base.changed = TRUE;
 
 	return Ok;
 }
@@ -1044,6 +1048,8 @@ GpStatus
 GdipGetLineGammaCorrection (GpLineGradient *brush, BOOL *useGammaCorrection)
 {
 	g_return_val_if_fail (brush != NULL, InvalidParameter);
+	g_return_val_if_fail (useGammaCorrection != NULL, InvalidParameter);
+
 	*useGammaCorrection = brush->gammaCorrection;
 
 	return Ok;
@@ -1053,6 +1059,7 @@ GpStatus
 GdipGetLinePresetBlendCount (GpLineGradient *brush, int *count)
 {
 	g_return_val_if_fail (brush != NULL, InvalidParameter);
+	g_return_val_if_fail (count != NULL, InvalidParameter);
 
 	/* If count is less than 2, we are not in a proper state 
 	 * to return presetblend property.
@@ -1108,7 +1115,7 @@ GdipSetLinePresetBlend (GpLineGradient *brush, GDIPCONST ARGB *blend, GDIPCONST
 		brush->blend->count = 0;
 	}
 
-	brush->changed = TRUE;
+	brush->base.changed = TRUE;
 	return Ok;
 }
 
@@ -1122,6 +1129,12 @@ GdipGetLinePresetBlend (GpLineGradient *brush, ARGB *blend, float *positions, in
 	g_return_val_if_fail (blend != NULL, InvalidParameter);
 	g_return_val_if_fail (positions != NULL, InvalidParameter);
 	g_return_val_if_fail (brush->presetColors->count == count, InvalidParameter);
+
+	/* If count is less than 2, we are not in a proper state 
+	 * to return presetblend property.
+	 */
+	if (brush->presetColors->count < 2)
+		return WrongState;
 	
 	memcpy (blend, brush->presetColors->colors, count * sizeof (ARGB));
 	memcpy (positions, brush->presetColors->positions, count * sizeof (float));
@@ -1136,7 +1149,7 @@ GdipSetLineColors (GpLineGradient *brush, ARGB color1, ARGB color2)
 
 	brush->lineColors[0] = color1;
 	brush->lineColors[1] = color2;
-	brush->changed = TRUE;
+	brush->base.changed = TRUE;
 	return Ok;
 }
 
@@ -1189,7 +1202,7 @@ GdipGetLineTransform (GpLineGradient *brush, GpMatrix *matrix)
 	if (brush->presetColors->count >= 2)
 		return WrongState;
 
-	*matrix = *(brush->matrix);
+	cairo_matrix_copy(matrix, brush->matrix);
 
 	return Ok;
 }
@@ -1200,8 +1213,10 @@ GdipSetLineTransform (GpLineGradient *brush, GDIPCONST GpMatrix *matrix)
 	g_return_val_if_fail (brush != NULL, InvalidParameter);
 	g_return_val_if_fail (matrix != NULL, InvalidParameter);
 
-	*(brush->matrix) = *matrix;
-	brush->changed = TRUE;
+	brush->matrix = cairo_matrix_create();
+	g_return_val_if_fail (brush->matrix != NULL, OutOfMemory);
+	cairo_matrix_copy(brush->matrix, matrix);
+	brush->base.changed = TRUE;
 	return Ok;
 }
 
@@ -1222,19 +1237,19 @@ GdipSetLineWrapMode (GpLineGradient *brush, GpWrapMode wrapMode)
 	g_return_val_if_fail (wrapMode != WrapModeClamp, InvalidParameter);
 
 	brush->wrapMode = wrapMode;
-	brush->changed = TRUE;
+	brush->base.changed = TRUE;
 	return Ok;
 }
 
 GpStatus
 GdipSetLineLinearBlend (GpLineGradient *brush, float focus, float scale)
 {
-	g_return_val_if_fail (brush != NULL, InvalidParameter);
-
 	float *blends;
 	float *positions;
 	int count = 3;
 
+	g_return_val_if_fail (brush != NULL, InvalidParameter);
+
 	if (focus == 0 || focus == 1) {
 		count = 2;
 	}
@@ -1256,6 +1271,14 @@ GdipSetLineLinearBlend (GpLineGradient *brush, float focus, float scale)
 		brush->blend->positions = positions;
 	}
 
+	/* we clear the preset colors when setting the blend */
+	if (brush->presetColors->count != 0) {
+		GdipFree (brush->presetColors->colors);
+		GdipFree (brush->presetColors->positions);
+		brush->presetColors->count = 0;
+	}
+
+	/* set the blend colors */
 	if (focus == 0) {
 		brush->blend->positions [0] = focus;
 		brush->blend->factors [0] = scale;
@@ -1280,6 +1303,57 @@ GdipSetLineLinearBlend (GpLineGradient *brush, float focus, float scale)
 	}
 
 	brush->blend->count = count;
+	brush->base.changed = TRUE;
+
+	return Ok;
+}
+
+/* This is defined in general.c */
+float gdip_erf (float x, float std, float mean);
+
+GpStatus
+GdipSetLineSigmaBlend (GpLineGradient *brush, float focus, float scale)
+{
+	float *blends;
+	float *positions;
+	float pos = 0.0;
+	int count = 511; /* total no of samples */
+	int index;
+	float sigma;
+	float mean;
+	float fall_off_len = 2.0; /* curve fall off length in terms of SIGMA */
+	float delta; /* distance between two samples */
+
+	/* we get a curve not starting from 0 and not ending at 1.
+	 * so we subtract the starting value and divide by the curve
+	 * height to make it fit in the 0 to scale range
+	 */
+	float curve_bottom;
+	float curve_top;
+	float curve_height;
+
+	g_return_val_if_fail (brush != NULL, InvalidParameter);
+
+	if (focus == 0 || focus == 1) {
+		count = 256;
+	}
+
+	if (brush->blend->count != count) {
+		blends = (float *) GdipAlloc (count * sizeof (float));
+		g_return_val_if_fail (blends != NULL, OutOfMemory);
+
+		positions = (float *) GdipAlloc (count * sizeof (float));
+		g_return_val_if_fail (positions != NULL, OutOfMemory);
+
+		/* free the existing values */
+		if (brush->blend->count != 0) {
+			GdipFree (brush->blend->factors);
+			GdipFree (brush->blend->positions);
+		}
+
+		brush->blend->factors = blends;
+		brush->blend->positions = positions;
+	}
 
 	/* we clear the preset colors when setting the blend */
 	if (brush->presetColors->count != 0) {
@@ -1288,17 +1362,132 @@ GdipSetLineLinearBlend (GpLineGradient *brush, float focus, float scale)
 		brush->presetColors->count = 0;
 	}
 
-	brush->changed = TRUE;
+	/* Set the blend colors. We use integral of the Normal Distribution,
+	 * i.e. Cumulative Distribution Function (CFD).
+	 *
+	 * Normal distribution:
+	 *
+	 * y (x) = (1 / sqrt (2 * PI * sq (sigma))) * exp (-sq (x - mu)/ (2 * sq (sigma)))
+	 *
+	 * where, y = height of normal curve, 
+	 *        sigma = standard deviation
+	 *        mu = mean
+	 * OR
+	 * y (x) = peak * exp ( - z * z / 2)
+	 * where, z = (x - mu) / sigma
+	 *
+	 * In this curve, peak would occur at mean i.e. for x = mu. This results in
+	 * a peak value of peak = (1 / sqrt (2 * PI * sq (sigma))).
+	 *
+	 * Cumulative distribution function:
+	 * Ref: http://mathworld.wolfram.com/NormalDistribution.html
+	 *
+	 * D (x) = (1 / 2) [1 + erf (z)]
+	 * where, z = (x - mu) / (sigma * sqrt (2))
+	 *
+	 */
+	if (focus == 0) {
+		/* right part of the curve with a complete fall in fall_off_len * SIGMAs */
+		sigma = 1.0 / fall_off_len;
+		mean = 0.5;
+		delta = 1.0 / 255.0;
+
+		curve_bottom = 0.5 * (1.0 - gdip_erf (1.0, sigma, mean));
+		curve_top = 0.5 * (1.0 - gdip_erf (focus, sigma, mean));
+		curve_height = curve_top - curve_bottom;
+
+		/* set the start */
+		brush->blend->positions [0] = focus;
+		brush->blend->factors [0] = scale;
+
+		for (index = 1, pos = delta; index < 255; index++, pos += delta) {
+			brush->blend->positions [index] = pos;
+			brush->blend->factors [index] = (scale / curve_height) * 
+				(0.5 * (1.0 - gdip_erf (pos, sigma, mean)) - curve_bottom);
+		}
+
+		/* set the end */
+		brush->blend->positions [count - 1] = 1.0;
+		brush->blend->factors [count - 1] = 0.0;
+	}
+
+	else if (focus == 1) {
+		/* left part of the curve with a complete rise in fall_off_len * SIGMAs */
+		sigma = 1.0 / fall_off_len;
+		mean = 0.5;
+		delta = 1.0 / 255.0;
+
+		curve_bottom = 0.5 * (1.0 + gdip_erf (0.0, sigma, mean));
+		curve_top = 0.5 * (1.0 + gdip_erf (focus, sigma, mean));
+		curve_height = curve_top - curve_bottom;
+
+		/* set the start */
+		brush->blend->positions [0] = 0.0;
+		brush->blend->factors [0] = 0.0;
+
+		for (index = 1, pos = delta; index < 255; index++, pos += delta) {
+			brush->blend->positions [index] = pos;
+			brush->blend->factors [index] = (scale / curve_height) * 
+				(0.5 * (1.0 + gdip_erf (pos, sigma, mean)) - curve_bottom);
+		}
+
+		/* set the end */
+		brush->blend->positions [count - 1] = focus;
+		brush->blend->factors [count - 1] = scale;
+	}
+
+	else {
+		/* left part of the curve with a complete fall in fall_off_len * SIGMAs */
+		sigma = focus / (2 * fall_off_len);
+		mean = focus / 2.0;
+		delta = focus / 255.0;
+
+		/* set the start */
+		brush->blend->positions [0] = 0.0;
+		brush->blend->factors [0] = 0.0;
+
+		curve_bottom = 0.5 * (1.0 + gdip_erf (0.0, sigma, mean));
+		curve_top = 0.5 * (1.0 + gdip_erf (focus, sigma, mean));
+		curve_height = curve_top - curve_bottom;
+
+		for (index = 1, pos = delta; index < 255; index++, pos += delta) {
+			brush->blend->positions [index] = pos;
+			brush->blend->factors [index] = (scale / curve_height) * 
+				(0.5 * (1.0 + gdip_erf (pos, sigma, mean)) - curve_bottom);
+		}
+
+		brush->blend->positions [index] = focus;
+		brush->blend->factors [index] = scale;
+
+		/* right part of the curve with a complete fall in fall_off_len * SIGMAs */
+		sigma = (1.0 - focus) / (2 * fall_off_len);
+		mean = (1.0 + focus) / 2.0;
+		delta = (1.0 - focus) / 255.0;
+
+		curve_bottom = 0.5 * (1.0 - gdip_erf (1.0, sigma, mean));
+		curve_top = 0.5 * (1.0 - gdip_erf (focus, sigma, mean));
+		curve_height = curve_top - curve_bottom;
+
+		index ++;
+		pos = focus + delta;
+
+		for (; index < 510; index++, pos += delta) {
+			brush->blend->positions [index] = pos;
+			brush->blend->factors [index] = (scale / curve_height) * 
+				(0.5 * (1.0 - gdip_erf (pos, sigma, mean)) - curve_bottom);
+		}
+
+		/* set the end */
+		brush->blend->positions [count - 1] = 1.0;
+		brush->blend->factors [count - 1] = 0.0;
+	}
+
+	brush->blend->count = count;
+	brush->base.changed = TRUE;
 
 	return Ok;
 }
 
-GpStatus
-GdipSetLineSigmaBlend (GpLineGradient *brush, float focus, float scale)
-{
-	return NotImplemented;
-}
-
 GpStatus
 GdipMultiplyLineTransform (GpLineGradient *brush, GpMatrix *matrix, GpMatrixOrder order)
 {
@@ -1308,7 +1497,7 @@ GdipMultiplyLineTransform (GpLineGradient *brush, GpMatrix *matrix, GpMatrixOrde
 	g_return_val_if_fail (matrix != NULL, InvalidParameter);
 
 	if ((status = GdipMultiplyMatrix (brush->matrix, matrix, order)) == Ok)
-		brush->changed = TRUE;
+		brush->base.changed = TRUE;
 	return status;
 }
 
@@ -1322,7 +1511,7 @@ GdipResetLineTransform (GpLineGradient *brush)
 	status = cairo_matrix_set_identity (brush->matrix);
 	s = gdip_get_status (status);
 	if (s == Ok)
-		brush->changed = TRUE;
+		brush->base.changed = TRUE;
 	return s;
 }
 
@@ -1340,7 +1529,7 @@ GdipRotateLineTransform (GpLineGradient *brush, float angle, GpMatrixOrder order
 	if ((status = GdipTranslateMatrix (brush->matrix, rect->Width, rect->Height, order)) == Ok)
 		if ((status = GdipRotateMatrix (brush->matrix, angle, order)) == Ok)
 			if ((status = GdipTranslateMatrix (brush->matrix, -rect->Width, -rect->Height, order)) == Ok)
-				brush->changed = TRUE;
+				brush->base.changed = TRUE;
 	return status;
 }
 
@@ -1351,7 +1540,7 @@ GdipScaleLineTransform (GpLineGradient *brush, float sx, float sy, GpMatrixOrder
 	g_return_val_if_fail (brush != NULL, InvalidParameter);
 
 	if ((status = GdipScaleMatrix (brush->matrix, sx, sy, order)) == Ok)
-		brush->changed = TRUE;
+		brush->base.changed = TRUE;
 	return status;
 }
 
@@ -1362,6 +1551,6 @@ GdipTranslateLineTransform (GpLineGradient *brush, float dx, float dy, GpMatrixO
 	g_return_val_if_fail (brush != NULL, InvalidParameter);
 
 	if ((status = GdipTranslateMatrix (brush->matrix, dx, dy, order)) == Ok)
-		brush->changed = TRUE;
+		brush->base.changed = TRUE;
 	return status;
 }
