@@ -170,6 +170,10 @@ gdip_load_png_image_from_file_or_stream (FILE *fp,
 
         stride = width * 4;
 
+        while (stride % sizeof(pixman_bits_t))
+                stride++;
+
+
         row_pointers = png_get_rows (png_ptr, info_ptr);
         rawdata = GdipAlloc (stride * height);
         rawptr = rawdata;
@@ -178,8 +182,12 @@ gdip_load_png_image_from_file_or_stream (FILE *fp,
 
 	case 4:
 		for (i = 0; i < height; i++) {
-			memcpy (rawptr, row_pointers[i], stride);
-			rawptr += stride;
+			png_bytep rowp = row_pointers[i];
+			for (j = 0; j < width; j++) {
+				set_pixel_bgra (rawptr, 0, rowp[0], rowp[1], rowp[2], rowp[3]);
+				rowp += 4;
+				rawptr += 4;
+			}
 		}
 		break;
 
@@ -187,10 +195,9 @@ gdip_load_png_image_from_file_or_stream (FILE *fp,
 		for (i = 0; i < height; i++) {
 			png_bytep rowp = row_pointers[i];
 			for (j = 0; j < width; j++) {
-				*rawptr++ = *rowp++;
-				*rawptr++ = *rowp++;
-				*rawptr++ = *rowp++;
-				*rawptr++ = 255; /* a */
+				set_pixel_bgra (rawptr, 0, rowp[0], rowp[1], rowp[2], 0xff);
+				rowp += 3;
+				rawptr += 4;
 			}
 		}
 		break;
@@ -200,10 +207,8 @@ gdip_load_png_image_from_file_or_stream (FILE *fp,
 			png_bytep rowp = row_pointers[i];
 			for (j = 0; j < width; j++) {
 				png_byte pix = *rowp++;
-				*rawptr++ = pix;
-				*rawptr++ = pix;
-				*rawptr++ = pix;
-				*rawptr++ = 255;
+				set_pixel_bgra (rawptr, 0, pix, pix, pix, 0xff);
+				rawptr += 4;
 			}
 		}
 		break;
@@ -213,7 +218,6 @@ gdip_load_png_image_from_file_or_stream (FILE *fp,
 
         img = gdip_bitmap_new ();
         img->image.type = imageBitmap;
-        img->image.graphics = 0;
         img->image.width = width;
         img->image.height = height;
 
@@ -225,7 +229,7 @@ gdip_load_png_image_from_file_or_stream (FILE *fp,
         img->data.Scan0 = rawdata;
         img->data.Reserved = GBD_OWN_SCAN0;
 
-        img->image.surface = cairo_surface_create_for_image (rawdata,
+        img->image.surface = cairo_surface_create_for_image ((char *)rawdata,
                                                              img->cairo_format,
                                                              img->image.width,
                                                              img->image.height,
@@ -253,6 +257,7 @@ gdip_load_png_image_from_file_or_stream (FILE *fp,
         GdipFree (pngsrc);
 
     *image = (GpImage *) img;
+
     return Ok;
 
   error:
@@ -298,13 +303,11 @@ gdip_save_png_image_to_file_or_stream (FILE *fp,
     GpBitmap *bitmap = (GpBitmap *) image;
     gdip_stream_png_source_ptr pngsrc = NULL;
 
-    int i;
+    int i, j;
 
     int bit_depth;
     int color_type;
 
-    png_bytep *row_pointers = NULL;
-
     png_ptr = png_create_write_struct (PNG_LIBPNG_VER_STRING,
                                        NULL, NULL, NULL);
     if (!png_ptr)
@@ -391,11 +394,6 @@ gdip_save_png_image_to_file_or_stream (FILE *fp,
                   PNG_COMPRESSION_TYPE_DEFAULT,
                   PNG_FILTER_TYPE_DEFAULT);
 
-    row_pointers = GdipAlloc(sizeof (png_bytep) * image->height);
-
-    for (i = 0; i < image->height; i++)
-        row_pointers[i] = bitmap->data.Scan0 + (bitmap->data.Stride * i);
-
     png_write_info (png_ptr, info_ptr);
 
     if (image->pixFormat != bitmap->data.PixelFormat) {
@@ -411,13 +409,31 @@ gdip_save_png_image_to_file_or_stream (FILE *fp,
 
     png_set_bgr(png_ptr);
 
-    png_write_rows (png_ptr, row_pointers, image->height);
+#ifdef WORDS_BIGENDIAN
+    {
+	    guchar *row_pointer = GdipAlloc (image->width * 4);
+
+	    for (i = 0; i < image->height; i++) {
+		for (j = 0; j < image->width; j++) {
+			row_pointer[j*4] = *((guchar *)bitmap->data.Scan0 + (bitmap->data.Stride * i) + (j*4) + 3);
+			row_pointer[j*4+1] = *((guchar *)bitmap->data.Scan0 + (bitmap->data.Stride * i) + (j*4) + 2);
+			row_pointer[j*4+2] = *((guchar *)bitmap->data.Scan0 + (bitmap->data.Stride * i) + (j*4) + 1);
+			row_pointer[j*4+3] = *((guchar *)bitmap->data.Scan0 + (bitmap->data.Stride * i) + (j*4) + 0);
+		}
+		png_write_row (png_ptr, row_pointer);
+	    }
+	    GdipFree (row_pointer);
+    }
+#else
+    for (i = 0; i < image->height; i++) {
+    	png_write_row (png_ptr, bitmap->data.Scan0 + (bitmap->data.Stride * i));
+    }
+#endif
+
     png_write_end (png_ptr, NULL);
 
     png_destroy_write_struct (&png_ptr, &info_ptr);
 
-    GdipFree (row_pointers);
-
     if (pngsrc)
         GdipFree (pngsrc);
 
@@ -425,8 +441,6 @@ gdip_save_png_image_to_file_or_stream (FILE *fp,
 error:
     if (png_ptr)
         png_destroy_write_struct (&png_ptr, info_ptr ? &info_ptr : NULL);
-    if (row_pointers)
-        GdipFree(row_pointers);
     if (pngsrc)
         GdipFree (pngsrc);
 
