@@ -48,7 +48,6 @@ gdip_texture_init (GpTexture *texture)
 	texture->matrix = cairo_matrix_create ();
 	texture->rectangle = NULL;
 	texture->pattern = NULL;
-	texture->changed = TRUE;
 }
 
 GpTexture*
@@ -410,8 +409,14 @@ gdip_texture_setup (GpGraphics *graphics, GpBrush *brush)
 	product = cairo_matrix_create ();
 	g_return_val_if_fail (product != NULL, OutOfMemory);
 
-	/* We do setup iff brush is changed or setup has not been done yet. */
-	if (texture->changed || (texture->pattern) == NULL) {
+	/* We create the new pattern for brush, if the brush is changed
+	 * or if pattern has not been created yet.
+	 */
+	if (texture->base.changed || (texture->pattern) == NULL) {
+
+		/* destroy the existing pattern */
+		if (texture->pattern)
+			cairo_pattern_destroy (texture->pattern);
 
 		switch (texture->wrapMode) {
 
@@ -442,7 +447,8 @@ gdip_texture_setup (GpGraphics *graphics, GpBrush *brush)
 
 	if (status == Ok) {
 		if (texture->pattern == NULL)
-			return GenericError;
+			status = GenericError;
+		else {
 
 		/* Cairo Bug: REPEAT and transformation do not go together.        */
 		/* To work around this we can create an intermediate surface first */
@@ -474,23 +480,21 @@ gdip_texture_setup (GpGraphics *graphics, GpBrush *brush)
 	
 		pattern = cairo_pattern_create_for_surface (temp);
 	*/
-		pattern = texture->pattern;
+			pattern = texture->pattern;
 
-		if (pattern != NULL) {
-			/* Use both the matrices */
-			cairo_matrix_multiply (product, texture->matrix, graphics->copy_of_ctm);
-			cairo_matrix_invert (product);
-			cairo_pattern_set_matrix (pattern, product);
-			cairo_set_pattern (ct, pattern);
-			/*	cairo_pattern_destroy (pattern); */
+			if (pattern != NULL) {
+				/* Use both the matrices */
+				cairo_matrix_multiply (product, texture->matrix, graphics->copy_of_ctm);
+				cairo_matrix_invert (product);
+				cairo_pattern_set_matrix (pattern, product);
+				cairo_set_pattern (ct, pattern);
+				/*	cairo_pattern_destroy (pattern); */
+			}
+			status = gdip_get_status (cairo_status (ct));
+			/*	cairo_surface_destroy (temp); */
 		}
-		status = gdip_get_status (cairo_status (ct));
-		/*	cairo_surface_destroy (temp); */
 	}
 
-	if (status == Ok)
-		texture->changed = FALSE;
-
 	cairo_matrix_destroy (product);
 
 	return status;
@@ -513,9 +517,10 @@ gdip_texture_clone (GpBrush *brush, GpBrush **clonedBrush)
 	result->wrapMode = texture->wrapMode;
 	result->image = texture->image;
 	result->matrix = cairo_matrix_create ();
-	result->changed = texture->changed;
-	/* We are not cloning the pattern. Let the clone create its own pattern. */
+
+	/* Let the clone create its own pattern. */
 	result->pattern = NULL;
+	result->base.changed = TRUE;
 
 	if (result->matrix == NULL) {
 		GdipFree (result);
@@ -578,7 +583,7 @@ GdipCreateTexture (GpImage *image, GpWrapMode wrapMode, GpTexture **texture)
 	else
 		return NotImplemented;
 
-	imageSurface = cairo_surface_create_for_image (bitmap->data.Scan0,
+	imageSurface = cairo_surface_create_for_image ((char *)bitmap->data.Scan0,
 					bitmap->cairo_format, bitmap->data.Width,
 					bitmap->data.Height, bitmap->data.Stride);
 
@@ -642,7 +647,7 @@ GdipCreateTexture2I (GpImage *image, GpWrapMode wrapMode, int x, int y, int widt
 		  || (bmpWidth < (x + width)) || (bmpHeight < (y + height)))
 	     return OutOfMemory;
 
-	original = cairo_surface_create_for_image (bitmap->data.Scan0, 
+	original = cairo_surface_create_for_image ((char *)bitmap->data.Scan0, 
 			bitmap->cairo_format, x + width, y + height, bitmap->data.Stride);
 
 	g_return_val_if_fail (original != NULL, OutOfMemory);
@@ -714,18 +719,19 @@ GdipGetTextureTransform (GpTexture *texture, GpMatrix *matrix)
 	g_return_val_if_fail (texture != NULL, InvalidParameter);
 	g_return_val_if_fail (matrix != NULL, InvalidParameter);
 
-	*matrix = *(texture->matrix);
+	cairo_matrix_copy(matrix, texture->matrix);
 	return Ok;
 }
 
 GpStatus
-GdipSetTextureTransform (GpTexture *texture, GpMatrix *matrix)
+GdipSetTextureTransform (GpTexture *texture, GDIPCONST GpMatrix *matrix)
 {
 	g_return_val_if_fail (texture != NULL, InvalidParameter);
 	g_return_val_if_fail (matrix != NULL, InvalidParameter);
 
-	*(texture->matrix) = *matrix;
-	texture->changed = TRUE;
+	texture->matrix = cairo_matrix_create();
+	cairo_matrix_copy(texture->matrix, matrix);
+	texture->base.changed = TRUE;
 
 	return Ok;
 }
@@ -740,7 +746,7 @@ GdipResetTextureTransform (GpTexture *texture)
 	status = cairo_matrix_set_identity (texture->matrix);
 	s = gdip_get_status (status);
 	if (s == Ok)
-		texture->changed = TRUE;
+		texture->base.changed = TRUE;
 	return s;
 }
 
@@ -754,7 +760,7 @@ GdipMultiplyTextureTransform (GpTexture *texture, GpMatrix *matrix, GpMatrixOrde
 	/* FIXME: How to take care of rotation here ? */
 	status = GdipMultiplyMatrix (texture->matrix, matrix, order);
 	if (status == Ok)
-		texture->changed = TRUE;
+		texture->base.changed = TRUE;
 	return status;
 }
 
@@ -766,7 +772,7 @@ GdipTranslateTextureTransform (GpTexture *texture, float dx, float dy, GpMatrixO
 
 	status = GdipTranslateMatrix (texture->matrix, dx, dy, order);
 	if (status == Ok)
-		texture->changed = TRUE;
+		texture->base.changed = TRUE;
 	return status;
 }
 
@@ -778,7 +784,7 @@ GdipScaleTextureTransform (GpTexture *texture, float sx, float sy, GpMatrixOrder
 
 	status = GdipScaleMatrix (texture->matrix, sx, sy, order);
 	if (status == Ok)
-		texture->changed = TRUE;
+		texture->base.changed = TRUE;
 	return status;
 }
 
@@ -786,6 +792,7 @@ GpStatus
 GdipRotateTextureTransform (GpTexture *texture, float angle, GpMatrixOrder order)
 {
 	GpStatus status;
+	GpPointF axis;
 
 	g_return_val_if_fail (texture != NULL, InvalidParameter);
 
@@ -794,7 +801,6 @@ GdipRotateTextureTransform (GpTexture *texture, float angle, GpMatrixOrder order
 	 * of rotation was suggested by Carl.
 	 */
 
-	GpPointF axis;
 	/* Our pattern size is 2*(texture->rect) hence its centre is following */
 	axis.X = texture->rectangle->Width;
 	axis.Y = texture->rectangle->Height;
@@ -802,7 +808,7 @@ GdipRotateTextureTransform (GpTexture *texture, float angle, GpMatrixOrder order
 	if ((status = GdipTranslateMatrix (texture->matrix, -axis.X, -axis.Y, order)) == Ok)
 		if ((status = GdipRotateMatrix (texture->matrix, angle, order)) == Ok)
 			if ((status = GdipTranslateMatrix (texture->matrix, axis.X, axis.Y, order)) == Ok)
-				texture->changed = TRUE;
+				texture->base.changed = TRUE;
 	return status;
 }
 
@@ -812,7 +818,7 @@ GdipSetTextureWrapMode (GpTexture *texture, GpWrapMode wrapMode)
 	g_return_val_if_fail (texture != NULL, InvalidParameter);
 
 	texture->wrapMode = wrapMode;
-	texture->changed = TRUE;
+	texture->base.changed = TRUE;
 
 	return Ok;
 }
