@@ -38,6 +38,7 @@
 #include <byteswap.h>
 #endif
 
+#include "gdipImage.h"
 #include "tiffcodec.h"
 
 #ifdef HAVE_LIBTIFF
@@ -63,6 +64,62 @@ typedef struct {
         SizeDelegate sizeFunc;
 } gdip_tiff_clientData;
 
+tsize_t 
+gdip_tiff_fileread (thandle_t clientData, tdata_t buffer, tsize_t size)
+{
+	return (tsize_t)fread(buffer, 1, size, (FILE*)clientData);
+}
+
+tsize_t 
+gdip_tiff_fileread_none (thandle_t clientData, tdata_t buffer, tsize_t size)
+{
+	return 0;
+}
+
+tsize_t 
+gdip_tiff_filewrite (thandle_t clientData, tdata_t buffer, tsize_t size)
+{
+	return (tsize_t)fwrite (buffer, 1, size, (FILE*)clientData);
+}
+
+toff_t 
+gdip_tiff_fileseek (thandle_t clientData, toff_t offSet, int whence)
+{
+	int seek_ok = fseek ((FILE*)clientData, offSet, whence);
+	if (seek_ok == 0)
+		return offSet;
+
+	return -1;
+}
+
+int 
+gdip_tiff_fileclose (thandle_t clientData)
+{
+	// This is a TIFF cleanup function; but we own the FILE* and close it in image.c so this is a null op
+	return 0;
+}
+
+toff_t 
+gdip_tiff_filesize (thandle_t clientData)
+{
+	long cur_pos = ftell ((FILE*)clientData);
+	fseek ((FILE*)clientData, 0, SEEK_END);
+	long ret = ftell ((FILE*)clientData);
+	fseek ((FILE*)clientData, cur_pos, SEEK_SET);
+	return (toff_t)ret;
+}
+
+int
+gdip_tiff_filedummy_map (thandle_t clientData, tdata_t *phase, toff_t* size)
+{
+	return 0;
+}
+
+void
+gdip_tiff_filedummy_unmap (thandle_t clientData, tdata_t base, toff_t size)
+{
+}
+
 tsize_t 
 gdip_tiff_read (thandle_t clientData, tdata_t buffer, tsize_t size)
 {
@@ -137,61 +194,74 @@ GpStatus
 gdip_save_tiff_image (TIFF* tiff, GpImage *image, GDIPCONST EncoderParameters *params)
 {
 	GpBitmap *bitmap = (GpBitmap *) image;	
-	unsigned char *buf = NULL;
-	int i, linebytes;
+	int i, j, k, l, linebytes;
 	guint32 *r32 = NULL;
 	size_t npixels;
+	int dimensionCount = 0;
+	int frameCount = 0;
+	BitmapData data;
+	int currPage = 0; /*First page wud already have been set */
+	int totalPages = 0;
 					
 	if (!tiff)
 		return InvalidParameter;		
-		
-	TIFFSetField (tiff, TIFFTAG_IMAGEWIDTH, bitmap->data.Width);  
-	TIFFSetField (tiff, TIFFTAG_IMAGELENGTH, bitmap->data.Height); 
-	TIFFSetField (tiff, TIFFTAG_SAMPLESPERPIXEL, 4); /* Hardcoded 32bbps*/
-	TIFFSetField (tiff, TIFFTAG_BITSPERSAMPLE, 8);  
-	TIFFSetField (tiff, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);   
-	TIFFSetField (tiff, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG); 
-	TIFFSetField (tiff, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB);
-	linebytes =  bitmap->data.Stride;     	
-	
-	if (TIFFScanlineSize (tiff) < linebytes) 
-		buf =(unsigned char *)_TIFFmalloc (linebytes); 
-	else 
-		buf = (unsigned char *)_TIFFmalloc (TIFFScanlineSize (tiff)); 
 
-	TIFFSetField (tiff, TIFFTAG_ROWSPERSTRIP, TIFFDefaultStripSize (tiff, linebytes)); 
-
-	/* We store data in ABGR format, but tiff stores it in ARGB, 
-	 * so convert before writing it. 
-	 */	 
-	npixels = bitmap->image.width * bitmap->image.height;
-	r32 = bitmap->data.Scan0;
-	/* flip bytes */
-	for (i = 0; i < npixels; i++) {
-		*r32 = (*r32 & 0xff000000) | ((*r32 & 0x00ff0000) >> 16) |
-				(*r32 & 0x0000ff00) | ((*r32 & 0x000000ff) << 16);
-		r32++;
-	}
+	dimensionCount = image->frameDimensionCount; 
+	for (j = 0; j < dimensionCount; j++)
+		totalPages += image->frameDimensionList [j].count;
 	
-	/*write data*/
-	for (i = 0; i < bitmap->data.Height; i++) {
-		if (TIFFWriteScanline (tiff, bitmap->data.Scan0 + i *bitmap->data.Stride, i, 0) < 0) 
-			break;
-	}
-		
-	/* image data might be reused again, so convert back to ABGR*/
-	r32 = bitmap->data.Scan0;
-	for (i = 0; i < npixels; i++) {
-		*r32 = (*r32 & 0xff000000) | ((*r32 & 0x00ff0000) >> 16) |
-				(*r32 & 0x0000ff00) | ((*r32 & 0x000000ff) << 16);
-		r32++;
-	}
-	
-	TIFFClose (tiff); 
-		
-	if (buf) 
-		_TIFFfree (buf);
+	for (j = 0; j < dimensionCount; j++) {
+                frameCount = image->frameDimensionList [j].count;
+                for (k = 0; k < frameCount; k++) {
+			if (k > 0 || j > 0) {
+				TIFFCreateDirectory (tiff);
+			}
+			
+			if (totalPages > 1) {
+				TIFFSetField (tiff, TIFFTAG_SUBFILETYPE, FILETYPE_PAGE);
+				TIFFSetField (tiff, TIFFTAG_PAGENUMBER, currPage++, totalPages);
+			}
+			data = image->frameDimensionList [j].frames [k];
+			TIFFSetField (tiff, TIFFTAG_IMAGEWIDTH, data.Width);  
+			TIFFSetField (tiff, TIFFTAG_IMAGELENGTH, data.Height); 
+			TIFFSetField (tiff, TIFFTAG_BITSPERSAMPLE, 8);  
+			TIFFSetField (tiff, TIFFTAG_COMPRESSION, COMPRESSION_NONE);  
+			TIFFSetField (tiff, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB);
+			TIFFSetField (tiff, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);   
+			TIFFSetField (tiff, TIFFTAG_SAMPLESPERPIXEL, 4); /* Hardcoded 32bbps*/
 
+			linebytes =  data.Stride;     	
+    			TIFFSetField (tiff, TIFFTAG_ROWSPERSTRIP, TIFFDefaultStripSize (tiff, linebytes)); 
+
+			TIFFSetField (tiff, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG); 
+			TIFFSetField (tiff, TIFFTAG_EXTRASAMPLES, EXTRASAMPLE_UNSPECIFIED);
+				
+			{
+				guchar *row_pointer = GdipAlloc (image->width * 4);
+
+				for (i = 0; i < image->height; i++) {
+					for (l = 0; l < image->width; l++) {
+#ifdef WORDS_BIGENDIAN
+						row_pointer[l*4+0] = *((guchar *)data.Scan0 + (data.Stride * i) + (l*4) + 1);
+						row_pointer[l*4+1] = *((guchar *)data.Scan0 + (data.Stride * i) + (l*4) + 2);
+						row_pointer[l*4+2] = *((guchar *)data.Scan0 + (data.Stride * i) + (l*4) + 3);
+						row_pointer[l*4+3] = *((guchar *)data.Scan0 + (data.Stride * i) + (l*4) + 0);
+#else
+						row_pointer[l*4+0] = *((guchar *)data.Scan0 + (data.Stride * i) + (l*4) + 2);
+						row_pointer[l*4+1] = *((guchar *)data.Scan0 + (data.Stride * i) + (l*4) + 1);
+						row_pointer[l*4+2] = *((guchar *)data.Scan0 + (data.Stride * i) + (l*4) + 0);
+						row_pointer[l*4+3] = *((guchar *)data.Scan0 + (data.Stride * i) + (l*4) + 3);
+#endif
+					}
+					TIFFWriteScanline (tiff, row_pointer, i, 0);
+				}
+				GdipFree (row_pointer);
+			}
+			TIFFWriteDirectory (tiff);			
+		}
+	}
+				
+	TIFFClose (tiff); 		
 	return Ok;
 }
 
@@ -200,93 +270,122 @@ gdip_load_tiff_image (TIFF *tif, GpImage **image)
 {
 	GpBitmap *img = NULL;
 	char *raster = NULL;
+	int dirCount = 0;
+	int j=0;
+	BitmapData data;
+
+	data.ByteCount = data.Top = data.Left = 0;
+	data.Bytes = NULL;
 
 	if (tif) {
+
 		TIFFRGBAImage tifimg;
 		char emsg[1024];
 
-		if (TIFFRGBAImageBegin (&tifimg, tif, 0, emsg)) {
-			size_t npixels;
+		/*Count the no of image frames present in input*/
+		do {
+			dirCount++;
+		} while (TIFFReadDirectory (tif));
+		
+		img = gdip_bitmap_new ();
+		img->image.type = imageBitmap;
+		/* libtiff expands stuff out to ARGB32 for us if we use this interface */
+		img->image.pixFormat = Format32bppArgb;
+		img->cairo_format = CAIRO_FORMAT_ARGB32;
 
-			img = gdip_bitmap_new ();
-			img->image.type = imageBitmap;
-			img->image.graphics = 0;
-			img->image.width = tifimg.width;
-			img->image.height = tifimg.height;
-			/* libtiff expands stuff out to ARGB32 for us if we use this interface */
-			img->image.pixFormat = Format32bppArgb;
-			img->cairo_format = CAIRO_FORMAT_ARGB32;
-			img->data.Stride = tifimg.width * 4;
-			img->data.PixelFormat = img->image.pixFormat;
-			img->data.Width = img->image.width;
-			img->data.Height = img->image.height;
+		img->image.frameDimensionCount = 1;
+		img->image.frameDimensionList = (FrameInfo *) GdipAlloc (sizeof (FrameInfo));
+		img->image.frameDimensionList [0].count = dirCount;
+		memcpy (&(img->image.frameDimensionList [0].frameDimension), 
+				&gdip_image_frameDimension_page_guid, sizeof (CLSID));
+		img->image.frameDimensionList [0].frames = (BitmapData *)
+						GdipAlloc (sizeof (BitmapData) * dirCount);
+
+		/*loop within all the directories and extract the frame info*/
+		for (j = 0; j < dirCount; j++) {
+			if (!TIFFSetDirectory (tif, j))
+				goto error;
 			
-			npixels = tifimg.width * tifimg.height;
-			/* Note that we don't use _TIFFmalloc */
-			raster = GdipAlloc (npixels * sizeof (guint32));
-			if (raster != NULL) {
-				/* Problem: the raster data returned here has the origin at bottom left,
-				* not top left.  The TIFF guys must be in cahoots with the OpenGL folks.
-				*
-				* Then, to add insult to injury, it's in ARGB format, not ABGR.
-				*/
-				if (TIFFRGBAImageGet (&tifimg, (uint32*) raster, tifimg.width, tifimg.height)) { 
-					guchar *onerow = GdipAlloc (img->data.Stride);
-					guint32 *r32 = (guint32*)raster;
-					int i;
-				
-					/* flip raster */
-					for (i = 0; i < tifimg.height / 2; i++) {
-						memcpy (onerow, raster + (img->data.Stride * i), img->data.Stride);
-						memcpy (raster + (img->data.Stride * i),
-							raster + (img->data.Stride * (tifimg.height - i - 1)),
-								img->data.Stride);
-						memcpy (raster + (img->data.Stride * (tifimg.height - i - 1)),
-								onerow, img->data.Stride);
-					}
-					/* flip bytes */
+			if (TIFFRGBAImageBegin (&tifimg, tif, 0, emsg)) {
+				size_t npixels;
 
-					for (i = 0; i < npixels; i++) {
-						*r32 = (*r32 & 0xff000000) | ((*r32 & 0x00ff0000) >> 16) |
-								(*r32 & 0x0000ff00) | ((*r32 & 0x000000ff) << 16);
-						r32++;
-					}
+				data.Stride = tifimg.width * 4;
+				data.PixelFormat = img->image.pixFormat;
+				data.Width = tifimg.width;
+				data.Height = tifimg.height;
+				data.Reserved = GBD_OWN_SCAN0;
+			
+				npixels = tifimg.width * tifimg.height;
+				/* Note that we don't use _TIFFmalloc */
+				raster = GdipAlloc (npixels * sizeof (guint32));
+				if (raster != NULL) {
+					/* Problem: the raster data returned here has the origin at bottom left,
+					* not top left.  The TIFF guys must be in cahoots with the OpenGL folks.
+					*
+					* Then, to add insult to injury, it's in ARGB format, not ABGR.
+					*/
+					if (TIFFRGBAImageGet (&tifimg, (uint32*) raster, tifimg.width, tifimg.height)) { 
+						guchar *onerow = GdipAlloc (data.Stride);
+						guint32 *r32 = (guint32*)raster;
+						int i;
 				
-					img->data.Scan0 = raster;
-					GdipFree (onerow);
-					img->data.Reserved = GBD_OWN_SCAN0;
+						/* flip raster */
+						for (i = 0; i < tifimg.height / 2; i++) {
+							memcpy (onerow, raster + (data.Stride * i), data.Stride);
+							memcpy (raster + (data.Stride * i), 
+									raster + (data.Stride * (tifimg.height - i - 1)),
+													data.Stride);
+							memcpy (raster + (data.Stride * (tifimg.height - i - 1)),
+												onerow, data.Stride);
+						}
+						/* flip bytes */
+
+						for (i = 0; i < npixels; i++) {
+							*r32 = (*r32 & 0xff000000) | ((*r32 & 0x00ff0000) >> 16) |
+									(*r32 & 0x0000ff00) | ((*r32 & 0x000000ff) << 16);
+							r32++;
+						}	
+						
+						GdipFree (onerow);
+
+						data.Scan0 = (byte *)raster;
+						img->image.frameDimensionList [0].frames [j] = data;
+					} else {
+						goto error;
+					} /*end if (TIFFRGBAImageGet */					
+				} else {
+					goto error;
+				} /*end if (raster != NULL) */
 				
-					img->image.surface = cairo_surface_create_for_image (raster,
-											img->cairo_format,
-											img->image.width,
-											img->image.height,
-											img->data.Stride);
-					img->image.horizontalResolution = 0;
-					img->image.verticalResolution = 0;
-				
-					img->image.imageFlags = ImageFlagsReadOnly |
-								ImageFlagsHasRealPixelSize |
-								ImageFlagsColorSpaceRGB;
-				
-					img->image.propItems = NULL;
-					img->image.palette = NULL;
-				}
+				TIFFRGBAImageEnd (&tifimg);				
 			} else {
 				goto error;
-			}
+			} /*end if (TIFFRGBAImageBegin())*/
+
+		} /*End For Loop for reading directories */
 		
-			TIFFRGBAImageEnd (&tifimg);
-		} else {
-			goto error;
-		}
-	
 		TIFFClose (tif);
+	
+		img->data = img->image.frameDimensionList [0].frames [0];
+		img->image.width = img->data.Width;
+		img->image.height = img->data.Height;
+		img->image.surface = cairo_surface_create_for_image ((char *)img->data.Scan0, 
+						img->cairo_format, img->image.width, 
+						img->image.height, img->data.Stride);
+		img->image.horizontalResolution = 0;
+		img->image.verticalResolution = 0;
+			
+		img->image.imageFlags = ImageFlagsReadOnly | ImageFlagsHasRealPixelSize | 
+								ImageFlagsColorSpaceRGB;
+				
+		img->image.propItems = NULL;
+		img->image.palette = NULL;
+
+		*image = (GpImage *) img;
+		return Ok;
 	} else {
 		goto error;
-	}
-
-	*image = (GpImage *) img;
-	return Ok;
+	} /*end if (tif) */
 
 error:
 	
@@ -308,7 +407,9 @@ gdip_load_tiff_image_from_file (FILE *fp, GpImage **image)
 {
 	TIFF *tif = NULL;
 	
-	tif = TIFFFdOpen (fileno (fp), "lose.tif", "r");
+	tif = TIFFClientOpen("lose.tif", "r", (thandle_t) fp, gdip_tiff_fileread, 
+				gdip_tiff_filewrite, gdip_tiff_fileseek, gdip_tiff_fileclose, 
+				gdip_tiff_filesize, gdip_tiff_filedummy_map, gdip_tiff_filedummy_unmap);
 	return gdip_load_tiff_image (tif, image);
 }
 
@@ -318,6 +419,7 @@ gdip_save_tiff_image_to_file (FILE *fp, GpImage *image, GDIPCONST EncoderParamet
 	TIFF* tiff;
 	
 	tiff = TIFFFdOpen (fileno (fp), "lose.tif", "w");
+
 	if (!tiff)
 		return FileNotFound;		
 		
@@ -341,7 +443,7 @@ gdip_load_tiff_image_from_stream_delegate (GetBytesDelegate getBytesFunc,
 	clientData.closeFunc = closeFunc;
 	clientData.sizeFunc = sizeFunc;
 	
-	tif = TIFFClientOpen("lose.tif", "r", &clientData, gdip_tiff_read, 
+	tif = TIFFClientOpen("lose.tif", "r", (thandle_t) &clientData, gdip_tiff_read, 
 				gdip_tiff_write, gdip_tiff_seek, gdip_tiff_close, 
 				gdip_tiff_size, gdip_tiff_dummy_map, gdip_tiff_dummy_unmap);
 	
@@ -366,7 +468,7 @@ gdip_save_tiff_image_to_stream_delegate (GetBytesDelegate getBytesFunc,
 	clientData.closeFunc = closeFunc;
 	clientData.sizeFunc = sizeFunc;
 	
-	tiff = TIFFClientOpen("lose.tif", "w", &clientData, gdip_tiff_read_none, 
+	tiff = TIFFClientOpen("lose.tif", "w", (thandle_t) &clientData, gdip_tiff_read_none, 
 			gdip_tiff_write, gdip_tiff_seek, gdip_tiff_close, 
 			gdip_tiff_size, gdip_tiff_dummy_map, gdip_tiff_dummy_unmap);
 	if (!tiff)
